#!/usr/bin/perl

# Script to run cross-validated HMM models on the data.
# Sergey Kirshner, 01/09/2004
# Revised: 05/03/2004 by Sergey Kirshner
# Revised: 10/14/2004 by Sergey Kirshner

use POSIX qw(floor);

$num_args=$#ARGV+1;

if( $num_args==0 ) {
  # No arguments passed -- displaying help
  print "mvnhmm-batch-learn-params.perl <model type> <output discrete dimensionality> <output distribution type> <number of sequences> <length of each sequence> <input dimensionality> <number of sequences to leave out (with cross-validation)> <precision> <number of restarts> <minimum hidden states> <maximum hidden states>\n";
  print "Model types: hmm, mixture, nhmm, nmixture\n";
  print "Output discrete dimensionality: >0\n";
  print "Output distribution type: independent, chowliu, condchowliu, chains, mixind, mixchains, nested, maxent, BNmaxent, BNcondmaxent\n";
  print "Number of sequences: >0\n";
  print "Length of each sequence: >0\n";
  print "Input data dimensionality: >0 (use 0 for models not using input data)\n";
  print "Number of sequences to leave out with cross-validation: >=0.  Use 0 for no cross-validation\n";
  print "EM precision: >0\n";
  print "Number of restarts: >0\n";
  print "Minimum number of hidden states: >0 (irrelevant for stateless)\n";
  print "Maximum number of hidden states: >0 (irrelevant for stateless)\n";

  print "\n";
  print "In order for the script to run, two directories must be present in the execution directory: 'data' and 'outputs'.  The results of the runs will be placed into 'outputs' directory.  'data' directory must contain the dataset in the file 'data' (link to a data set), and an input data set (for models requiring inputs) in the file 'input_data' (link is fine).\n";
  exit( 0 );
} elsif( $num_args<11 ) {
  # Not enough arguments passed
  print "Need eleven(!) arguments passed\n";
  exit( -1 );
} else {
  # Collecting necessary parameters from the command line

  # Model type
  $model_type=$ARGV[0];
  print "Model type: ";
  if( $model_type eq "hmm" ) {
    print "HMM\n";
  } elsif( $model_type eq "nhmm" ) {
    print "non-homogeneous HMM\n";
  } elsif( $model_type eq "mixture" ) {
    print "mixture model\n";
  } elsif( $model_type eq "nmixture" ) {
    print "non-homogeneous mixture model\n";
  } else {
    print "Model type: unrecognized\n";
    exit( -1 );
  }

  # Output dimensionality
  $output_dim=$ARGV[1];
  if( $output_dim<1 ) {
    print "Invalid dimensionality of the data set: $output_dim\n";
    exit( -1 );
  }
  print "Dimensionality of the data: $output_dim\n";

  # Output distribution type
  $output_dist=$ARGV[2];
  print "Output distribution type: ";
  if( $output_dist eq "independent" ) {
    print "conditionally independent Bernoulli\n";
  } elsif( $output_dist eq "chowliu" ) {
    print "Chow-Liu tree\n";
  } elsif( $output_dist eq "condchowliu" ) {
    print "conditional Chow-Liu forest\n";
  } elsif( $output_dist eq "chains" ) {
    print "double-chains (conditional Bernoulli)\n";
  } elsif( $output_dist eq "mixind" ) {
    print "mixtures of 2 CI Bernoulli\n";
  } elsif( $output_dist eq "mixchains" ) {
    print "mixtures of 2 AR Bernoulli\n";
  } elsif( $output_dist eq "nested" ) {
    print "nested-divide distribution\n";
  } elsif( $output_dist eq "maxent" ) {
    print "full bivariate binary MaxEnt\n";
  } elsif( $output_dist eq "BNmaxent" ) {
    print "BN bivariate binary MaxEnt\n";
  } elsif( $output_dist eq "BNcondmaxent" ) {
    print "BN bivariate binary conditional MaxEnt\n";
  } else {
    print "unrecognized\n";
    exit( -1 );
  }

  # Number of sequences
  $num_seqs=$ARGV[3];
  if( $num_seqs<1 ) {
    print "Invalid number of sequences: $num_seqs\n";
    exit( -1 );
  }
  print "Number of sequences: $num_seqs\n";

  # Length of each sequence
  $seq_length=$ARGV[4];
  if( $seq_length<1 ) {
    print "Invalid length of each sequence: $seq_length\n";
    exit( -1 );
  }
  print "Length of each sequence: $seq_length\n";

  # Input dimensionality
  $input_dim=$ARGV[5];
  if( $model_type eq "nhmm" | $model_type eq "nmixture" ) {
    if( $input_dim<1 ) {
      print "Invalid dimensionality of the input data: $input_dim\n";
      exit( -1 );
    }
    print "Input data dimensionality: $input_dim\n";
  } else {
    print "Input data dimensionality: irrelevant\n";
  }

  # Number of sequences to leave out for cross-validation
  $xval_out=$ARGV[6];
  if( $xval_out<0 ) {
    print "Invalid number of sequences to leave out: $xval_out\n";
    exit( -1 );
  }
  if( $xval_out==0 ) {
    print "Cross-validation type: none\n";
    $xval_type="none\n";
  } elsif( $num_seqs/$xval_out<2 ) {
    print "Number of sequences to leave-out is too large -- should be at most a half of all examples\n";
    exit( -1 );
  } else {
    print "Cross-validation type: leave-$xval_out-out\n";
    $xval_type="leave_n_out";
  }

  # EM precision threshold
  $em_precision=$ARGV[7];
  if( $em_precision<=0 ) {
    print "Invalid EM precision: $em_precision\n";
    exit( -1 );
  } else {
    print "EM precision: $em_precision\n";
  }

  # Number of restarts
  $num_restarts=$ARGV[8];
  if( $num_restarts<=0 ) {
    print "Invalid number of restarts: $num_restarts\n";
    exit( -1 );
  } else {
    print "Number of restarts: $num_restarts\n";
  }

  # Range for the number of hidden states
  $min_states=$ARGV[9];
  $max_states=$ARGV[10];
  if( !($min_states>0 & $max_states>0) ) {
    print "Invalid number of hidden states: $min_states $max_states\n";
    exit( -1 );
  } elsif( $min_states>$max_states ) {
    print "The number of minimum hidden states is larger than the number of maximum hidden states\n";
    exit( -1 );
  } else {
    print "The range for the number of hidden states: $min_states to $max_states\n";
  }
}

# Removing previously created parameter file
`rm -f params_current`;

for( $num_states=$min_states; $num_states<=$max_states; $num_states=$num_states+1 ) {
  # Looping over the number of value for hidden states

  # Building up the parameter file

  # Creating a file
  `touch params_current`;

  # Number of hidden states
  `echo 'num_states $num_states' >> params_current`;

  # Type of the model
  `echo 'model_type $model_type' >> params_current`;

  # Action type
  `echo 'action learn' >> params_current`;

  `echo 'num_discrete_data_components $output_dim' >> params_current`;
  `echo 'num_real_data_components 0' >> params_current`;

  `echo 'emission' >> params_current`;
  if( $output_dist eq 'independent' ) {
    # Number of emission components
    `echo 'independent $output_dim' >> params_current`;
	
    # Types of parameters for emission components
    `echo 'bernoulli 2' >> params_current`;
  } elsif( $output_dist eq 'chains' ) {
    # Number of emission components
    `echo 'independent $output_dim' >> params_current`;

    # Types of parameters for emission components
    `echo 'chain-bernoulli 2' >> params_current`;
  } elsif( $output_dist eq 'chowliu' ) {
    `echo 'chow-liu $output_dim 2' >> params_current`;
  } elsif( $output_dist eq 'condchowliu' ) {
    `echo 'conditional-chow-liu $output_dim 2' >> params_current`;
  } elsif( $output_dist eq 'maxent' ) {
    `echo 'maxent-full $output_dim' >> params_current`;
  } elsif( $output_dist eq 'BNmaxent' ) {
    `echo 'BN-maxent $output_dim' >> params_current`;
  } elsif( $output_dist eq 'BNcondmaxent' ) {
    `echo 'BN-cond-maxent $output_dim' >> params_current`;
  }	elsif( $output_dist eq 'mixind' ) {
    `echo 'mixture 2' >> params_current`;
    `echo 'independent $output_dim' >> params_current`;
    `echo 'bernoulli 2' >> params_current`;
  }	elsif( $output_dist eq 'mixchains' ) {
    `echo 'mixture 2' >> params_current`;
    `echo 'independent $output_dim' >> params_current`;
    `echo 'chain-bernoulli 2' >> params_current`;
  } elsif( $output_dist eq 'nested' ) {
    `echo 'nested-divide $output_dim 2' >> params_current`;
  }

  # Number of input components
  if( $model_type eq 'nhmm' ) {
    # Types of parameters for input distributions
    `echo 'input_distribution' >> params_current`;
    `echo 'transition-logistic ${num_states} $input_dim' >> params_current`;
  } elsif( $model_type eq 'nmixture' ) {
    # Types of parameters for input distributions
    `echo 'input_distribution' >> params_current`;
    `echo 'logistic ${num_states} $input_dim' >> params_current`;
  }
	
  # Define the suffix for filenames
  $state_suffix="_states${num_states}";

  # Data file
  # Data file is stored in data/data
  `echo 'data data/data' >> params_current`;

  # Number of data sequences
  # Number of data sequences just as passed
  `echo 'num_data_sequences $num_seqs' >> params_current`;

  # Length of each data sequence
  `echo 'data_sequence_length $seq_length' >> params_current`;

  if( $model_type eq 'nhmm' | $model_type eq 'nmixture' ) {
    # Need input data information

    # Not analyzing data

    # Input data file
    `echo 'input_filename data/input_data' >> params_current`;

    # Number of input sequences
    `echo 'num_input_sequences $num_seqs' >> params_current`;

    # Length of each input sequence
    `echo 'input_sequence_length $seq_length' >> params_current`;

    # Number finite-valued components of the input set
    `echo 'num_discrete_input_components 0' >> params_current`;

    # Number real-valued components of the input set
    `echo 'num_real_input_components $input_dim' >> params_current`;
  }

  # Learning the parameters

  # Number of random restarts
  `echo 'num_restarts $num_restarts' >> params_current`;

  # EM precision
  `echo 'em_precision $em_precision' >> params_current`;

  # Cross-validation type
  `echo 'xval_type $xval_type' >> params_current`;
	
  if( $xval_type eq 'leave_n_out' ) {
    # Number of examples to leave-out for cross-validation
    `echo 'examples_out $xval_out' >> params_current`;
  }

  # Output file
  # Learning
  `echo 'output outputs/model${state_suffix}' >> params_current`;

  # Running the experiment
  # Learning
  print "Generating parameter file for ${num_states} hidden states\n";

  #  `/bin/nice -19 mvnhmm params_current`;

  # Deleting the parameter file
  `mv params_current params_states$num_states`;
}

