/*
 * Copyright (c) 2002-2005, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
 */

/* data.cpp */

#include "struct.h"

Sequence::Sequence( long length ) {
  /* Index variable(s) */
  long t;

  seq_length=length;
  entry=new DataPoint*[seq_length];
  /* Initializing entries */
  for( t=0; t<seq_length; t++ )
    entry[t]=NULL;

  num_total_entries=0;
  num_missing_discrete_entries=0;
  missing_discrete=NULL;
}

Sequence::~Sequence() {
  /* Index variable(s) */
  long i;
  
  for( i=0; i<seq_length; i++ )
    if( entry[i] )
      delete entry[i];

  if( num_missing_discrete_entries>0 )
    {
      for( i=0; i<num_missing_discrete_entries; i++ )
	delete missing_discrete[i];
      delete missing_discrete;
    }

  delete entry;
}

void Sequence::WriteToFile( FILE* output ) {
  /* Writing the sequence to a file */
  
  /* Index variable(s) */
  long i;

  for( i=0; i<seq_length; i++ )
    {
      if( entry[i] )
	entry[i]->WriteToFile( output );
      fprintf( output, "\n" );
    }

  return;
}

long Sequence::total_entries() {
  /* Index variable(s) */
  long t;

  if( num_total_entries==0 )
    for( t=0; t<seq_length; t++ )
      if( entry[t] )
	num_total_entries+=entry[t]->total_entries();

  return( num_total_entries );
}

double Sequence::weighted_total_entries( double* weight ) {
  /* Index variable(s) */
  long t;

  double total=0.0;

  for( t=0; t<seq_length; t++ )
    if( entry[t] )
      total+=weight[t]*(double)(entry[t]->total_entries());

  return( total );
}

void Sequence::CatalogueMissingEntries() {
  /* Index variable(s) */
  long t,i;

  long** temp;

  /* Allocating dummy array to store missing entries */
  temp=new long*[seq_length*entry[0]->ddim];

  if( num_missing_discrete_entries>0 )
    { /* Resetting the list of missing variables */
      for( i=0; i<num_missing_discrete_entries; i++ )
	delete missing_discrete[i];
      delete missing_discrete;

      num_missing_discrete_entries=0;
    } /* Resetting the list of missing variables */
    
  /* Counting the number of missing entries */
  for( t=0; t<seq_length; t++ )
    for( i=0; i<entry[t]->ddim; i++ )
      if( is_missing( entry[t]->ddata[i] ) )
	{
	  temp[num_missing_discrete_entries]=new long[2];
	  temp[num_missing_discrete_entries][0]=t;
	  temp[num_missing_discrete_entries][1]=i;
	  num_missing_discrete_entries++;
	
	}

  if( num_missing_discrete_entries>0 )
    {
      missing_discrete=new long*[num_missing_discrete_entries];
      for( i=0; i<num_missing_discrete_entries; i++ )
	missing_discrete[i]=temp[i];
    }

  delete temp;
}

Data::Data( long n_seqs, const char* tag, long is_p ) {
  /* Individual sequences are initialized in other functions:
     ReadData
  */

  /* Index variable(s) */
  long n;

  num_seqs=n_seqs;
  is_primary=is_p;

  /* Allocating the array of sequences */
  sequence=new Sequence*[num_seqs];
  for( n=0; n<num_seqs; n++ )
    sequence[n]=NULL;

  /* Name tag */
  strcpy( nametag, tag );
}

Data::~Data() {
  /* Index variable(s) */
  long i;

  if( is_primary )
    for( i=0; i<num_seqs; i++ )
      if( sequence[i] )
	delete sequence[i];

  delete sequence;
}

void Data::WriteToFile( FILE* output ) {
  /* Writing data to a file */

  /* Index variable(s) */
  long i;

  for( i=0; i<num_seqs; i++ )
    if( sequence[i] )
      sequence[i]->WriteToFile( output );

  fflush( output );

  return;
}

long Data::num_points() {
  /* Returns the number of vectors in the data set */
  long num=0;

  /* Index variables */
  long i;
  
  for( i=0; i<num_seqs; i++ )
    if( sequence[i] )
      num+=sequence[i]->seq_length;

  return( num );
}

long Data::total_entries() {
  long num_total_entries=0;

  /* Index variable(s) */
  long n;

  for( n=0; n<num_seqs; n++ )
    if( sequence[n] )
      num_total_entries+=sequence[n]->total_entries();
  
  return( num_total_entries );
}

double Data::weighted_total_entries( double** weight ) {
  double total=0.0;

  /* Index variable(s) */
  long n;

  for( n=0; n<num_seqs; n++ )
    if( sequence[n] )
      total+=sequence[n]->weighted_total_entries( weight[n] );
  
  return( total );
}

DataPoint::DataPoint( long dim1, long dim2 ) {
  /* !!! No safety checking !!! */
  ddim=dim1;
  rdim=dim2;

  /* Allocating data vectors */
  if( ddim>0 )
    ddata=new long[ddim];
  else
    ddata=NULL;

  if( rdim>0 )
    rdata=new double[rdim];
  else
    rdata=NULL;

  ddata_ind=NULL;
  rdata_ind=NULL;
}

DataPoint::~DataPoint() {
  /* Deallocating data structures */
  if( ddata )
    delete ddata;

  if( rdata )
    delete rdata;

  if( ddata_ind )
    delete ddata_ind;
  
  if( rdata_ind )
    delete rdata_ind;
}

void DataPoint::WriteToFile( FILE* output ) {
  /* Writing the entry into a file */
  
  /* Index variable(s) */
  long i;

  for( i=0; i<ddim; i++ )
    fprintf( output, "%ld ", ddata[i] );
  
  for( i=0; i<rdim; i++ )
    fprintf( output, "%0.12le ", rdata[i] );
  
  return;
}

long DataPoint::total_entries() {
  /* Calculating the total number of initialized components */
  long num_total_entries=0;
  
  /* Index variable(s) */
  long i;

  for( i=0; i<ddim; i++ )
    if( !is_missing( ddata[i] ) )
      num_total_entries++;

  for( i=0; i<rdim; i++ )
    if( !is_missing( rdata[i] ) )
      num_total_entries++;

  return( num_total_entries );
}

void Data::MissingEntryIndicators() {
  /* Creating indicators for missing entries */

  /* Index variable(s) */
  long n;

  for( n=0; n<num_seqs; n++ )
    sequence[n]->MissingEntryIndicators();

  return;
}

void Sequence::MissingEntryIndicators() {
  /* Creating indicators for missing entries */

  /* Index variable(s) */
  long t;

  for( t=0; t<seq_length; t++ )
    entry[t]->MissingEntryIndicators();

  return;
}

void DataPoint::MissingEntryIndicators() {
  /* Creating indicators for missing entries */

  /* Index variable(s) */
  long i;

  if( !ddata_ind )
    ddata_ind=new bool[ddim];

  for( i=0; i<ddim; i++ )
    if( is_missing( ddata[i] ) )
      ddata_ind[i]=1;
    else
      ddata_ind[i]=0;

  if( !rdata_ind )
    rdata_ind=new bool[rdim];

  for( i=0; i<rdim; i++ )
    if( is_missing( rdata[i] ) )
      rdata_ind[i]=1;
    else
      rdata_ind[i]=0;

  return;
}

long is_missing( double d ) {
  return( isnan( d ) );
}

long is_missing( long d ) {
  return( d==CONST_MISSING_VALUE );
}

double missing_value( double d ) {
  double a=0.0;
  //return( 0 );
  return( a/a );
  //  return( nan("") );
}

long missing_value( long d ) {
  return( CONST_MISSING_VALUE );
}
