/*
 * Copyright (c) 2002-2006, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* em.cpp */

#include "em.h"

void HMM::EM( Data* data, Data* input_data, double precision ) {
  /* Index variable(s) */
  long i,n,t,j;

  /* Accumulator variable(s) */
  double sum;

  /* Variables */
  double old_lp_normalized; // Previous log-posterior

  /* Previous emission variables */
  Distribution*** prev_emission;

  long iteration;           // Current iteration number

  prev_emission=new Distribution**[num_variables];
  for( i=0; i<num_variables; i++ )
    prev_emission[i]=new Distribution*[num_states[i]];

  is_done=0;

  /* Since likelihood can be estimated after performing an E-step, doing just that. */

  /* Allocating necessary structures */
  AllocateEMStructures( data, input_data );
  
  /* E-step */
  /* Need to perform forward-backward algorithm */
  EStep( data, input_data );

  iteration=0;
  if( em_verbose )
    printf( "%ld:\tll=%0.12le\tlp=%0.12le\tper-dimension-lp=%0.12le\n", 
	    iteration, ll, lp, lp_normalized );

  //  WriteToFile( stdout );  

  if( !finite( lp_normalized ) ) {
    if( em_verbose )
      {
	fprintf( stderr, "Log-posterior is infinite\n" );
	num_failed++;
	is_done=1;
	//	WriteToFile( stderr );
      }   
  }

  while( !is_done )
    { /* Main loop */
      iteration++;
 
      /* M-step */

      /* Storing the previous value of the log-likelihood */
      old_lp_normalized=lp_normalized;

      /* Calculating necessary summaries */
      CalculateSummaries( data );
      
      /* Updating emission probabilities */
      for( i=0; i<num_variables; i++ )
	for( j=0; j<num_states[i]; j++ )
	  {
	    prev_emission[i][j]=emission[i][j]->copy();
	    emission[i][j]->UpdateEmissionParameters( data, transition->uni_prob[i][j], norm_const[i][j] );
	  }

      /* Updating transition layer parameters */
      transition->UpdateTransitionParameters( data, input_data, iteration );

      /* E-step */
      /* Need to perform forward-backward algorithm */
      EStep( data, input_data );

      if( em_verbose )
	printf( "%ld:\tll=%0.12le\tlp=%0.12le\tper-dimension-lp=%0.12le\n", 
		iteration, ll, lp, lp_normalized );

	
      if( !finite( lp_normalized ) )
	{ /* Log-posterior is infinite -- run failed */
	  num_failed++;
	  is_done=1;
	} /* Log-posterior is infinite -- run failed */
      else if( lp_normalized-old_lp_normalized<precision )
	{ /* Checking whether the end condition is reached */
	  if( lp_normalized>=old_lp_normalized )
	    is_done=1;
	  else
	    {
	      if( em_verbose )
		printf( "Log-posterior went down\n" );

	      /* Analyzing why log-likelihood went down */
	      /* Most likely, one of the emissions changed to a worse set of parameters */
	      for( i=0; i<num_variables; i++ )
		for( j=0; j<num_states[i]; j++ )
		  {		    
		    if( prev_emission[i][j]->weighted_ll( data, transition->uni_prob[i][j] )>
			emission[i][j]->weighted_ll( data, transition->uni_prob[i][j] ) )
		      { /* Emission i got worse */
			delete emission[i][j];
			emission[i][j]=prev_emission[i][j]->copy();
		      } /* Emission i got worse */
		  }
	      /* Redoing the E-step */
	      EStep( data, input_data );

	      if( em_verbose )
		printf( "Corrected ll=%0.12le\tlp=%0.12le\tper-dimension lp=%0.12le\n", 
			ll, lp, lp_normalized );

	      if( !finite( lp_normalized ) )
		{ /* Log-posterior is infinite -- run failed */
		  num_failed++;
		  is_done=1;
		} /* Log-posterior is infinite -- run failed */
	      else if( lp_normalized-old_lp_normalized<precision )
		{ /* Checking whether the end condition is reached */
		  if( lp_normalized>=old_lp_normalized )
		    is_done=1;
		  else
		    {
		      if( em_verbose )
			printf( "Corrected log-posterior still went down\n" );
		      is_done=1;
		    }
		} /* Checking whether the end condition is reached */
	    }
	} /* Checking whether the end condition is reached */
	  
      for( i=0; i<num_variables; i++ ) 
	for( j=0; j<num_states[i]; j++ )
	  delete prev_emission[i][j];
    } /* Main loop */
  
  /* Deallocating arrays */
  DeallocateEMStructures( data, input_data );

  for( i=0; i<num_variables; i++ )
    delete prev_emission[i];
  delete prev_emission;
  
  return;
}

void HMM::AllocateEMStructures( Data* data, Data* input ) {
  /* Allocating structures needed for forward-backward algorithm */

  /* Index variables */
  long i,j,n;

  AllocateAlpha( data, input );
  transition->AllocateEMStructures( data, input );

  return;
}

void Distribution::AllocateEMStructures( Data* data, Data* input ) {
  /* Index variable(s) */
  long i,j,n;

  switch( type ) {
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
  case DIST_TRANSLOGISTIC:
    for( i=0; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	{
	  joint_prob[i][j]=new double*[data->num_seqs];
	  for( n=0; n<data->num_seqs; n++ )
	    joint_prob[i][j][n]=new double[data->sequence[n]->seq_length];
	}

    for( i=0; i<num_states; i++ )
      {
	bkd_update[i]=new double*[data->num_seqs];
	for( n=0; n<data->num_seqs; n++ )
	  bkd_update[i][n]=new double[data->sequence[n]->seq_length];
      }

    break;
  default:;
  }

  return;
}

void HMM::AllocateAlpha( Data* data, Data* input ) {
  /* Allocating structures needed to calculate scaled alphas for forward-backward algorithm */

  /* Index variable(s) */
  long i,n,j;

  /* Allocation block */

  /* Allocating array of P(R_nt|S_nt=ij) (or P(R_nt,X_nt|S_nt=ij) ) */  
  for( i=0; i<num_variables; i++ )
    for( j=0; j<num_states[i]; j++ )
      { /* For each state */
	pb[i][j]=new double*[data->num_seqs];
	for( n=0; n<data->num_seqs; n++ )
	  /* For each sequence */
	  pb[i][j][n]=new double[data->sequence[n]->seq_length];
      } /* For each state */
  
  transition->AllocateForwardPassStructures( data, input );

  return;
}

void Distribution::AllocateForwardPassStructures( Data* data, Data* input ) {
  /* Index variable(s) */
  long i,j,n;
  
  switch( type ){
  case DIST_BERNOULLI:
    /* Allocating array of prosterior univariate probabilities */
    for( i=0; i<num_states; i++ )
      {
	uni_prob[0][i]=new double*[data->num_seqs];
	for( n=0; n<data->num_seqs; n++ )
	  uni_prob[0][i][n]=new double[data->sequence[n]->seq_length];
      }

    /* Allocating inverse scaling factors */
    upd_scale=new double*[data->num_seqs];
    for( n=0; n<data->num_seqs; n++ )
      /* For each sequence */
      upd_scale[n]=new double[data->sequence[n]->seq_length];

    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    /* Allocating scaled forward pass updates */
    for( i=0; i<num_states; i++ )
      { /* For each state */
	fwd_update[i]=new double*[data->num_seqs];
	for( n=0; n<data->num_seqs; n++ )
	  fwd_update[i][n]=new double[data->sequence[n]->seq_length];
      } /* For each state */
      
    /* Allocating array of prosterior univariate probabilities */
    for( i=0; i<num_states; i++ )
      {
	uni_prob[0][i]=new double*[data->num_seqs];
	for( n=0; n<data->num_seqs; n++ )
	  uni_prob[0][i][n]=new double[data->sequence[n]->seq_length];
      }

    /* Allocating inverse scaling factors */
    upd_scale=new double*[data->num_seqs];
    for( n=0; n<data->num_seqs; n++ )
      /* For each sequence */
      upd_scale[n]=new double[data->sequence[n]->seq_length];

    break;
  case DIST_LOGISTIC:
    /* Allocating array of prosterior univariate probabilities */
    for( i=0; i<num_states; i++ )
      {
	uni_prob[0][i]=new double*[data->num_seqs];
	for( n=0; n<data->num_seqs; n++ )
	  uni_prob[0][i][n]=new double[data->sequence[n]->seq_length];
      }

    /* Allocating inverse scaling factors */
    upd_scale=new double*[data->num_seqs];
    for( n=0; n<data->num_seqs; n++ )
      /* For each sequence */
      upd_scale[n]=new double[data->sequence[n]->seq_length];

    /* Allocating the array of probabilities P(S_nt|S_n,t-1,X_nt) */
    for( i=0; i<num_states; i++ )
      {
	p_tr[i][0]=new double*[input->num_seqs];
	for( n=0; n<input->num_seqs; n++ )
	  p_tr[i][0][n]=new double[input->sequence[n]->seq_length];
      }
    
    /* Allocating the array of log-unnormalized P(S_nt|S_n,t-1,X_nt) */
    for( i=1; i<num_states; i++ )
      {
	log_un_prob[i][0]=new double*[input->num_seqs];
	for( n=0; n<input->num_seqs; n++ )
	  log_un_prob[i][0][n]=new double[input->sequence[n]->seq_length];
      }
    
    break;
  case DIST_TRANSLOGISTIC:
    /* Allocating scaled forward pass updates */
    for( i=0; i<num_states; i++ )
      { /* For each state */
	fwd_update[i]=new double*[data->num_seqs];
	for( n=0; n<data->num_seqs; n++ )
	  fwd_update[i][n]=new double[data->sequence[n]->seq_length];
      } /* For each state */

    /* Allocating array of prosterior univariate probabilities */
    for( i=0; i<num_states; i++ )
      {
	uni_prob[0][i]=new double*[data->num_seqs];
	for( n=0; n<data->num_seqs; n++ )
	  uni_prob[0][i][n]=new double[data->sequence[n]->seq_length];
      }

    /* Allocating inverse scaling factors */
    upd_scale=new double*[data->num_seqs];
    for( n=0; n<data->num_seqs; n++ )
      /* For each sequence */
      upd_scale[n]=new double[data->sequence[n]->seq_length];

      /* Allocating the array of probabilities P(S_nt|S_n,t-1,X_nt) */
      for( i=0; i<num_states; i++ )
	for( j=0; j<num_states; j++ )
	  {
	    p_tr[i][j]=new double*[input->num_seqs];
	    for( n=0; n<input->num_seqs; n++ )
	      /* For each sequence */
	      p_tr[i][j][n]=new double[input->sequence[n]->seq_length];
	  }

      /* Allocating the array of log-unnormalized P(S_nt|S_n,t-1,X_nt) */
      for( i=1; i<num_states; i++ )
	for( j=0; j<num_states; j++ )
	  {
	    log_un_prob[i][j]=new double*[input->num_seqs];
	    for( n=0; n<input->num_seqs; n++ )
	      /* For each sequence */
	      log_un_prob[i][j][n]=new double[input->sequence[n]->seq_length];
	    }

    break;
  default:;
  }
  
  return;
}

void HMM::EStep( Data* data, Data* input ) {

  /* Index variable(s) */
  long n,i;
#ifdef BLAH
  ll=0.0;
  for( n=0; n<data->num_seqs; n++ )
    {
      if( input )
	ll+=log_likelihood_sequence( data->sequence[n], input->sequence[n], n, 0 );
      else
	ll+=log_likelihood_sequence( data->sequence[n], NULL, n, 0 );
      transition->CalculateBackwardUpdatesSequence( data->sequence[n], n, pb );
    }
#endif
  lp=log_posterior( data, input );
  lp_normalized=lp/data->total_entries();

  for( n=0; n<data->num_seqs; n++ )
    transition->CalculateBackwardUpdatesSequence( data->sequence[n], n, pb );

  return;
}

double HMM::log_likelihood_sequence( Sequence* data, Sequence* input, long n, long start_index ) {
  double ll=0.0;

  /* Index variable(s) */
  long i,j,t;

  /* Temporary variable(s) */
  double sum;

  /*** Calculating needed values (forward) ***/

  /* Calculating emission probabilities */
  CalculateProbEmissionsSequence( data, n, start_index );

  /* Calculating P(S_nt|S_n,t-1, X_nt) */
  transition->CalculateTransitionValuesSequence( input, n, start_index );

  /* Need to perform forward pass -- calculating scaled alpha values */
  transition->CalculateForwardUpdatesSequence( data, n, start_index, pb );
   
  /* Calculating the log-likelihood using scaled alphas */
  for( t=0; t<data->seq_length; t++ )
    ll-=log(transition->upd_scale[n][t]);

  return( ll );
}

void HMM::CalculateProbEmissionsSequence( Sequence* data, long n, long start_index ) {
  /* Index variable(s) */
  long i,j,t;

  /* Initializing array of P(R_nt|S_nt=ij) */
  for( t=start_index; t<data->seq_length; t++ )
    if( t==0 )
      for( i=0; i<num_variables; i++ )	
	for( j=0; j<num_states[i]; j++ )
	  pb[i][j][n][t]=emission[i][j]->prob(data->entry[t], NULL);
    else
      for( i=0; i<num_variables; i++ )	
	for( j=0; j<num_states[i]; j++ )
	  pb[i][j][n][t]=emission[i][j]->prob(data->entry[t], data->entry[t-1]);

  return;
}

void Distribution::CalculateTransitionValuesSequence( Sequence* input, long n, long start_index ) {
  /* Index variable(s) */
  long i,j,t;

  /* Temporary variable(s) */
  double sum;
  double max;

  /* Calculating P(S_t|S_{t-1}, X_t ) */

  /* !!! Hack speeding up the calculation of transition probabilities !!! */
  switch( type ) {
  case DIST_LOGISTIC:
    for( t=start_index; t<input->seq_length; t++ )
      {
	/* Calculating P(S_nt|X_nt) */
	max=0.0;
	for( i=1; i<num_states; i++ )
	  {
	    log_un_prob[i][0][n][t]=log_un_probLogistic( input->entry[t], i, -1 );
	    if( max<log_un_prob[i][0][n][t] )
	      max=log_un_prob[i][0][n][t];
	  }

	sum=exp(-max);
	for( i=1; i<num_states; i++ )
	  sum+=exp(log_un_prob[i][0][n][t]-max);	    
	
	p_tr[0][0][n][t]=exp(-max)/sum;
	for( i=1; i<num_states; i++ )
	  p_tr[i][0][n][t]=exp(log_un_prob[i][0][n][t]-max)/sum;
      }

    break;
  case DIST_TRANSLOGISTIC:
    if( start_index==0 )
      {
	/* Calculating P(S_n1|X_n1) */
	max=0.0;
	for( i=1; i<num_states; i++ )
	  {
	    log_un_prob[i][0][n][0]=log_un_probLogistic( input->entry[0], i, -1 );
	    if( max<log_un_prob[i][0][n][0] )
	      max=log_un_prob[i][0][n][0];
	  }
	
	sum=exp(-max);
	for( i=1; i<num_states; i++ )
	  sum+=exp(log_un_prob[i][0][n][0]-max);	    
	
	p_tr[0][0][n][0]=exp(-max)/sum;
	for( i=1; i<num_states; i++ )
	  p_tr[i][0][n][0]=exp(log_un_prob[i][0][n][0]-max)/sum;
	
	/* Updating start_index */
	start_index=1;
      }

    /* Calculating P(S_{nt}|S_{n,t-1},X_{nt}) */
    for( t=start_index; t<input->seq_length; t++ )
      for( j=0; j<num_states; j++ )
	{
	  max=0.0;
	  for( i=1; i<num_states; i++ )
	    {
	      log_un_prob[i][j][n][t]=log_un_probLogistic( input->entry[t], i, j );
	      if( max<log_un_prob[i][j][n][t] )
		max=log_un_prob[i][j][n][t];
	    }

	  sum=exp(-max);
	  for( i=1; i<num_states; i++ )
	    sum+=exp(log_un_prob[i][j][n][t]-max);
	  
	  p_tr[0][j][n][t]=exp(-max)/sum;
	  for( i=1; i<num_states; i++ )
	    p_tr[i][j][n][t]=exp(log_un_prob[i][j][n][t]-max)/sum;
	}
    break;
  default:;
  }
  return;
}

void Distribution::CalculateForwardUpdatesSequence( Sequence* data, long n, long start_index, double**** pb ) {
  /* Index variable(s) */
  long i,j,t;

  /* Temporary variable(s) */
  double sum;

  switch( type ) {
  case DIST_BERNOULLI:
    for( t=start_index; t<data->seq_length; t++ )
      {
	sum=0.0;
	for( i=0; i<num_states; i++ )
	  {
	    uni_prob[0][i][n][t]=state_prob[i]*pb[0][i][n][t];
	    sum+=uni_prob[0][i][n][t];
	  }
	
	/* Calculating the scaling factor */
	upd_scale[n][t]=1.0/sum;
      }
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    if( start_index==0 )
      {
	/* First alpha */
	sum=0.0;
	for( i=0; i<num_states; i++ )
	  { 
	    fwd_update[i][n][0]=state_prob[i]*pb[0][i][n][0];
	    sum+=fwd_update[i][n][0];
	  }
	
	/* Calculating first c */
	upd_scale[n][0]=1.0/sum;
	
	/* Rescaling first alpha */
	for( i=0; i<num_states; i++ )
	  fwd_update[i][n][0]*=upd_scale[n][0];
	
	/* Updating start_index */
	start_index=1;
      }
    
    /* Calculating the rest of alphas */
    for( t=start_index; t<data->seq_length; t++ )
      {
	sum=0.0;
	for( i=0; i<num_states; i++ )
	  { /* Summing over S_t */
	    fwd_update[i][n][t]=0.0;
	    for( j=0; j<num_states; j++ )
	      /* Summing over S_{t-1} */
	      fwd_update[i][n][t]+=(fwd_update[j][n][t-1]*cond_state_prob[j][i]);
	    
	    fwd_update[i][n][t]*=pb[0][i][n][t];
	    sum+=fwd_update[i][n][t];
	  } /* Summing over S_t */
	
	upd_scale[n][t]=1.0/sum;
	
	/* Rescaling alphas */
	for( i=0; i<num_states; i++ )
	  fwd_update[i][n][t]*=upd_scale[n][t];
      }
    break;
  case DIST_LOGISTIC:
    for( t=start_index; t<data->seq_length; t++ )
      {
	sum=0.0;
	for( i=0; i<num_states; i++ )
	  {
	    uni_prob[0][i][n][t]=p_tr[i][0][n][t]*pb[0][i][n][t];
	    sum+=uni_prob[0][i][n][t];
	  }

	upd_scale[n][t]=1.0/sum;
      }
    break;
  case DIST_TRANSLOGISTIC:
    if( start_index==0 )
      {
	/* First alpha */
	sum=0.0;
	for( i=0; i<num_states; i++ )
	  { 
	    fwd_update[i][n][0]=p_tr[i][0][n][0]*pb[0][i][n][0];
	    sum+=fwd_update[i][n][0];
	  }
	
	/* Calculating first c */
	upd_scale[n][0]=1.0/sum;
	
	/* Rescaling first alpha */
	for( i=0; i<num_states; i++ )
	  fwd_update[i][n][0]*=upd_scale[n][0];
	
	/* Updating start_index */
	start_index=1;
      }
    
    /* Calculating the rest of alphas */
    for( t=start_index; t<data->seq_length; t++ )
      {
	sum=0.0;
	for( i=0; i<num_states; i++ )
	  { /* Summing over S_t */
	    fwd_update[i][n][t]=0.0;
	    for( j=0; j<num_states; j++ )
	      /* Summing over S_{t-1} */
	      fwd_update[i][n][t]+=(fwd_update[j][n][t-1]*p_tr[i][j][n][t]);
	    
	    fwd_update[i][n][t]*=pb[0][i][n][t];
	    sum+=fwd_update[i][n][t];
	  } /* Summing over S_t */
	
	upd_scale[n][t]=1.0/sum;
	
	/* Rescaling alphas */
      for( i=0; i<num_states; i++ )
	fwd_update[i][n][t]*=upd_scale[n][t];
      }
    break;
  default:;
  }

  return;
}

void Distribution::CalculateBackwardUpdatesSequence( Sequence* data, long n, double**** pb ) {
  /* Index variable(s) */
  long i,j,t;

  switch( type ) {
  case DIST_BERNOULLI:
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    /* Initializing beta_T */
    for( i=0; i<num_states; i++ )
      bkd_update[i][n][data->seq_length-1]=upd_scale[n][data->seq_length-1];
      
    /* Calculating remaining betas */
    for( t=data->seq_length-2; t>=0; t-- )
      {
	for( i=0; i<num_states; i++ )
	  { /* Summing over S_t */
	    bkd_update[i][n][t]=0.0;
	    for( j=0; j<num_states; j++ )
	      /* Summing over S_{t+1} */
	      bkd_update[i][n][t]+=(bkd_update[j][n][t+1]*cond_state_prob[i][j]*pb[0][j][n][t+1]);
	  } /* Summing over S_t */
	
	/* Rescaling betas */
	for( i=0; i<num_states; i++ )
	  bkd_update[i][n][t]*=upd_scale[n][t];
      }
    break;
  case DIST_LOGISTIC:
    break;
  case DIST_TRANSLOGISTIC:
    /* Initializing beta_T */
    for( i=0; i<num_states; i++ )
      bkd_update[i][n][data->seq_length-1]=upd_scale[n][data->seq_length-1];
      
    /* Calculating remaining betas */
    for( t=data->seq_length-2; t>=0; t-- )
      {
	for( i=0; i<num_states; i++ )
	  { /* Summing over S_t */
	    bkd_update[i][n][t]=0.0;
	    for( j=0; j<num_states; j++ )
	      /* Summing over S_{t+1} */
	      bkd_update[i][n][t]+=(bkd_update[j][n][t+1]*p_tr[j][i][n][t+1]*pb[0][j][n][t+1]);
	  } /* Summing over S_t */
	
	/* Rescaling betas */
	for( i=0; i<num_states; i++ )
	  bkd_update[i][n][t]*=upd_scale[n][t];
      }
    break;
  deafult:;
  }

  return;
}

void HMM::CalculateSummaries( Data* data ) {
  /* Index variable(s) */
  long i,j,n,t;

  transition->CalculateSummaries( data, pb );
  
  /* Calculating aggregate weights */
  for( i=0; i<num_variables; i++ )
    for( j=0; j<num_states[i]; j++ )
      {
	norm_const[i][j]=0.0;
	for( n=0; n<data->num_seqs; n++ )
	  for( t=0; t<data->sequence[n]->seq_length; t++ )
	    norm_const[i][j]+=transition->uni_prob[i][j][n][t];
      }
  
    return;
}

void Distribution::CalculateSummaries( Data* data, double**** pb ) {
  /* Index variable(s) */
  long i,j,n,t;

  switch( type ) {
  case DIST_BERNOULLI:
    /* Univariate probabilities */
    for( i=0; i<num_states; i++ )
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  uni_prob[0][i][n][t]*=upd_scale[n][t];

    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    /* Joint probabilities */
    for( i=0; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	for( n=0; n<data->num_seqs; n++ )
	  for( t=1; t<data->sequence[n]->seq_length; t++ )
	    joint_prob[i][j][n][t]=fwd_update[j][n][t-1]*bkd_update[i][n][t]*pb[0][i][n][t]*cond_state_prob[j][i];

    /* Univariate probabilities */
    for( i=0; i<num_states; i++ )
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  uni_prob[0][i][n][t]=fwd_update[i][n][t]*bkd_update[i][n][t]/upd_scale[n][t];


    break;
  case DIST_LOGISTIC:
    /* Univariate probabilities */
    for( i=0; i<num_states; i++ )
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  uni_prob[0][i][n][t]*=upd_scale[n][t];

    break;
  case DIST_TRANSLOGISTIC:
    /* Joint probabilities */
    for( i=0; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	for( n=0; n<data->num_seqs; n++ )
	  for( t=1; t<data->sequence[n]->seq_length; t++ )
	    joint_prob[i][j][n][t]=fwd_update[j][n][t-1]*bkd_update[i][n][t]*pb[0][i][n][t]*p_tr[i][j][n][t];

    /* Univariate probabilities */
    for( i=0; i<num_states; i++ )
      for( n=0; n<data->num_seqs; n++ )
	for( t=0; t<data->sequence[n]->seq_length; t++ )
	  uni_prob[0][i][n][t]=fwd_update[i][n][t]*bkd_update[i][n][t]/upd_scale[n][t];

    break;
  default:;
  }
	   
  return;
}

void HMM::DeallocateEMStructures( Data* data, Data* input ) {
  /* Deallocating structures needed for forward-backward algorithm */

  /* Index variables */
  long i,j,n;

  DeallocateAlpha( data, input );
  transition->DeallocateEMStructures( data, input );

  return;
}

void Distribution::DeallocateEMStructures( Data* data, Data* input ) {
  /* Index variable(s) */
  long i,j,n;

  switch( type ) {
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
  case DIST_TRANSLOGISTIC:
    for( i=0; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	{
	  for( n=0; n<data->num_seqs; n++ )
	    delete joint_prob[i][j][n];
	  delete joint_prob[i][j];
	}

    for( i=0; i<num_states; i++ )
      { /* For each state */
	for( n=0; n<data->num_seqs; n++ )
	  /* For each sequence */
	  delete bkd_update[i][n];
	delete bkd_update[i];
      } /* For each state */
    break;
  }

  return;
}

void HMM::DeallocateAlpha( Data* data, Data* input_vars ) {
  /* Deallocating arrays needed to calculate scaled alphas in forward-backward algorithm */
  
  /* Index variable(s) */
  long i,n,j;

  /* pb */
  for( i=0; i<num_variables; i++ )
    for( j=0; j<num_states[i]; j++ )
      { /* For each state */
	for( n=0; n<data->num_seqs; n++ )
	  delete pb[i][j][n];
	delete pb[i][j];
      } /* For each state */

  transition->DeallocateForwardPassStructures( data, input_vars );

  return;
}

void Distribution::DeallocateForwardPassStructures( Data* data, Data* input ) {
  /* Index variable(s) */
  long i,j,n;

  switch( type ) {
  case DIST_BERNOULLI:
    /* Deallocating univariate posterior probabilities */
    for( i=0; i<num_states; i++ )
      {
	for( n=0; n<data->num_seqs; n++ )
	  delete uni_prob[0][i][n];
	delete uni_prob[0][i];
      }

    for( n=0; n<data->num_seqs; n++ )
      delete upd_scale[n];
    delete upd_scale;

    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    for( i=0; i<num_states; i++ )
      { /* For each state */
	for( n=0; n<data->num_seqs; n++ )
	  delete fwd_update[i][n];
	delete fwd_update[i];
      } /* For each state */
    
    /* Deallocating univariate posterior probabilities */
    for( i=0; i<num_states; i++ )
      {
	for( n=0; n<data->num_seqs; n++ )
	  delete uni_prob[0][i][n];
	delete uni_prob[0][i];
      }

    for( n=0; n<data->num_seqs; n++ )
      delete upd_scale[n];
    delete upd_scale;

    break;
  case DIST_LOGISTIC:
    /* Deallocating univariate posterior probabilities */
    for( i=0; i<num_states; i++ )
      {
	for( n=0; n<data->num_seqs; n++ )
	  delete uni_prob[0][i][n];
	delete uni_prob[0][i];
      }

    for( n=0; n<data->num_seqs; n++ )
      delete upd_scale[n];
    delete upd_scale;

    /* Normalized transition probability */
    for( i=0; i<num_states; i++ )
      {
	for( n=0; n<input->num_seqs; n++ )
	  delete p_tr[i][0][n];
	delete p_tr[i][0];
      }

    /* Log-unnormalized transition probability */
    for( i=1; i<num_states; i++ )
      {
	for( n=0; n<input->num_seqs; n++ )
	  delete log_un_prob[i][0][n];
	delete log_un_prob[i][0];
      }

    break;
  case DIST_TRANSLOGISTIC:
    for( i=0; i<num_states; i++ )
      { /* For each state */
	for( n=0; n<data->num_seqs; n++ )
	  delete fwd_update[i][n];
	delete fwd_update[i];
      } /* For each state */
    
    /* Deallocating univariate posterior probabilities */
    for( i=0; i<num_states; i++ )
      {
	for( n=0; n<data->num_seqs; n++ )
	  delete uni_prob[0][i][n];
	delete uni_prob[0][i];
      }

    for( n=0; n<data->num_seqs; n++ )
      delete upd_scale[n];
    delete upd_scale;

    /* Normalized transition probability */
    for( i=0; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	{
	  for( n=0; n<input->num_seqs; n++ )
	    delete p_tr[i][j][n];
	  delete p_tr[i][j];
	}

    /* Log-unnormalized transition probability */
    for( i=1; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	{
	  for( n=0; n<input->num_seqs; n++ )
	    delete log_un_prob[i][j][n];
	  delete log_un_prob[i][j];
	}      

    break;
  default:;
  }

  return;
}
