/*
 * Copyright (c) 2002-2005, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* envelop.cpp */

/* Enveloping functions */

#include "struct.h"

extern long em_verbose;

void RunLearnParameters( Data* output, 
			 Data* input, 
			 HMM* theta, 
			 HMM** passed_theta, 
			 Parameters* params, 
			 FILE* output_file ) {
  
  /* Enveloping routines for em algorithm */
  
  HMM* final_theta;  // Final parameters of the model
  HMM* temp_theta;   // Current model
  
  extern long num_failed; // Number of failed runs
  long prev_num_failed;   // Temporary variable
  long noninit_index;     // Index of first non-initialized run
  
  /* Index variable(s) */
  long i;
  
  /* Temporary variables */
  HMM* dummy_theta;  // Temporary variable to swap temp_theta and final_theta pointers
  long sum;

  /* Initializing */
  final_theta=NULL;
  num_failed=0;

  /* !!! Determining whether to run regular EM or EM-MCMC !!! */
  sum=0;
  for( i=0; i<output->num_seqs; i++ )
    sum+=output->sequence[i]->num_missing_discrete_entries;

  /* Checking whether initial conditions have been passed */
  if( passed_theta )
    {
      if( params->num_models<params->num_restarts )
	noninit_index=params->num_models;
      else
	noninit_index=params->num_restarts;
      
      for( i=0; i<noninit_index; i++ )
	{
	  if( em_verbose )
	    printf( "Starting restart #%ld\n", i+1 );

	  prev_num_failed=num_failed;
	  	  
	  temp_theta=passed_theta[i];
	  
	  /* Learning the parameters */
	  temp_theta->EM( output, input, params->em_precision );
	  
	  /* Updating the best model */
	  if( num_failed==prev_num_failed )
	    { /* Run finished ok */
	      if( !final_theta )
		final_theta=temp_theta;		  
	      else if( temp_theta->ll<final_theta->ll )
		final_theta=temp_theta;
	    } /* Run finished ok */
	}
    }
  else 
    noninit_index=0;
  
  /* Allocating a model */
  temp_theta=theta->copy();

  for( i=noninit_index; i<params->num_restarts; i++ )
    {
      if( em_verbose )
	printf( "Starting restart #%ld\n", i+1 );

      prev_num_failed=num_failed;

      /* Initializing the parameters */
      switch( params->initialization_type ) {
      case INIT_RANDOM:
	temp_theta->Initialize( output, input );
	break;
      case INIT_EM:
	temp_theta->InitializeByEM( output, input );
	break;
      default:;
      }
      
      /* Learning the parameters */
      temp_theta->EM( output, input, params->em_precision );

      if( num_failed==prev_num_failed )
	/* Run finished ok */
	if( !final_theta )
	  {
	    final_theta=temp_theta;
	    temp_theta=theta->copy();
	  }
	else if( temp_theta->ll>final_theta->ll )
	  { /* Updating the best model */
	    dummy_theta=final_theta;
	    final_theta=temp_theta;
	    temp_theta=dummy_theta;
	  } /* Updating the best model */
    }

  if( num_failed<params->num_restarts )
    { /* At least one run is successful */
      /* Calculating BIC score */
      final_theta->bic=final_theta->BIC( output->num_points() );
      
      final_theta->PostProcess();

      final_theta->WriteToFile( output_file );
      
      /* Displaying the statistics about the run */
      fprintf( output_file, "%c Failure rate: %ld out of %ld.\n", 
	       COMMENT_SYMBOL, num_failed, params->num_restarts );
    } /* At least one run is successful */
  else
    { /* All runs failed */
      fprintf( output_file, "%c All runs are unsuccessful\n", 
	       COMMENT_SYMBOL );
    } /* All runs failed */

  /* Deallocating the model */
  if( final_theta )
    delete final_theta;

  delete temp_theta;
  
  return;
}

void RunLogLikelihoodData( Data* output_data, Data* input_data, HMM* theta, FILE* output_file ) {

  /* Allocating necessary structures */
  theta->AllocateEMStructures( output_data, input_data );

  /* Calculating the log-likelihood of the data */
  fprintf( output_file, "%0.12le\n", theta->log_likelihood( output_data, input_data ) );

  /* Deallocating arrays */
  theta->DeallocateEMStructures( output_data, input_data );
}

void RunSimulateData( Data* input_data, 
		      HMM* theta, 
		      Parameters* params, 
		      FILE* state_output_file, 
		      FILE* output_file,
		      long num_seqs ) {
  
  long** sim_states; // Simulated states
  Data* output_data; // Simulated data 
  
  /* Index variable(s) */
  long n,t;

  /* State sequences first */
  sim_states=theta->simulate_states( input_data, 
				     num_seqs, 
				     params->length_data_seqs );
  
  if( state_output_file )
    { /* Displaying the states */
      for( n=0; n<num_seqs; n++ )
	for( t=0; t<params->length_data_seqs; t++ )
	  fprintf( state_output_file, "%ld\n", sim_states[n][t] );
    } /* Displaying the states */
  
  /* Having generated the state sequences, can generate the data */
  output_data=new Data( num_seqs );
  
  /* Simulating the outputs */
  output_data->Simulate( theta->emission[0], sim_states, params->length_data_seqs, params->output_datum );
  
  output_data->WriteToFile( output_file );
  
  /* Deallocating states */
  for( n=0; n<num_seqs; n++ )
    delete sim_states[n];
  delete sim_states;
  
  /* Deallocating the data */
  delete output_data;
  
  return;
}

void RunHoleData( Data* output_data, Data* input_data, HMM* theta, FILE* output_file, long hole_type ) {

  /* Allocating necessary structures */
  //theta->AllocateAlpha( output_data, input_data );
  theta->AllocateEMStructures( output_data, input_data );

  switch( hole_type ) {
  case POKING_TYPE_LOGP:
  case POKING_TYPE_PRED:
    /* Calculating the predictive score of the model */
    fprintf( output_file, "%0.12le\n", theta->hole_score( output_data, input_data, hole_type ) );
    break;
  case MISSING_IND_PROB:
    theta->MissingIndividualMarginals( output_data, input_data, output_file );
    break;
  case HIDDEN_STATE_PROB:
    theta->StateProbabilities( output_data, input_data, output_file );
    break;
  }

  /* Deallocating arrays */
  //  theta->DeallocateAlpha( output_data, input_data );
  theta->DeallocateEMStructures( output_data, input_data );
  
}

void RunPredictionData( Data* data, Data* input, HMM* theta, long lookahead, FILE* output ) {
  double ll;
  
  /* Index variable(s) */
  long n,t,tau;

  theta->AllocateEMStructures( data, input );  

  for( n=0; n<data->num_seqs; n++ )
    {
      /* Calculating emission probabilities */
      theta->CalculateProbEmissionsSequence( data->sequence[n], n, 0 );

      /* Calculating P(S_nt|S_n,t-1, X_nt) */
      if( input )
	theta->transition->CalculateTransitionValuesSequence( input->sequence[n], n, 0 );
      else
	theta->transition->CalculateTransitionValuesSequence( NULL, n, 0 );

      /* Calculating scaled alpha values */
      theta->transition->CalculateForwardUpdatesSequence( data->sequence[n], n, 0, theta->pb );

      for( t=0; t<data->sequence[n]->seq_length-lookahead+1; t++ )
	{
	  ll=0.0;
	  for( tau=t; tau<t+lookahead; tau++ )
	    ll-=log(theta->transition->upd_scale[n][t]);
	  fprintf( output, "%0.12le\n", ll );
	}
    }

  theta->DeallocateEMStructures( data, input );

  return;
}

void RunEvaluate( Data* output_data, long option, FILE* output_file ) {
  double* mean;
  double*** cov;
  double** pers;
  long** dist;
  double** MI;
  double** lo;

  long dim;

  /* Index variable(s) */
  long i,j;

  /* !!! !!! */
  dim=output_data->sequence[0]->entry[0]->ddim;

  switch( option ) {
  case ANALYSIS_MEAN:
    /* Calculating mean value for each component */
    dim=output_data->sequence[0]->entry[0]->ddim+output_data->sequence[0]->entry[0]->rdim;
    mean=output_data->mean();
    for( i=0; i<dim; i++ )
      fprintf( output_file, "%0.8le\t", mean[i] );
    fprintf( output_file, "\n" );

    delete mean;
    
    break;
  case ANALYSIS_CORR:
    /* Calculating correlation of the components */

    /* First, calculating the covariance matrices */
    cov=output_data->covariance();

    /* corr(i,j)=cov(i,j)/sqrt(cov(i,i)*cov(j,j)) */
    if( cov[0] )
      {
	dim=output_data->sequence[0]->entry[0]->ddim;
	for( i=0; i<dim; i++ )
	  {
	    for( j=0; j<dim; j++ )
	      fprintf( output_file, "%0.8le\t", cov[0][i][j]/(sqrt(cov[0][i][i]*cov[0][j][j])) );
	    fprintf( output_file, "\n" );
	  }

	for( i=0; i<dim; i++ )
	  delete cov[0][i];
	delete cov[0];
      }
    if( cov[1] )
      {
	dim=output_data->sequence[0]->entry[0]->rdim;
	for( i=0; i<dim; i++ )
	  {
	    for( j=0; j<dim; j++ )
	      fprintf( output_file, "%0.8le\t", cov[1][i][j]/(sqrt(cov[1][i][i]*cov[1][j][j])) );
	    fprintf( output_file, "\n" );
	  }

	for( i=0; i<dim; i++ )
	  delete cov[1][i];
	delete cov[1];
      }

    delete cov;

    break;
  case ANALYSIS_PERS:
    /* Calculating probabilities of persisting */
    pers=output_data->persistence();

    for( i=0; i<dim; i++ )
      {
	for( j=0; j<2 /* !!! !!! */; j++ )
	  fprintf( output_file, "%0.8le\t", pers[i][j] );
	fprintf( output_file, "\n" );
      }

    for( i=0; i<dim; i++ )
      delete pers[i];
    delete pers;

    break;
  case ANALYSIS_DRY:
    /* Calculating the distribution of the dry spell length */
    dist=output_data->spell( 0 );

    for( i=0; i<dim; i++ )
      {
	for( j=0; j<output_data->sequence[0]->seq_length /* !!! !!! */; j++ )
	  fprintf( output_file, "%ld\t", dist[i][j] );
	fprintf( output_file, "\n" );
      }

    for( i=0; i<dim; i++ )
      delete dist[i];
    delete dist;

    break;
  case ANALYSIS_WET:
    /* Calculating the distribution of the wet spell length */
    dist=output_data->spell( 1 );

    for( i=0; i<dim; i++ )
      {
	for( j=0; j<output_data->sequence[0]->seq_length /* !!! !!! */; j++ )
	  fprintf( output_file, "%ld\t", dist[i][j] );
	fprintf( output_file, "\n" );
      }

    for( i=0; i<dim; i++ )
      delete dist[i];
    delete dist;

    break;
  case ANALYSIS_MI:
    /* Calculating mutual information of the components */

    /* Calculating mutual information */
    MI=output_data->mutual_information();

    for( i=0; i<dim; i++ )
      {
	for( j=0; j<dim; j++ )
	  fprintf( output_file, "%0.8le\t", MI[i][j] );
	fprintf( output_file, "\n" );
      }

    for( i=0; i<dim; i++ )
      delete MI[i];
    delete MI;

    break;
  case ANALYSIS_LO:
    /* Calculating log-odds ratio for the pairs of components */
    lo=output_data->log_odds_ratio();

    for( i=0; i<dim; i++ )
      {
	for( j=0; j<dim; j++ )
	  fprintf( output_file, "%0.8le\t", lo[i][j] );
	fprintf( output_file, "\n" );
      }

    for( i=0; i<dim; i++ )
      delete lo[i];
    delete lo;

    break;
  default:;
  }
  
  fflush( output_file );

  return;
}
