/*
 * Copyright (c) 2002-2005, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* eval.cpp */

/* Contains routines used in evaluation */

#include "struct.h"

double* Data::mean() {
  /* Calculates percentage of rain occurences for each station */

  long ddim, rdim;
  long* num_present;  /* Number of observations for each component */
  double* total;      /* Running sum per component */

  /* Index variable(s) */
  long i,n,t;
  
  /* Initializing hits and total arrays */
  /* !!! Assumes same dimensionality for all entries !!! */
  ddim=sequence[0]->entry[0]->ddim;
  rdim=sequence[0]->entry[0]->rdim;

  num_present=new long[ddim+rdim];
  total=new double[ddim+rdim];
  for( i=0; i<ddim+rdim; i++ )
    {
      num_present[i]=0;
      total[i]=0.0;
    }

  for( n=0; n<num_seqs; n++ )
    for( t=0; t<sequence[n]->seq_length; t++ )
      {      
	for( i=0; i<ddim; i++ )
	  if( !is_missing( sequence[n]->entry[t]->ddata[i] ) )
	    {
	      num_present[i]++;
	      if( sequence[n]->entry[t]->ddata[i]==1 )
		total[i]+=1.0;
	    }
	for( i=0; i<rdim; i++ )
	  if( !is_missing( sequence[n]->entry[t]->rdata[i] ) )
	    {
	      num_present[ddim+i]+1;
	      total[ddim+i]+=sequence[n]->entry[t]->rdata[i];
	    }
      }
	
  /* Computing the array of outputs */
  for( i=0; i<ddim+rdim; i++ )
    total[i]/=(double)num_present[i];

  /* Deallocating hits arrays */
  delete num_present;

  return( total );
}

double*** Data::covariance() {
  /* Calculating covariance */
  double* mu;
  double*** cov;
  long** num_present_d=NULL;
  long** num_present_r=NULL;
  long ddim, rdim;

  /* Index variable(s) */
  long i,j,n,t;

  /* First, need to know the mean */
  mu=mean();

  /* Initializing covariance matrix */
  /* !!! Assumes same dimensionality for all entries !!! */
  ddim=sequence[0]->entry[0]->ddim;
  rdim=sequence[0]->entry[0]->rdim;

  cov=new double**[2];
  if( ddim>0 )
    {
      num_present_d=new long*[ddim];
      for( i=0; i<ddim; i++ )
	{
	  num_present_d[i]=new long[ddim];
	  for( j=0; j<ddim; j++ )
	    num_present_d[i][j]=0;
	}

      cov[0]=new double*[ddim];
      for( i=0; i<ddim; i++ )
	{	
	  cov[0][i]=new double[ddim];
	  for( j=0; j<ddim; j++ )
	    cov[0][i][j]=0.0;
	}
    }
  else
    cov[0]=NULL;

  if( rdim>0 )
    {
      num_present_r=new long*[rdim];
      for( i=0; i<rdim; i++ )
	{
	  num_present_r[i]=new long[rdim];
	  for( j=0; j<rdim; j++ )
	    num_present_r[i][j]=0;
	}

      cov[1]=new double*[rdim];
      for( i=0; i<rdim; i++ )
	{
	  cov[1][i]=new double[rdim];
	  for( j=0; j<rdim; j++ )
	    cov[1][i][j]=0.0;
	}
    }
  else
    cov[1]=NULL;

  for( n=0; n<num_seqs; n++ )
    for( t=0; t<sequence[n]->seq_length; t++ )
      {
	for( i=0; i<ddim; i++ )
	  for( j=0; j<=i; j++ )
	    if( !is_missing( sequence[n]->entry[t]->ddata[i] ) && 
		!is_missing( sequence[n]->entry[t]->ddata[j] ) )
	      {
		num_present_d[i][j]+=1;
		cov[0][i][j]+=((double)sequence[n]->entry[t]->ddata[i]-mu[i])*
		  ((double)sequence[n]->entry[t]->ddata[j]-mu[j]);
	      }
	for( i=0; i<rdim; i++ )
	  for( j=0; j<=i; j++ )
	    if( !is_missing( sequence[n]->entry[t]->rdata[i] ) &&
		!is_missing( sequence[n]->entry[t]->rdata[j] ) )
	      {
		num_present_r[i][j]+=1;
		cov[1][i][j]+=(sequence[n]->entry[t]->rdata[i]-mu[ddim+i])*
		  (sequence[n]->entry[t]->rdata[j]-mu[ddim+j]);
	      }
      }

  if( ddim>0 )
    {
      for( i=0; i<ddim; i++ )
	for( j=0; j<=i; j++ )
	  {
	    cov[0][i][j]/=(double)num_present_d[i][j];
	    cov[0][j][i]=cov[0][i][j];
	  }
      for( i=0; i<ddim; i++ )
	delete num_present_d[i];
      delete num_present_d;
    }

  if( rdim>0 )
    {
      for( i=0; i<rdim; i++ )
	for( j=0; j<=i; j++ )
	  {
	    cov[1][i][j]/=(double)num_present_r[i][j];
	    cov[1][j][i]=cov[1][j][i];
	  }
      for( i=0; i<rdim; i++ )
	delete num_present_r[i];
      delete num_present_r;
    }

  delete mu;

  return( cov );
}

double** Data::persistence() {
  /* Calculating the persistence of values for each component */
  
  long dim;
  long** num_present;
  double** pers;

  /* Index variable(s) */
  long i,j,n,t;

  /* Allocating and initializing arrays */
  dim=sequence[0]->entry[0]->ddim;
  num_present=new long*[dim];
  for( i=0; i<dim; i++ )
    {
      num_present[i]=new long[2]; /* !!! !!! */
      for( j=0; j<2; j++ )
	num_present[i][j]=0;
    }
  pers=new double*[dim];
  for( i=0; i<dim; i++ )
    {
      pers[i]=new double[2]; /* !!! !!! */
      for( j=0; j<2; j++ )
	pers[i][j]=0.0;
    }

  /* Calculating counts */
  for( n=0; n<num_seqs; n++ )
    for( t=1; t<sequence[n]->seq_length; t++ )
      for( i=0; i<dim; i++ )
	if( !is_missing( sequence[n]->entry[t-1]->ddata[i] ) && !is_missing( sequence[n]->entry[t]->ddata[i] ) )
	  {
	    if( sequence[n]->entry[t-1]->ddata[i]==0 )
	      {
		num_present[i][0]++;
		if( sequence[n]->entry[t]->ddata[i]==0 )
		  pers[i][0]+=1.0;
	      }
	    else if( sequence[n]->entry[t-1]->ddata[i]==1 )
	      {
		num_present[i][1]++;
		if( sequence[n]->entry[t]->ddata[i]==1 )
		  pers[i][1]+=1.0;
	      }
	  }

  /* Calculating persistence probabilities */
  for( i=0; i<dim; i++ )
    for( j=0; j<2 /* !!! !!! */; j++ )
      pers[i][j]/=(double)num_present[i][j];

  /* Deallocating counts */
  for( i=0; i<dim; i++ )
    delete num_present[i];
  delete num_present;

  return( pers );
}

long** Data::spell( long outcome ) {

  long dim;
  long** length_dist;
  long spell_length;

  /* Index variable(s) */
  long i,j,k;

  /* Allocating and initializing the output array */
  dim=sequence[0]->entry[0]->ddim;
  
  /* First dimension is the components, second is the length of the spell */
  /* !!! Second component max size is the length of the first sequence !!! */
  length_dist=new long*[dim];
  for( i=0; i<dim; i++ )
    length_dist[i]=new long[sequence[0]->seq_length];

  for( i=0; i<dim; i++ )
    for( j=0; j<sequence[0]->seq_length; j++ )
      length_dist[i][j]=0;

  for( i=0; i<dim; i++ )
    { /* For each component (station) */
      spell_length=0;

      for( j=0; j<num_seqs; j++ )
	{ /* For each sequence */
	  for( k=0; k<sequence[j]->seq_length; k++ )
	    /* For each entry in the sequence */
	    if( sequence[j]->entry[k]->ddata[i]==outcome )
	      /* Spell continues */
	      spell_length++;
	    else if( spell_length>0 )
	      { /* Spell ended */

		/* Updating the counts */
		length_dist[i][spell_length-1]++;

		/* Resetting the length of the current spell */
		spell_length=0;
	      } /* Spell ended */
	  
	  /* Checking whether the sequence has ended in a spell */
	  if( spell_length>0 )
	    /* Updating the counts */
	    length_dist[i][spell_length-1]++;

	  /* Resetting the length */
	  spell_length=0;

	} /* For each sequence */

    } /* For each component (station) */
  
  return( length_dist );
}

double** Data::mutual_information() {
  /* Calculating mutual information for each pair of dimensions */
  double**** counts;
  double** MI;
  long dim;

  /* Index variable(s) */
  long i,j,k,l;

  /* Temporary variable(s) */
  double sum;

  dim=sequence[0]->entry[0]->ddim;

  /* Allocating the arrays of counts */
  counts=new double***[dim];
  for( i=0; i<dim; i++ )
    {
      counts[i]=new double**[i+1];
      for( j=0; j<i+1; j++ )
	{
	  counts[i][j]=new double*[2]; /* !!! !!! */
	  for( k=0; k<2 /* !!! */; k++ )
	    counts[i][j][k]=new double[2]; /* !!! !!! */
	}	
    }

  /* Allocating the MI matrix */
  MI=new double*[dim];
  for( i=0; i<dim; i++ )
    MI[i]=new double[dim];
  
  /* Initializing the counts */
  for( i=0; i<dim; i++ )
    for( j=0; j<i+1; j++ )
      for( k=0; k<2 /* !!! !!! */; k++ )
	for( l=0; l<2 /* !!! !!! */; l++ )
	  counts[i][j][k][l]=0.0;

  /* Calculating the counts */
  for( k=0; k<num_seqs; k++ )
    for( l=0; l<sequence[k]->seq_length; l++ )
      for( i=0; i<dim; i++ )
	for( j=0; j<i+1; j++ )
	  if( !is_missing( sequence[k]->entry[l]->ddata[i] ) &&
	      !is_missing( sequence[k]->entry[l]->ddata[j] ) )
	    counts[i][j][sequence[k]->entry[l]->ddata[i]][sequence[k]->entry[l]->ddata[j]]+=1.0;

  /* Calculating mutual information */
  for( i=0; i<dim; i++ )
    for( j=0; j<i+1; j++ )
      {
	MI[i][j]=0.0;

	/* !!! Assuming binary data !!! */
	sum=0.0;
	for( k=0; k<2; k++ )
	  for( l=0; l<2; l++ )
	    sum+=counts[i][j][k][l];
	
	if( sum>0.0 )
	  {
	    /* Making probabilities out of the counts */
	    for( k=0; k<2; k++ )
	      for( l=0; l<2; l++ )
		counts[i][j][k][l]/=sum;
	    
	    /* Calculating MI */
	    MI[i][j]+=xlogx(counts[i][j][0][0]);
	    MI[i][j]+=xlogx(counts[i][j][0][1]);
	    MI[i][j]+=xlogx(counts[i][j][1][0]);
	    MI[i][j]+=xlogx(counts[i][j][1][1]);
	    MI[i][j]-=xlogx(counts[i][j][0][0]+counts[i][j][0][1]);
	    MI[i][j]-=xlogx(counts[i][j][1][0]+counts[i][j][1][1]);
	    MI[i][j]-=xlogx(counts[i][j][0][0]+counts[i][j][1][0]);
	    MI[i][j]-=xlogx(counts[i][j][0][1]+counts[i][j][1][1]);
	  }
	MI[j][i]=MI[i][j];
      }

  /* Deallocating the array of counts */
  for( i=0; i<dim; i++ )
    {
      for( j=0; j<i+1; j++ )
	{
	  for( k=0; k<2 /* !!! */; k++ )
	    delete counts[i][j][k];
	  delete counts[i][j];
	}	
      delete counts[i];
    }
  delete counts;

  return( MI );
}

double** Data::log_odds_ratio() {
  /* Calculating log-odds ratio for each pair of dimensions */
  double**** counts;
  double** lor;
  long dim;

  /* Index variable(s) */
  long i,j,k,l;

  dim=sequence[0]->entry[0]->ddim;

  /* Allocating the arrays of counts */
  counts=new double***[dim];
  for( i=0; i<dim; i++ )
    {
      counts[i]=new double**[i+1];
      for( j=0; j<i+1; j++ )
	{
	  counts[i][j]=new double*[2]; /* !!! !!! */
	  for( k=0; k<2 /* !!! */; k++ )
	    counts[i][j][k]=new double[2]; /* !!! !!! */
	}	
    }

  /* Allocating the MI matrix */
  lor=new double*[dim];
  for( i=0; i<dim; i++ )
    lor[i]=new double[dim];
  
  /* Initializing the counts */
  for( i=0; i<dim; i++ )
    for( j=0; j<i+1; j++ )
      for( k=0; k<2 /* !!! !!! */; k++ )
	for( l=0; l<2 /* !!! !!! */; l++ )
	  counts[i][j][k][l]=0.0;

  /* Calculating the counts */
  for( k=0; k<num_seqs; k++ )
    for( l=0; l<sequence[k]->seq_length; l++ )
      for( i=0; i<dim; i++ )
	for( j=0; j<i+1; j++ )
	  if( !is_missing( sequence[k]->entry[l]->ddata[i] ) && 
	      !is_missing( sequence[k]->entry[l]->ddata[j] ) )
	    counts[i][j][sequence[k]->entry[l]->ddata[i]][sequence[k]->entry[l]->ddata[j]]+=1.0;

  /* Calculating mutual information */
  for( i=0; i<dim; i++ )
    for( j=0; j<i; j++ )
      {
	/* !!! Works only for binary data !!! */
	/* Log-odds ratio = log(11)+log(00)-log(01)-log(10) */
	lor[i][j]=log(counts[i][j][1][1])+log(counts[i][j][0][0])
	  -log(counts[i][j][0][1])-log(counts[i][j][1][0]);

#ifdef LINUX
	/* !!! Temporarily removing this section for non-linux compilations !!! */
	switch(  isinf(lor[i][j]) ) {
	case 1:
	  lor[i][j]=1e+100; /* Large positive number */
	  break;
	case -1:
	  lor[i][j]=-1e+100; /* Large negative number */
	  break;
	default:;
	}
#endif
	lor[j][i]=lor[i][j];
      }

  for( i=0; i<dim; i++ )
    /* Diagonal entries cannot be made sense of */
    lor[i][i]=0.0;

  /* Deallocating the array of counts */
  for( i=0; i<dim; i++ )
    {
      for( j=0; j<i+1; j++ )
	{
	  for( k=0; k<2 /* !!! */; k++ )
	    delete counts[i][j][k];
	  delete counts[i][j];
	}	
      delete counts[i];
    }
  delete counts;

  return( lor );
}
