/*
 * Copyright (c) 2002-2005, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* matrix.cpp */
#include "matrix.h"

double find_det( double** matrix, long dim ) {
  /* Calculating the determinant of a matrix */
  double** lu;
  double output=1.0;
  
  
  /* Index variable(s) */
  long i;
  
  lu=new double*[dim];
  for( i=0; i<dim; i++ )
    lu[i]=new double[dim];

  ludcmp( matrix, dim, NULL, NULL, lu );

  if( SINGULAR_FAIL )
    /* Matrix is singualr or close to it */
    return( 0.0 );

  for( i=0; i<dim; i++ )
    output*=lu[i][i];
  
  /* Deallocating the decomposition */
  for( i=0; i<dim; i++ )
    delete lu[i];
  delete lu;
  
  return( output );
}

void find_inv( double** a, long dim, double** output ) {
  double** lu;                 // LU decomposition of the matrix
  double* col;
  double* temp_col;
  
  /* Index variable(s) */
  long i,j;

  lu=new double*[dim];
  for( i=0; i<dim; i++ )
    lu[i]=new double[dim];
  
  ludcmp( a, dim, NULL, NULL, lu );

  if( SINGULAR_FAIL )
    /* Problem with decomposition */
    return;

  /* Initializing column vector */
  col=new double[dim];
  for( i=0; i<dim; i++ )
    col[i]=0.0;

  temp_col=new double[dim];
  
  for( j=0; j<dim; j++ )
    { /* Finding inverse of a column */
      col[j]=1.0;
      lubksb( lu, dim, NULL, col, temp_col );
      if( SINGULAR_FAIL )
	/* Problem with back-substitution */
	return;

      /* Copying the values of the inverse column */
      for( i=0; i<dim; i++ )
	output[i][j]=temp_col[i];
      
      col[j]=0.0;
    } /* Finding inverse of a column */
  
  delete temp_col;
  delete col;

  for( i=0; i<dim; i++ )
    delete lu[i];
  delete lu;
  
  return;
}

void ludcmp(double** a, long dim, long* perm, double* flip_perm, double** output ) {
  /* LU decomposition using Crout's method as described on pp. 46-47
     in Numerical Recipes in C, 2nd edition, Cambridge University Press,
     1992
  */
  
  /* Index variable(s) */
  long i,j,k;
  
  SINGULAR_FAIL=0;

  /* copying the values of the original matrix into the new one */
  for( i=0; i<dim; i++ )
    for( j=0; j<dim; j++ )
      output[i][j]=a[i][j];
  
  /* The output for both lower-triangular and upper-triangular matrices
     is stored in the same matrix. */
  
  /* Coefficients are calculated column by column */
  for( j=0; j<dim ; j++ )
    { /* For each column */
      
      /* Handling of the rows is broken down into three cases:
	 upper triangular part, diagonal, and lower triangular part. */
      
      /* Case 1 -- row<column (upper-triangle)) */
      for( i=0; i<j; i++ )
				/* Calculating entries of the upper triangular matrix for row j */
	for( k=0; k<i; k++ )
	  output[i][j]-=(output[i][k]*output[k][j]);
      
      /* Both diagonal and lower triangular entries requires same initial
	 calculations.  Lower triangular entries also require rescaling by the
	 appropriate diagonal element. */
      
      /* Calculating unscaled values */
      for( i=j; i<dim; i++ )
	for( k=0; k<j; k++ )
          output[i][j]-=(output[i][k]*output[k][j]);
      
      /* !!! No pivoting in this version !!! */
      /* Pivoting should be inserted here */
      
      /* Dividing the lower triangular part by the diagonal entry */
      if( fabs(output[j][j])<MATRIX_TINY )
	{
	  // fprintf( stderr, "Need pivoting in LU decomposition!\n" );
	  SINGULAR_FAIL=1;
	  return;
	}
      
      for( i=j+1; i<dim; i++ )
	output[i][j]/=output[j][j];
      
    } /* For each column */
  
  return;	
}

void lubksb( double** a, long dim, long* perm, double* b, double* output ) {
  /* LU back-substitution procedure */
  
  /* Index variable(s) */
  long i,j;

  SINGULAR_FAIL=0;
  
  /* Forward substitution first */
  for( i=0; i<dim; i++ )
    {
      output[i]=b[i];
      for( j=0; j<i; j++ )
	output[i]-=(a[i][j]*output[j]);
    }

  /* Back-substitution */
  for( i=dim-1; i>=0; i-- )
    {
      for( j=i+1; j<dim; j++ )
	output[i]-=(a[i][j]*output[j]);
      
      if( fabs(a[i][i])<MATRIX_TINY )
	{
	  //  fprintf( stderr, "Matrix is too close to being singular.  May need pivoting.\n" );
	  SINGULAR_FAIL=1;
	  return;
	}
      
      output[i]/=a[i][i];	
    }

  return;
}

void cholesky( double** a, long dim, double** output ) {
  /* Find Cholesky decomposition C of a positive definite matrix A.
     C is a upper-triangular matrix (in agreement with Matlab) such that 
     C'*C=A. 
  */
  
  /* Index variable(s) */
  long i,j,k;
  
  for( i=0; i<dim; i++ )
    {
      output[i][i]=0.0;
      for( k=0; k<i; k++ )
	output[i][i]+=output[k][i]*output[k][i];
      output[i][i]=sqrt(a[i][i]-output[i][i]);

      for( j=i+1; j<dim; j++ )
	{
	  output[i][j]=0.0;
	  output[j][i]=0.0;
	  for( k=0; k<i; k++ )
	    output[i][j]+=output[k][j]*output[k][i];
	  output[i][j]=(a[i][j]-output[i][j])/output[i][i];
	}
    }

  return;
}
