/*
 * Copyright (c) 2002-2006, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* model.cpp */
/* Declarations for class HMM */
#include "struct.h"

extern long short_dist_display; // Flag for not writing the params with dimension indices

HMM::HMM( long nv, long* ns ) {
  /* Basic instantiation -- transition and emission added separately */

  /* Index variable(s) */
  long i;
	
  /* Initializing the number of variables */
  num_variables=nv;

  num_states=new long[nv];

  /* Initializing the number of states */  
  for( i=0; i<num_variables; i++ )
    num_states[i]=ns[i];

  emission=new Distribution**[num_variables];
  for( i=0; i<num_variables; i++ )
    emission[i]=new Distribution*[num_states[i]];

  /* Initializing EM arrays */
  
  /* Allocating array of probabilities of data for each value of each latent variable */
  pb=new double***[num_variables];
  for( i=0; i<num_variables; i++ )
    pb[i]=new double**[num_states[i]];

  /* Allocating array of aggregates of weights for each value of each latent variable */
  norm_const=new double*[num_variables];
  for( i=0; i<num_variables; i++ )
    norm_const[i]=new double[num_states[i]];
  
  ll=NEG_INF;
  lp=NEG_INF;
  lp_normalized=NEG_INF;
  bic=NEG_INF;
}

HMM::~HMM() {
  /* Index variable(s) */
  long i,j;

  /* Deallocating emission structure */
  for( i=0; i<num_variables; i++ )
    {
      for( j=0; j<num_states[i]; j++ )
	delete emission[i][j];
      delete emission[i];
    }
  delete emission;

  /* Deallocating transition structure */
  delete transition;

  /* Deallocating arrays of probabilities and weights */
  for( i=0; i<num_variables; i++ )
    delete pb[i];
  delete pb;
  for( i=0; i<num_variables; i++ )
    delete norm_const[i];
  delete norm_const;

  /* Deallocating number of states */
  delete num_states;
}

HMM* HMM::copy() {
  /* Copy constructor */	
  HMM* model;

  /* Index variable(s) */
  long i,j;

  model=new HMM( num_variables, num_states );
  
  /* Copying emissions */
  for( i=0; i<num_variables; i++ )
    for( j=0; j<num_states[i]; j++ )
      model->emission[i][j]=emission[i][j]->copy();

  /* Copying input distributions */
  model->transition=transition->copy();
    
  model->ll=ll;
  model->lp=lp;
  model->lp_normalized=lp_normalized;
  model->bic=bic;

  return( model );
}

void HMM::Initialize( Data* data, Data* input_data ) {
  /* Index variable(s) */
  long i,j;
  double sum;
  
  /* Initializing parameters for the HMM */
  transition->Initialize( input_data );
     
  /* Initializing emission distribution */
  for( i=0; i<num_variables; i++ )
    for( j=0; j<num_states[i]; j++ )
      emission[i][j]->Initialize( data );
  
  return;
}

void HMM::InitializeByEM( Data* data, Data* input_data ) {
  /* Initializes the model by estimating the parameters from the data according to
     a simpler sub-model */

  Distribution* trans_copy;  // Copy of the transition distribution
  extern long num_failed;    // Number of failed runs so far -- should not change in the end
  long num_prev_failed;      // Temporary value of num_failed;

  /* Saving the original distribution structure */
  trans_copy=transition;
  
  /* Simplifying the model by removing the dependence on the input variable for latent state chain */
  transition=trans_copy->Project();

  if( transition )
    { /* Models with complex input layer */
      num_prev_failed=num_failed;
      do
	{
	  /* Initializing the compressed model */
	  Initialize( data, input_data );
	  
	  EM( data, input_data, EM_EPSILON );
	  if( num_failed!=num_prev_failed )
	    /* Run failed -- setting the counter of failed EM runs back to the previous value */
	    num_failed=num_prev_failed;
	  else
	    /* Run is ok -- making the counters differ */
	    num_prev_failed--;
	}
      while( num_failed==num_prev_failed );
      
      /* Expanding the model to the original type */
      transition->Expand( trans_copy );
      
      /* Updating the pointers */
      delete transition;
      transition=trans_copy;

    } /* Models with complex input layer */
  else
    transition=trans_copy;

  return;
}

double HMM::log_likelihood( Data* data, Data* input_vars ) {

  /* Index variable(s) */
  long n;

  ll=0.0;
  for( n=0; n<data->num_seqs; n++ )
    if( input_vars )
      ll+=log_likelihood_sequence( data->sequence[n], input_vars->sequence[n], n, 0 );
    else
      ll+=log_likelihood_sequence( data->sequence[n], NULL, n, 0 );

  //  ll_normalized=ll/data->total_entries();

  return( ll );
}

double HMM::log_posterior( Data* data, Data* input ) {
  /* Computing log-posterior of the parameter set (with unnormalized prior) */

  double log_prior=0.0;

  /* Index variable(s) */
  long i,v;

  /* First, computing the log-likelihood */
  ll=log_likelihood( data, input );

  /* Computing log-prior from emission distributions */
  for( v=0; v<num_variables; v++ )
    for( i=0; i<num_states[v]; i++ )
      /* Computing log-prior contribution for state i of variable v */
      log_prior+=emission[v][i]->log_prior();

  /* Computing log-prior from transition distribution */
  log_prior+=transition->log_prior();

  lp=ll+log_prior;
  lp_normalized=lp/data->total_entries();

  return( lp );
}

double HMM::hole_score( Data* data, Data* input_data, long analysis_type ) {
  double score=0.0;

  /* Index variable(s) */
  long i;
  
  /* Sequences are assumed iid */
  for( i=0; i<data->num_seqs; i++ )
    if( input_data )
      score+=hole_score_sequence( data->sequence[i], input_data->sequence[i], analysis_type, i );
    else
      score+=hole_score_sequence( data->sequence[i], NULL, analysis_type, i );

  return( score );
}

double HMM::hole_score_sequence( Sequence* data, Sequence* input_data, long analysis_type, long index ) {
  double score=0.0;            // Per bit log-likelihood score of the data
  double ll_unaltered;         // Log-likelihood of the unaltered sequence

  /* Index variable(s) */
  long t,m;

  /* Temporary variable(s) */
  long covered_value;
  long entry_dim;
  double current_ll;

  /* Calculating the log-likelihood of the sequence */
  ll_unaltered=log_likelihood_sequence( data, input_data, index, 0 );

#ifdef HOLE_VERBOSE
  printf(" Log-likelihood for sequence %ld = %0.6le\n", index+1, ll_unaltered );
#endif

  /*** Calculating the scores ***/
  /* !!! Determining the dimension of the discrete-valued entries !!! */
  entry_dim=data->entry[0]->ddim;

  /* Starting changing values in the end of the sequence and moving towards the front */
  for( t=0; t<data->seq_length; t++ )
    {
      /* !!! Hack -- should be replaced by something modular !!! */
      for( m=0; m<entry_dim; m++ )
	{
	  /* !!! Assuming two possible values (0 and 1) !!! */

	  /* Making part of the entry different from before */
	  covered_value=data->entry[t]->ddata[m];
	  if( covered_value==0 )
	    data->entry[t]->ddata[m]=1;
	  if( covered_value==1 )
	    data->entry[t]->ddata[m]=0;
	  
	  if( !is_missing( covered_value ) )
	    {
	      /* Calculating the score for the altered sequence */
	      current_ll=log_likelihood_sequence( data, input_data, index, t );
#ifdef HOLE_VERBOSE
	      printf( "%0.2le ", current_ll );
#endif      
	      /* Updating the score */
	      switch( analysis_type ) {
	      case POKING_TYPE_LOGP:
		/* Log-probability of the correct prediction */
		score+=-log(exp(current_ll-ll_unaltered)+1.0);
		break;
	      case POKING_TYPE_PRED:
		/* Number of correct predictions */
		if( ll_unaltered>current_ll )
		  score+=1.0;
		break;
	      default:;
	      }
	    }
	  else
	    {
#ifdef HOLE_VERBOSE
	      printf( "-1 " );
#endif
	    }

	  /* Putting the original value back */
	  data->entry[t]->ddata[m]=covered_value;
	}
#ifdef HOLE_VERBOSE
      printf( "\n" );
#endif
      /* Resetting intermediate values needed to compute the log-likelihood */
      ll_unaltered=log_likelihood_sequence( data, input_data, index, 0 );
    }

  return( score );
}

#ifdef BLAH2
void HMM::MissingIndividualMarginals( Data* data, Data* input_data, FILE* output ) {
  /* Computing and saving to a file marginal probabilities for all missing entries */

  long num_values;           // Number of possible values for the current entry
  double prob[MAXNUMVALUES]; // Marginal probability for the current missing value
  
  double baseline_ll;        // log-likelihood of a sequence in question

  long prev_t=-1;

  /* Index variable(s) */
  long index,n,t,i,j,b,k;

  /* Temporary variable(s) */
  double sum;

  /* !!! Assuming all variables have the same number of values -- binary !!! */
  num_values=2;

  for( n=0; n<data->num_seqs; n++ )
    data->sequence[n]->CatalogueMissingEntries();

  EStep( data, input_data );
  CalculateSummaries( data );

  /* Computing the probabilities */
  for( n=0; n<data->num_seqs; n++ )
    {
#ifdef BLAH
      /* Starting a new sequence */
      if( input_data )
	baseline_ll=log_likelihood_sequence( data->sequence[n], input_data->sequence[n],
					     n, 0 );
      else
	baseline_ll=log_likelihood_sequence( data->sequence[n], NULL,
					     n, 0 );
#endif
      prev_t=-1;
      
      for( index=0; index<data->sequence[n]->num_missing_discrete_entries; index++ )
	{
	  t=data->sequence[n]->missing_discrete[index][0];
	  i=data->sequence[n]->missing_discrete[index][1];
#ifdef BLAH	  
	  if( prev_t!=t && prev_t!=-1)
	    if( input_data )
	      baseline_ll=log_likelihood_sequence( data->sequence[n], input_data->sequence[n],
						   n, prev_t );
	    else
	      baseline_ll=log_likelihood_sequence( data->sequence[n], NULL,
						   n, prev_t );
#endif
	  prev_t=t;
	  
	  sum=0.0;
	  for( b=0; b<num_values-1; b++ )
	    { /* Looping over the value of current missing entry */
	      /* Setting the value */
	      data->sequence[n]->entry[t]->ddata[i]=b;
#ifdef BLAH
	      if( input_data )
		prob[b]=log_likelihood_sequence( data->sequence[n], input_data->sequence[n], n, t );
	      else
		prob[b]=log_likelihood_sequence( data->sequence[n], NULL, n, t );

	      prob[b]=exp(prob[b]-baseline_ll );
#endif
	      prob[b]=0.0;
	      /* !!! !!! */
	      for( j=0; j<num_variables; j++ )
		for( k=0; k<num_states[j]; k++ )
		  if( t==0 )
		    prob[b]+=transition->uni_prob[j][k][n][t]*
		      emission[j][k]->prob( data->sequence[n]->entry[t], NULL )
		      /pb[j][k][n][t];
		  else
		    prob[b]+=transition->uni_prob[j][k][n][t]*
		      emission[j][k]->prob( data->sequence[n]->entry[t], data->sequence[n]->entry[t-1] )
		      /pb[j][k][n][t];

	      sum+=prob[b];
	    } /* Looping over the value of current missing entry */
	  
	  /* Computing the value for the remaining entry */
	  prob[num_values-1]=1.0-sum;
	  
	  /* Placing missing value back */
	  data->sequence[n]->entry[t]->ddata[i]=missing_value( (long)0 );
	  
	  /* Outputting the value */
	  fprintf( output, "%ld %ld %ld ", n, t, i );
	  /* !!! Hack -- only outputting the probability of 1 !!! */
	  //	  for( b=0; b<num_values; b++ )
	  //	    fprintf( output, "\t%0.12le", prob[b] );
#ifdef BLAH
	  fprintf( output, "%0.12le", prob[1] );
	  fprintf( output, "\n" );
#endif
	  /* Outputting the predicted value */
	  if( prob[1]>0.5 )
	    fprintf( output, "1\n" );
	  else
	    fprintf( output, "0\n" );
	}
    }  

  return;
}
#endif

void HMM::MissingIndividualMarginals( Data* data, Data* input_data, FILE* output ) {
  /* Computing and saving to a file marginal probabilities for all missing entries */

  double prob;               // Probability of one for missing data
  
  double baseline_ll;        // log-likelihood of a sequence in question
  double new_ll;

  /* Index variable(s) */
  long index,n,t,i;

  /* !!! Assuming all variables have the same number of values -- binary !!! */
  for( n=0; n<data->num_seqs; n++ )
    { /* For each sequence */
      if( input_data )
	baseline_ll=log_likelihood_sequence( data->sequence[n], input_data->sequence[n], n, 0 );
      else
	baseline_ll=log_likelihood_sequence( data->sequence[n], NULL, n, 0 );

      for( t=data->sequence[n]->seq_length-1; t>=0; t-- )
	{
	  for( i=0; i<data->sequence[n]->entry[t]->ddim; i++ )
	    if( is_missing( data->sequence[n]->entry[t]->ddata[i] ) )
	      { /* Computing probability of missing entry being 1 */
		data->sequence[n]->entry[t]->ddata[i]=1;
		if( input_data )
		  new_ll=log_likelihood_sequence( data->sequence[n], input_data->sequence[n], n, t );
		else
		  new_ll=log_likelihood_sequence( data->sequence[n], NULL, n, t );

		prob=exp(new_ll-baseline_ll);
		fprintf( output, "%0.12le\n", prob );
		
		/* Returning the value */
		data->sequence[n]->entry[t]->ddata[i]=missing_value( (long)0 );
	      } /* Computing probability of missing entry being 1 */

	  /* Recomputing alphas */
	  if( input_data )
	    log_likelihood_sequence( data->sequence[n], input_data->sequence[n], n, t );
	  else
	    log_likelihood_sequence( data->sequence[n], NULL, n, t );
	}
    } /* For each sequence */

  return;
}


void HMM::StateProbabilities( Data* data, Data* input_data, FILE* output ) {
  /* Computing and writing to a file P(S_nt|R_n,X_n) */

  /* Index variable(s) */
  long n,t,i;

  EStep( data, input_data );
  CalculateSummaries( data );

  /* Computing the probabilities */
  /* !!! !!! */
  for( n=0; n<data->num_seqs; n++ )
    for( t=0; t<data->sequence[n]->seq_length; t++ )
      {
	for( i=0; i<num_states[0]; i++ )
	  fprintf( output, "%0.12le\t", transition->uni_prob[0][i][n][t] );
	fprintf( output, "\n" );
      }

  return;
}

void HMM::WriteToFile( FILE* output ) {
  /* Records model parameters into a file as 12-digit doubles */

  /* Index variable(s) */
  long i,j;

  /* File clarification */
  fprintf( output, "%c This file contains the parameters of the HMM\n", COMMENT_SYMBOL );

  /* Transition distribution */
  fprintf( output, "%c Transition probability:\n", COMMENT_SYMBOL );
  transition->WriteToFile2( output );
  fprintf( output, "\n" );

  /* Emission distributions */
  fprintf( output, "%c Parameters for the emission distributions for different possible hidden states:\n",
	   COMMENT_SYMBOL );
  for( i=0; i<num_variables; i++ )
    {
      fprintf( output, "%c Variable #%ld:\n", COMMENT_SYMBOL, i+1 );
      for( j=0; j<num_states[i]; j++ )
	{
	  fprintf( output, "%c State #%ld:\n", COMMENT_SYMBOL, j+1 );
	  if( !short_dist_display )
	    emission[i][j]->WriteToFile( output );
	  else
	    emission[i][j]->WriteToFile2( output );
	}
      fprintf( output, "\n" );
    }
  fprintf( output, "\n" );

  /* Displaying the log-likelihood for the data set */
  fprintf( output, "%c Log-likelihood of the data set: %0.12le\n", COMMENT_SYMBOL, ll );

  /* Displaying the log-posterior */
  fprintf( output, "%c Log-posterior of the data set: %0.12le\n", COMMENT_SYMBOL, lp );

  /* Displaying the per-dim log-likelihood for the data set */
  fprintf( output, "%c Per-dimension log-posterior of the data set: %0.12le\n", 
	   COMMENT_SYMBOL, lp_normalized );

  /* Displaying number of free parameters */
  fprintf( output, "%c Number of free (real-valued) parameters: %ld\n",
	   COMMENT_SYMBOL, num_params() );

  /* Displaying BIC score for the model and set */
  fprintf( output, "%c Bayes Information Criterion score: %0.12le\n", COMMENT_SYMBOL, bic ); 

  fflush( output );

  return;
}

long HMM::num_params() {
  /* Calculating the number of free parameters for the model */
  long num=0;

  /* Index variables */
  long i,j;

  /* Accounting for output layer */
  for( i=0; i<num_variables; i++ )
    for( j=0; j<num_states[i]; j++ )
      num+=emission[i][j]->num_params();

  /* Accounting for transition layer */
  num+=transition->num_params();

  return( num );
}

double HMM::BIC( long num_points ) {
  /* Calculating Bayes Inference Criterion score */
  
  return( log(double(num_points))*double(num_params())-2*ll );
}

void HMM::PostProcess() {
  /* Post training rearrangement of parameters */

  /* Index variable(s) */
  long i,j;
  
  /* So far, may need to rearrange only the emission probabilities */
  for( i=0; i<num_variables; i++ )
    for( j=0; j<num_states[i]; j++ )
      emission[i][j]->PostProcess();

  return;
}
