/*
 * Copyright (c) 2002-2006, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* simulation.cpp */

#include "struct.h"

static long is_first_normal_value=1;

long** HMM::simulate_states( Data* input, long num_seqs, long num_entries ) {
  /* Simulating sequences of hidden states according to the model theta */

  /* Variables */
  long** states;                // Output structure

  double* prob;                 // Probability distribution for S_nt given previous S_n,t-1 and Xs

  /* Index variable(s) */
  long n,t;

  /* Allocating the structure for the state sequences */
  if( input )
    states=new long*[input->num_seqs];
  else
    states=new long*[num_seqs];

  /* Allocating probability distribution for P(S_nt|S_n,t-1,Xs) */
  prob=new double[num_states[0]];

  for( n=0; n<num_seqs; n++ )
    { /* Generating state sequence n */
      
      /* Allocating states[i] */
      if( input )
	num_entries=input->sequence[n]->seq_length;

      states[n]=new long[num_entries];

      for( t=0; t<num_entries; t++ )
	{ /* Generating state S_nt */	  
	
	  /* Resetting probability distribution for the state */
	  switch( transition->type ) {
	  case DIST_CONDBERNOULLI:
	  case DIST_CONDBERNOULLIG:
	    if( t==0 )
	      prob=transition->state_prob;
	    else
	      prob=transition->cond_state_prob[states[n][t-1]];
	    break;
	  case DIST_TRANSLOGISTIC:
	    if( t==0 )
	      transition->probLogistic( input->sequence[n]->entry[t], -1, prob );
	    else
	      transition->probLogistic( input->sequence[n]->entry[t], states[n][t-1], prob );
	    break;
	  case DIST_BERNOULLI:
	    prob=transition->state_prob;
	    break;
	  case DIST_LOGISTIC:
	    transition->probLogistic( input->sequence[n]->entry[t], -1, prob );
	    break;
	  default:;
	  }

	  /* Simulating state S_nt */
	  states[n][t]=generateBernoulli( prob, num_states[0] );
	} /* Generating state S_nt */
      
    } /* Generating state sequence n */

      /* Deallocating structures */
  if( input )
    /* Deallocating distribution P(S_nt|S_n,t-1,Xs) */
    delete prob;

  return( states );
}

long* HMM::simulate_state_sequence_gibbs( Sequence* data, Sequence* input, long index, long do_init ) {
  /* Simulating a sequence of hidden states according to the model and the 
     COMPLETE input and output sequence */

  /* Variables */
  long* state;                 // Output structure
  double* prob;                // Probability distribution for S_t given everyting instanstiated

  /* Index variable(s) */
  long t,i;

  /* Temporary variable(s) */
  double sum;
  
  /* Allocating the structure for the state sequences */
  state=new long[data->seq_length];
  
  /* Allocating the array of probabilities */
  prob=new double[num_states[0]];

  if( do_init ) 
    {
    
      /* Calculating emission probabilities */
      CalculateProbEmissionsSequence( data, index, 0 );
      
      switch( transition->type ) {
      case DIST_LOGISTIC:
	transition->CalculateTransitionValuesSequence( input, index, 0 );
	break;
      case DIST_TRANSLOGISTIC:
	transition->CalculateTransitionValuesSequence( input, index, 0 );
	transition->CalculateForwardUpdatesSequence( data, index, 0, pb );
	break;
      case DIST_CONDBERNOULLI:
      case DIST_CONDBERNOULLIG:
	transition->CalculateForwardUpdatesSequence( data, index, 0, pb );
	break;
      default:;
      }
    }

  for( t=data->seq_length-1; t>=0; t-- )
    { /* Simulating hidden state S_t */
      sum=0.0;
      switch( transition->type ) {
      case DIST_BERNOULLI:
	/* Homogeneous mixture model */
	for( i=0; i<num_states[0]; i++ )
	  {
	    prob[i]=pb[0][i][index][t]*transition->state_prob[i];
	    sum+=prob[i];
	  }
	break;
      case DIST_LOGISTIC:
	/* Nonhomogeneous mixture model */
	for( i=0; i<num_states[0]; i++ )
	  {
	    prob[i]=pb[0][i][index][t]*transition->p_tr[i][0][index][t];
	    sum+=prob[i];
	  }
	break;
      case DIST_CONDBERNOULLI:
      case DIST_CONDBERNOULLIG:
	/* Hidden Markov model */
	if( t==data->seq_length-1 )
	  /* Last hidden variable */
	  for( i=0; i<num_states[0]; i++ )
	    {
	      prob[i]=transition->fwd_update[i][index][t];
	      sum+=prob[i];
	    }
	else
	  /* Not the last hidden variable */
	  for( i=0; i<num_states[0]; i++ )
	    {
	      prob[i]=transition->fwd_update[i][index][t]*transition->cond_state_prob[i][state[t+1]];
	      sum+=prob[i];
	    }
	break;
      case DIST_TRANSLOGISTIC:
	/* Nonhomogeneous hidden Markov model */
	if( t==data->seq_length-1 )
	  /* Last hidden variable */
	  for( i=0; i<num_states[0]; i++ )
	    {
	      prob[i]=transition->fwd_update[i][index][t];
	      sum+=prob[i];
	    }
	else
	  /* Not the last hidden variable */
	  for( i=0; i<num_states[0]; i++ )
	    {
	      prob[i]=transition->fwd_update[i][index][t]*transition->p_tr[state[t+1]][i][index][t+1];
	      sum+=prob[i];
	    }
	break;
      default:;
      }	

      /* Normalizing probabilities */
      for( i=0; i<num_states[0]; i++ )
	prob[i]/=sum;

      /* Simulating state t */
      state[t]=generateBernoulli( prob, num_states[0] );
    } /* Simulating hidden state S_t */

  delete prob;

  return( state );
}

long* HMM::simulate_missing_discrete_sequence_gibbs( Sequence* data, long* state ) {
  long* missing;
  double* prob;

  /* Index variables */
  long n,t,i,b,e,m1,m2,mi;

  /* Temporary variable(s) */
  double sum;
  Distribution* dist;

  if( data->num_missing_discrete_entries>0 )
    /* Allocating the array for new values */
    missing=new long[data->num_missing_discrete_entries];
  else
    missing=NULL;

  /* !!! Assuming Conditional Chow-Liu Forests for emissions !!! */

  for( n=0; n<data->num_missing_discrete_entries; n++ )
    { /* Looping over missing values */
      t=data->missing_discrete[n][0];
      i=data->missing_discrete[n][1];
      
      //      printf( "Missing value at time %ld at component %ld\n", t, i );

      dist=emission[0][state[t]];
      prob=new double[dist->num_states];
      
      /* Since all interactions are pairwise, can some over all edges
	 with R_t^i as a vertex. */
      /* Processing P(R_t|R_t-1,S_t) */
      //      printf( "S(%ld)=%ld\n", t, state[t] );
      if( dist->type==DIST_CONDCHOWLIU )
	{
	  /* Locating the index corresponding to i */
	  for( mi=0; mi<dist->dim; mi++ )
	    if( dist->dim_index[mi]==i )
	      break;

	  //	  printf( "Corresponding index=%ld\n", mi );
	  //	  printf( "Considering auto-regressive transition %ld->%ld\n", t-1, t );

	  for( b=0; b<dist->num_states; b++ )
	    prob[b]=dist->ind_prob[mi][b];
	  
	  for( e=0; e<dist->num_edges; e++ )
	    {
	      m1=dist->edge[e][0];
	      m2=dist->edge[e][1];
	      if( m1>=dist->dim && m1-dist->dim==mi )
		{ /* i is the first node for edge e */
		  if( m2>=dist->dim )
		    //		    {
		    //		  printf( "Edge connects %ld and %ld\n", m1, m2 );

		    /* Both nodes are in the current time slice */
		    for( b=0; b<dist->num_states; b++ )
		      prob[b]*=
			dist->edge_prob[e][b][data->entry[t]->ddata[dist->dim_index[m2-dist->dim]]]/
			dist->ind_prob[m1][b];
		  //		    }
		  else if( t>0 )
		    //		    {
		    //		  printf( "Edge connects %ld and %ld\n", m1, m2 );

		    /* The second node is in the previous time slice */
		    for( b=0; b<dist->num_states; b++ )
		      prob[b]*=
			dist->edge_prob[e][b][data->entry[t-1]->ddata[dist->dim_index[m2]]]/
			dist->ind_prob[m1][b];
		  //	    }			  
		} /* i is the first node for edge e */
	      else if( m2>=dist->dim && m2-dist->dim==mi )
		{ /* i is the second node for edge e */
		  if( m1>=dist->dim )
		    //  {
		    //	  printf( "Edge connects %ld and %ld\n", m1, m2 );
		    /* Both nodes are in the current time slice */
		    for( b=0; b<dist->num_states; b++ )
		      prob[b]*=
			dist->edge_prob[e][data->entry[t]->ddata[dist->dim_index[m1-dist->dim]]][b]/
			dist->ind_prob[m2][b];
		  //	    }
		  else if( t>0 )
		    //	    {
		    //	  printf( "Edge connects %ld and %ld\n", m1, m2 );
		    /* The first node is in the previous time slice */
		    for( b=0; b<dist->num_states; b++ )
		      prob[b]*=
			dist->edge_prob[e][data->entry[t-1]->ddata[dist->dim_index[m1]]][b]/
			dist->ind_prob[m2][b];
		  //	    }
		} /* i is the second node for edge e */	    
	    }
	}
	 
      /* Processing P(R_{t+1}|R_t,S_t) */
      if( t<data->seq_length-1 )
	{
	  //	  printf( "Considering auto-regressive transition %ld->%ld\n", t, t+1 );
	  //	  printf( "S(%ld)=%ld\n", t+1, state[t+1] );

	  dist=emission[0][state[t+1]];
	  if( dist->type==DIST_CONDCHOWLIU )
	    for( e=0; e<dist->num_edges; e++ )
	      {
		m1=dist->edge[e][0];
		m2=dist->edge[e][1];
		if( m1<dist->dim && m1==mi )
		  //  {
		  //   printf( "Edge connects %ld and %ld\n", m1, m2 );
		  /* i is the first node for edge e */
		  /* m2 must be from the next time slice */
		  for( b=0; b<dist->num_states; b++ )
		    prob[b]*=
		      dist->edge_prob[e][b][data->entry[t+1]->ddata[dist->dim_index[m2-dist->dim]]]/
		      dist->ind_prob[m1][b];
		//	  }
		else if( m2<dist->dim && m2==mi )
		  //	  {
		  //	    printf( "Edge connects %ld and %ld\n", m1, m2 );
		  /* i is the second node for edge e */
		  /* m1 must be from the next time slice */
		  
		  for( b=0; b<dist->num_states; b++ )
		    prob[b]*=
		      dist->edge_prob[e][data->entry[t+1]->ddata[dist->dim_index[m1-dist->dim]]][b]/
		      dist->ind_prob[m2][b];
		//	  }
	      } /* i is the second node for edge e */	    
	}
 
      /* Normalizing the probability */
      sum=0.0;
      for( b=0; b<dist->num_states; b++ )
	sum+=prob[b];
      for( b=0; b<dist->num_states; b++ )
	prob[b]/=sum;

      missing[n]=generateBernoulli( prob, dist->num_states );

      delete prob;      
    } /* Looping over missing values */
  
  return( missing );
}

long*** HMM::simulate_latent_sequence_po( Data* data, Data* input, long num_sims ) {
  long*** S;              // Sequence of latent states
  double* prob;           // Current probability vector

  /* Index variable(s) */
  long s,i,j,n,t;

  /* Temporary variable(s) */
  double sum;

  prob=new double[num_states[0]];

  /* Allocating array of latent state values */
  S=new long**[num_sims];
  for( s=0; s<num_sims; s++ )
    {
      S[s]=new long*[data->num_seqs];
      for( n=0; n<data->num_seqs; n++ )
	S[s][n]=new long[data->sequence[n]->seq_length];
    }

  /* Allocating necessary structures */
  AllocateAlpha( data, input );

  for( n=0; n<data->num_seqs; n++ )
    {
      /* Computing forward updates */
      CalculateProbEmissionsSequence( data->sequence[n], n, 0 );
      transition->CalculateTransitionValuesSequence( input->sequence[n], n, 0 );
      transition->CalculateForwardUpdatesSequence( data->sequence[n], n, 0, pb );
    }

  for( s=0; s<num_sims; s++ )
    {
      for( t=data->sequence[n]->seq_length-1; t>=0; t-- )
	{ /* Simulating hidden state S_t */

	  sum=0.0;
	  switch( transition->type ) {
	  case DIST_BERNOULLI:
	    /* Mixture */
	    for( i=0; i<num_states[0]; i++ )
	      {
		prob[i]=pb[0][i][n][t]*transition->state_prob[i];
		sum+=prob[i];
	      }
	    break;
	  case DIST_LOGISTIC:
	    /* Non-homogeneous mixture */
	    for( i=0; i<num_states[0]; i++ )
	      {
		prob[i]=pb[0][i][n][t]*transition->p_tr[i][0][n][t];
		sum+=prob[i];
	      }
	    break;
	  case DIST_CONDBERNOULLI:
	  case DIST_CONDBERNOULLIG:	    
	    /* HMM */
	    if( t==data->sequence[n]->seq_length-1 )
	      /* Last hidden variable */
	      for( i=0; i<num_states[0]; i++ )
		{
		  prob[i]=transition->fwd_update[i][n][t];
		  sum+=prob[i];
		}
	    else
	      /* Not the last hidden variable */
	      for( i=0; i<num_states[0]; i++ )
		{
		  prob[i]=transition->fwd_update[i][n][t]*transition->cond_state_prob[i][S[s][n][t+1]];
		  sum+=prob[i];
		}
	    break;
	  case DIST_TRANSLOGISTIC:
	    /* NHMM */
	    if( t==data->sequence[n]->seq_length-1 )
	      /* Last hidden variable */
	      for( i=0; i<num_states[0]; i++ )
		{
		  prob[i]=transition->fwd_update[i][n][t];
		  sum+=prob[i];
		}
	    else
	      /* Not the last hidden variable */
	      for( i=0; i<num_states[0]; i++ )
		{
		  prob[i]=transition->fwd_update[i][n][t]*transition->p_tr[S[s][n][t+1]][i][n][t+1];
		  sum+=prob[i];
		}
	    break;
	  default:;
	  }
	  
	  /* Normalizing probabilities */
	  for( i=0; i<num_states[0]; i++ )
	    prob[i]/=sum;

	  /* Simulating latent state value */
	  S[s][n][t]=generateBernoulli( prob, num_states[0] );
	} /* Simulating hidden state S_t */
    }

  /* Deallocating necessary structures */
  DeallocateAlpha( data, input );

  delete prob;

  return( S );
}
						    
void HMM::InitializeMissingEntriesSequence( Sequence* data ) {
  /* !!! Assinging initial values uniformly at random */
  /* !!! Assuming binary values */
						
  /* Index variables */
  long m,t,i;

  for( m=0; m<data->num_missing_discrete_entries; m++ )
    {
      t=data->missing_discrete[m][0];
      i=data->missing_discrete[m][1];

      if( drand48()<0.5 )
	data->entry[t]->ddata[i]=0;
      else
	data->entry[t]->ddata[i]=1;
    }

  return;
}

long generateBernoulli( double* dist, long dim ) {
  /* Generates a state (1 through dim) accrding to a Bernoulli distribution with dim states and distribution
     dist for the states.  Assumes that the total probability mass in dist adds up to 1. */

  long state=0;     // Generated state
  double sum=0.0;   // Probability mass counter
  double dice_roll; // Random number

  dice_roll=drand48();
  while( sum<dice_roll )
    {
      sum+=dist[state];
      state++;
    }

  if( state==0 )
    return( 0 );
  else
    return( state-1 );
}

double generateUnitNormal( void ) {
  /* Generating a univariate normally distributed value with zero mean and unit variance */ 
  /* Box-Muller */

  /* Following "Numerical Recipes in C", 2nd edition */

  /* Two generated random variables */
  double first_value;              // First
  static double second_value;      // Second value (to be carried over for the next run)

  double v1, v2;                   // Uniformly distributed in (-1,1)
  double d;                        // Length of (v1,v2)
  double fac;                      // Multiplication factor

  if( is_first_normal_value )
    { /* Simulating two normal values */
      do 
	{
	  v1=2.0*drand48()-1.0;
	  v2=2.0*drand48()-1.0;
	  d=v1*v1+v2*v2;
	} 
      while( d>=1.0 || d==0.0 );

      fac=sqrt(-2.0*log(d)/d);
      first_value=v1*fac;
      second_value=v2*fac;
      is_first_normal_value=0;

      return( first_value );
    } /* Simulating two normal values */
  else
    { /* Using left-over value */
      is_first_normal_value=1;
      return( second_value );
    } /* Using left-over value */
}

double generateExponential( double lambda ) {
  /* Generating a sample from a univariate exponential distribution f(x)=l*exp(-l*x) */
  return( -log(1.0-drand48())/lambda );
}

double generateGamma( double a, double b ) {
  /* Generating a sample from a Gamma distribution f(x)=x^(a-1)exp(-b*x)*b^a/G(a) */

  double x=0.0;
  double u,y,v,c,p;

  long done;

  /* First sampling from f(x)=x^(a-1)exp(-x)/G(a) */

  /* The method implmented below is based on rejection method as sugested by 
     Ahrens and summarized in "The Art of Computer Programming" by D. Knuth */

  done=0;

  /* Cases of a>1 and 0<a<=1 are treated separately */
  if( a>1.0 )
    while( !done )
      {
	u=drand48();
	y=tan(M_PI*u);
	x=sqrt(2.0*a-1.0)*y+a-1.0;
	if( x>0 )
	  {
	    /* Rejection in action */
	    v=drand48();
	    if( v<=(1.0+y*y)*exp((a-1.0)*log(x/(a-1.0))-sqrt(2.0*a-1.0)*y) )
	      /* Value under the curve -- accepting */
	      done=1;
	  }
      }
  else if( a>0.0 )
    {
      c=(exp(1.0)+a)/exp(1.0);
      while( !done )
	{
	  u=drand48();
	  p=c*u;
	  if( p>1.0 )
	    {
	      x=-log((c-p)/a);
	      v=drand48();
	      if( v<=pow(x,a-1.0) )
		done=1;            
	    }
	  else
	    {
	      x=pow(p,1.0/a);
	      u=drand48();
	      if( u<=exp(-x) )
		done=1;
	    }
	}
    }

  return( x/b );
}

double generateLognormal( double M, double S2 ) {
  double x;

  x=exp(sqrt(S2)*generateUnitNormal()+M);
  
  return( x );
}

void SimulateNormal( double* mu, double** Sigma, double* x, long dim ) {
  /* Simulating x according to the multivariate normal distribution with mean mu
     and upper-triangle Cholesky decomposition Sigma of the covariance matrix */

  double* unitnormal;

  /* Index variable(s) */
  long i,j;

  if( dim!=1 )
    { /* Multivariate normal */
      unitnormal=new double[dim];

      for( i=0; i<dim; i++ )
	{
	  unitnormal[i]=generateUnitNormal();
	  x[i]=mu[i];
	  for( j=0; j<=i; j++ )
	    x[i]+=unitnormal[j]*Sigma[j][i];
	}

      /* Applying the covariance structure */

      delete unitnormal;
    } /* Multivariate normal */
  else
    x[0]=mu[0]+Sigma[0][0]*generateUnitNormal();

  return;
}

void Data::Simulate( Distribution** dist, long** states, long num_entries, DataPoint* datum ) {
  /* Simulating sequences of data according to distribution dist and already generated hidden states */

  /* Index variable(s) */
  long n;
  
  for( n=0; n<num_seqs; n++ )
    {
      /* Initializing sequence n */
      if( sequence[n] )
	delete sequence[n];
      sequence[n]=new Sequence( num_entries );
            
#ifdef SIM_VERBOSE
      printf( "Sequence %ld", n );
      fflush( stdout );
#endif 
      /* Simulating sequence n */
      sequence[n]->Simulate( dist, states[n], datum );
#ifdef SIM_VERBOSE
      printf( "Done\n" );
#endif
    }

  return;
}

void Sequence::Simulate( Distribution** dist, long* states, DataPoint* datum ) {
  /* Simulating a sequence of data according to distribution dist and already generated hidden states */

  /* Index variable(s) */
  long t;
  
  for( t=0; t<seq_length; t++ )
    {
      /* Initializing observation t */
      if( entry[t] )
	delete entry[t];
      entry[t]=new DataPoint( datum->ddim, datum->rdim );

      /* Simulating observation t */
      if( t==0 )
	dist[states[t]]->Simulate( entry[t], NULL );
      else
	dist[states[t]]->Simulate( entry[t], entry[t-1] );
#ifdef SIM_VERBOSE
      printf( "." );
#endif
    }  

  return;
}

void Distribution::Simulate( DataPoint* sim, DataPoint* prev_datum ) {
  /* Simulating an observation according to distribution */

  long* generated;
  double* prob_vector;
  long num_gibbs_sims;

  /* Index variable(s) */
  long i,j,e,sim_index;

  /* Temporary variables */
  long temp_value;
  long* ddata;
  double* rdata;
  double sum;
  double temp;

  switch( type ) {
  case DIST_FACTOR:
    for( i=0; i<dim; i++ )
      subdist[i]->Simulate( sim, prev_datum );
    break;
  case DIST_MIXTURE:
    /* Picking a mixture component */
    i=generateBernoulli( state_prob, num_states );

    /* Simulating a data point from the component */
    subdist[i]->Simulate( sim, prev_datum );

    break;
  case DIST_CLMIXTURE:
    /* Tree-structured mixture */

    ddata=new long[dim];
    
    generated=new long[dim];
    for( i=0; i<dim; i++ )
      generated[i]=0;
    
    prob_vector=new double[num_states];

    for( e=0; e<num_edges; e++ )
      {
	/* Two possibilities:
	   both nodes need to be simulated, or just one */
	if( !generated[edge[e][0]] && !generated[edge[e][1]] )
	  { /* Both nodes need to be simulated */
	    /* Simulating the first node of the edge */
	    ddata[edge[e][0]]=generateBernoulli( ind_prob[edge[e][0]], num_states );
	    generated[edge[e][0]]=1;

	    /* Simulating the second node of the edge */
	    /* Creating the probability distribution first */
	    for( i=0; i<num_states; i++ )
	      prob_vector[i]=edge_prob[e][ddata[edge[e][0]]][i]/ind_prob[edge[e][0]][ddata[edge[e][0]]];

	    ddata[edge[e][1]]=generateBernoulli( prob_vector, num_states );
	    generated[edge[e][1]]=1;
	  } /* Both nodes need to be simulated */
	else
	  { /* Only one node needs to be simulated */
	    if( !generated[edge[e][0]] )
	      { /* Simulating the first node of the edge given the second */
		/* Creating the probability distribution first */
		for( i=0; i<num_states; i++ )
		  prob_vector[i]=edge_prob[e][i][ddata[edge[e][1]]]/ind_prob[edge[e][1]][ddata[edge[e][1]]];
		
		ddata[edge[e][0]]=generateBernoulli( prob_vector, num_states );
		generated[edge[e][0]]=1;
	      } /* Simulating the first node of the edge given the second */
	    
	    if( !generated[edge[e][1]] )
	      { /* Simulating the second node of the edge given the first */
		/* Creating the probability distribution first */
		for( i=0; i<num_states; i++ )
		  prob_vector[i]=edge_prob[e][ddata[edge[e][0]]][i]/ind_prob[edge[e][0]][ddata[edge[e][0]]];

		ddata[edge[e][1]]=generateBernoulli( prob_vector, num_states );
		generated[edge[e][1]]=1;
	      } /* Simulating the second node of the edge given the first */
	  } /* Only one node needs to be simulated */    
      }

    /* Making sure all nodes have been simulated */
    for( i=0; i<dim; i++ )
      if( !generated[i] )
	/* Simulating the node form individual probabilities */
	ddata[i]=generateBernoulli( ind_prob[i], num_states );
      
    delete prob_vector;

    delete generated;

    /* Simulating values for each of the mixtures */
    for( i=0; i<dim; i++ )
      subdist2[i][ddata[i]]->Simulate( sim, prev_datum );

    delete ddata;
    
    break;
  case DIST_BERNOULLI:
    if( dim==1 )
      sim->ddata[dim_index[0]]=generateBernoulli( state_prob, num_states );
    else
      {
	fprintf( stderr, "Multi-dimensional Bernoulli distribution simulation is not yet implemented!\n" );
	exit( -1 );
      }
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    if( dim==1 )
      if( prev_datum )
	/* Not the first sequence entry */
	sim->ddata[dim_index[0]]=generateBernoulli( cond_state_prob[prev_datum->ddata[dim_index[0]]], num_states );
      else
	/* First sequence entry */
	sim->ddata[dim_index[0]]=generateBernoulli( state_prob, num_states );
    else
      {
	fprintf( stderr, "Multi-dimensional conditional Bernoulli distribution simulation is not yet implemented!\n" );
	exit( -1 );
      }
    break;
  case DIST_UNICONDMVME:
    prob_vector=new double[num_states];
    for( i=0; i<num_states; i++ )
      prob_vector[i]=0.0;

    /* Computing probabilities for all outcomes */
    for( i=0; i<dim; i++ )
      {
	temp=1.0;
	for( j=1; j<num_features[i]; j++ )
	  if( feature_index[i][j]<prev_datum->ddim )
	    { /* Categorical feature */
	      if( feature_value[i][j]!=prev_datum->ddata[feature_index[i][j]] )
		{
		  temp=0.0;
		  j=num_features[i];
		}
	    } /* Categorical feature */
	  else
	    /* Real-valued feature */
	    temp*=prev_datum->rdata[feature_index[i][j]-prev_datum->ddim];

	prob_vector[feature_value[i][0]]+=temp;
      }

    temp=prob_vector[0];
    for( i=1; i<num_states; i++ )
      if( prob_vector[i]>temp )
	temp=prob_vector[i];

    sum=0.0;
    for( i=0; i<num_states; i++ )
      sum+=exp(prob_vector[i]-temp);

    for( i=0; i<num_states; i++ )
      prob_vector[i]=exp(prob_vector[i]-temp)/sum;

    sim->ddata[0]=generateBernoulli( prob_vector, num_states );

    delete prob_vector;
    break;
  case DIST_CHOWLIU:
    /* When generating the model, assuming that postprocessing step is run as well.
       Generating the nodes in the order in which they appear in the edges. */

    ddata=new long[dim];

    generated=new long[dim];
    for( i=0; i<dim; i++ )
      generated[i]=0;

    prob_vector=new double[num_states];

    for( e=0; e<num_edges; e++ )
      {
	/* Two possibilities:
	   both nodes need to be simulated, or just one */
	if( !generated[edge[e][0]] && !generated[edge[e][1]] )
	  { /* Both nodes need to be simulated */
	    /* Simulating the first node of the edge */
	    ddata[edge[e][0]]=generateBernoulli( ind_prob[edge[e][0]], num_states );
	    generated[edge[e][0]]=1;

	    /* Simulating the second node of the edge */
	    /* Creating the probability distribution first */
	    for( i=0; i<num_states; i++ )
	      prob_vector[i]=edge_prob[e][ddata[edge[e][0]]][i]/ind_prob[edge[e][0]][ddata[edge[e][0]]];

	    ddata[edge[e][1]]=generateBernoulli( prob_vector, num_states );
	    generated[edge[e][1]]=1;
	  } /* Both nodes need to be simulated */
	else
	  { /* Only one node needs to be simulated */
	    if( !generated[edge[e][0]] )
	      { /* Simulating the first node of the edge given the second */
		/* Creating the probability distribution first */
		for( i=0; i<num_states; i++ )
		  prob_vector[i]=edge_prob[e][i][ddata[edge[e][1]]]/ind_prob[edge[e][1]][ddata[edge[e][1]]];
		
		ddata[edge[e][0]]=generateBernoulli( prob_vector, num_states );
		generated[edge[e][0]]=1;
	      } /* Simulating the first node of the edge given the second */
	    
	    if( !generated[edge[e][1]] )
	      { /* Simulating the second node of the edge given the first */
		/* Creating the probability distribution first */
		for( i=0; i<num_states; i++ )
		  prob_vector[i]=edge_prob[e][ddata[edge[e][0]]][i]/ind_prob[edge[e][0]][ddata[edge[e][0]]];

		ddata[edge[e][1]]=generateBernoulli( prob_vector, num_states );
		generated[edge[e][1]]=1;
	      } /* Simulating the second node of the edge given the first */
	  } /* Only one node needs to be simulated */    
      }

    /* Making sure all nodes have been simulated */
    for( i=0; i<dim; i++ )
      if( !generated[i] )
	/* Simulating the node form individual probabilities */
	ddata[i]=generateBernoulli( ind_prob[i], num_states );
      
    delete prob_vector;

    delete generated;

    for( i=0; i<dim; i++ )
      sim->ddata[dim_index[i]]=ddata[i];
    delete ddata;
    
    break;
  case DIST_CONDCHOWLIU:
    /* When generating the model, assuming that postprocessing step is run as well.
       Generating the nodes in the order in which they appear in the edges. */

    ddata=new long[dim];

    generated=new long[2*dim];
    for( i=0; i<dim; i++ )
      {
	generated[i]=1;
	generated[i+dim]=0;
      }

    prob_vector=new double[num_states];

    for( e=0; e<num_edges; e++ )
      {
	/* Two possibilities:
	   both nodes need to be simulated, or just one */
	if( !generated[edge[e][0]] && !generated[edge[e][1]] )
	  { /* Both nodes need to be simulated */
	    /* Simulating the first node of the edge */
	    ddata[edge[e][0]-dim]=generateBernoulli( ind_prob[edge[e][0]], num_states );
	    generated[edge[e][0]]=1;

	    /* Simulating the second node of the edge */
	    /* Creating the probability distribution first */
	    for( i=0; i<num_states; i++ )
	      prob_vector[i]=edge_prob[e][ddata[edge[e][0]-dim]][i]/
		ind_prob[edge[e][0]-dim][ddata[edge[e][0]-dim]];

	    ddata[edge[e][1]-dim]=generateBernoulli( prob_vector, num_states );
	    generated[edge[e][1]]=1;
	  } /* Both nodes need to be simulated */
	else
	  { /* Only one node needs to be simulated */
	    if( !generated[edge[e][0]] )
	      { /* Simulating the first node of the edge given the second */
		/* Creating the probability distribution first */
		if( edge[e][1]<dim )
		  if( prev_datum )
		    temp_value=prev_datum->ddata[dim_index[edge[e][1]]];
		  else
		    temp_value=missing_value( (long)0 );
		else
		  temp_value=ddata[edge[e][1]-dim];

		if( !is_missing( temp_value ) )
		  for( i=0; i<num_states; i++ )
		    prob_vector[i]=edge_prob[e][i][temp_value]/ind_prob[edge[e][1]][temp_value];
		else
		  for( i=0; i<num_states; i++ )
		    prob_vector[i]=ind_prob[edge[e][0]][i];

		ddata[edge[e][0]-dim]=generateBernoulli( prob_vector, num_states );
		generated[edge[e][0]]=1;
	      } /* Simulating the first node of the edge given the second */
	    
	    if( !generated[edge[e][1]] )
	      { /* Simulating the second node of the edge given the first */
		/* Creating the probability distribution first */
		if( edge[e][0]<dim )
		  if( prev_datum )		    
		    temp_value=prev_datum->ddata[dim_index[edge[e][0]]];
		  else
		    temp_value=missing_value( (long)0 );
		else
		  temp_value=ddata[edge[e][0]-dim];

		if( !is_missing( temp_value ) )
		  for( i=0; i<num_states; i++ )
		    prob_vector[i]=edge_prob[e][temp_value][i]/ind_prob[edge[e][0]][temp_value];
		else
		  for( i=0; i<num_states; i++ )
		    prob_vector[i]=ind_prob[edge[e][1]][i];

		ddata[edge[e][1]-dim]=generateBernoulli( prob_vector, num_states );
		generated[edge[e][1]]=1;
	      } /* Simulating the second node of the edge given the first */
	  } /* Only one node needs to be simulated */    
      }

    /* Making sure all nodes have been simulated */
    for( i=dim; i<2*dim; i++ )
      if( !generated[i] )
	/* Simulating the node form individual probabilities */
	ddata[i-dim]=generateBernoulli( ind_prob[i], num_states );
      
    delete prob_vector;

    delete generated;

    for( i=0; i<dim; i++ )
      sim->ddata[dim_index[i]]=ddata[i];
    delete ddata;
    
    break;
  case DIST_ME_BIVAR:
    /* Simulation by Gibbs sampling */
    if( is_sim_initialized )
      /* Already performed burn-off */
      num_gibbs_sims=GIBBS_LAG;
    else
      { /* Need to perform the burn-off first */
	is_sim_initialized=1;
	last_sim=new long[dim];

	/* Initializing at random */
	for( i=0; i<dim; i++ )
	  if( drand48()>0.5 )
	    last_sim[i]=1;
	  else
	    last_sim[i]=0;

	num_gibbs_sims=GIBBS_BURNIN;
      } /* Need to perform the burn-off first */

    for( sim_index=0; sim_index<num_gibbs_sims; sim_index++ )
      /* Performing Gibbs sampling */
      for( i=0; i<dim; i++ )
	{ /* Resampling component i */
	  /* Calculating probability of outcome 0: 1/(exp(sum)+1) */
	  sum=0.0;
	  for( j=0; j<i; j++ )
	    if( last_sim[j]==1 )
	      sum+=sigma[i][j];
	  sum+=sigma[i][i];
	  for( j=i+1; j<dim; j++ )
	    if( last_sim[j]==1 )
	      sum+=sigma[j][i];
	  
	  /* Sampling */
	  if( drand48()<1.0/(exp(sum)+1.0) )
	    last_sim[i]=0;
	  else
	    last_sim[i]=1;
	} /* Resampling component i */

    /* Copying the last sample into the output */
    for( i=0; i<dim; i++ )
      sim->ddata[dim_index[i]]=last_sim[i];

    break;
  case DIST_BN_ME_BIVAR:
    for( i=0; i<dim; i++ )
      {
	/* Computing the probability of outcome 0: 1/(exp(sum)+1) */
	sum=sigma[sim_order[i]][0];
	for( j=1; j<num_features[sim_order[i]]; j++ )
	  if( sim->ddata[dim_index[feature_index[sim_order[i]][j]]]==1 )
	    sum+=sigma[sim_order[i]][j];
	
	/* Sampling */
	if( drand48()<1.0/(1.0+exp(sum)) )
	  sim->ddata[dim_index[sim_order[i]]]=0;
	else
	  sim->ddata[dim_index[sim_order[i]]]=1;
      }
    break;
  case DIST_BN_CME_BIVAR:
    for( i=0; i<dim; i++ )
      if( !prev_datum )
	{ /* First entry in a sequence -- using state_prob */
	  if( drand48()<state_prob[i] )
	    sim->ddata[dim_index[i]]=1;
	  else
	    sim->ddata[dim_index[i]]=0;
	} /* First entry in a sequence -- using state_prob */
      else
	{
	  /* Computing the probability of outcome 0: 1/(exp(sum)+1) */
	  sum=sigma[sim_order[i]][0];
	  for( j=1; j<num_features[sim_order[i]]; j++ )
	    if( feature_index[sim_order[i]][j]<dim )
	      {
		if( sim->ddata[dim_index[feature_index[sim_order[i]][j]]]==1 )
		  sum+=sigma[sim_order[i]][j];
	      }
	    else
	      {
		if( prev_datum->ddata[dim_index[feature_index[sim_order[i]][j]-dim]]==1 )
		  sum+=sigma[sim_order[i]][j];
	      }
	      
	  /* Sampling */
	  if( drand48()<1.0/(1.0+exp(sum)) )
	    sim->ddata[dim_index[sim_order[i]]]=0;
	  else
	    sim->ddata[dim_index[sim_order[i]]]=1;
	}
    break;
  case DIST_DELTAEXP:
    i=generateBernoulli( mix_prob, num_states );
    if( i==0 )
      sim->rdata[dim_index[0]]=0.0;
    else
      sim->rdata[dim_index[0]]=generateExponential( exp_param[i-1] );
    break;
  case DIST_DELTAGAMMA:
    i=generateBernoulli( mix_prob, num_states );
    if( i==0 )
      sim->rdata[dim_index[0]]=0.0;
    else
      sim->rdata[dim_index[0]]=generateGamma( gamma_param1[i-1], gamma_param2[i-1] );
    break;
  case DIST_DIRACDELTA:
    sim->rdata[dim_index[0]]=delta_value;
    break;
  case DIST_EXP:
    sim->rdata[dim_index[0]]=generateExponential( exp_param1 );
    break;
  case DIST_GAMMA:
    sim->rdata[dim_index[0]]=generateGamma( gamma1, gamma2 );
    break;
  case DIST_NORMAL:
    /* If Cholesky decomposition is not computed yet, doing so now */
    if( !chol_sigma )
      {
	chol_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  chol_sigma[i]=new double[dim];

	cholesky( sigma, dim, chol_sigma );
      }

    rdata=new double[dim];
    SimulateNormal( mu, chol_sigma, rdata, dim );
    for( i=0; i<dim; i++ )
      sim->rdata[dim_index[i]]=rdata[i];
    delete rdata;
    break;
  case DIST_NORMALCHAIN:
    /* If Cholesky decomposition is not computed yet, doing so now */
    if( !chol_sigma )
      {
	chol_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  chol_sigma[i]=new double[dim];

	cholesky( sigma, dim, chol_sigma );
      }
    if( !chol_first_sigma )
      {
	chol_first_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  chol_first_sigma[i]=new double[dim];

	cholesky( first_sigma, dim, chol_first_sigma );
      }

    rdata=new double[dim];

    if( prev_datum )
      { /* Not the first observation in the sequence */

	/* Allocating mean vector */
	prob_vector=new double[dim];

	/* Calculating the value for the mean vector */
	for( i=0; i<dim; i++ )
	  {
	    prob_vector[i]=mu[i];
	    for( e=0; e<dim; e++ )
	      prob_vector[i]+=W[i][e]*prev_datum->rdata[dim_index[e]];
	  }

	SimulateNormal( prob_vector, chol_sigma, rdata, dim );

	delete prob_vector;
      } /* Not the first observation in the sequence */
    else
      /* First observation in the sequence */
      SimulateNormal( first_mu, chol_first_sigma, rdata, dim );

    for( i=0; i<dim; i++ )
      sim->rdata[dim_index[i]]=rdata[i];
    delete rdata;

    break;
  case DIST_NORMALCL:
    /* !!! May be replaced by more efficient sampling !!! */
    /* If Cholesky decomposition is not computed yet, doing so now */
    if( !chol_sigma )
      {
	chol_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  chol_sigma[i]=new double[dim];

	cholesky( sigma, dim, chol_sigma );
      }

    rdata=new double[dim];
    SimulateNormal( mu, chol_sigma, rdata, dim );
    for( i=0; i<dim; i++ )
      sim->rdata[dim_index[i]]=rdata[i];
    delete rdata;
    break;
  case DIST_LOGISTIC:
  case DIST_TRANSLOGISTIC:
    fprintf( stderr, "Simulation from the logistic data is not yet implemented!\n" );
    exit( -1 );
    break;
  default:;
  }

  return;
}

void Distribution::SimulateMissingEntries( DataPoint* datum, DataPoint* prev_datum ) {
  /* Simulating an observation according to distribution */

  double** mult;
  double* prob;
  long* sim_value;

  long** node_index;

  /* Index variable(s) */
  long i,j,env,m1,b,e,ei,i1,i2,value1,value2;

  /* Temporary variables */
  double sum;

  switch( type ) {
  case DIST_FACTOR:
    for( i=0; i<dim; i++ )
      subdist[i]->SimulateMissingEntries( datum, prev_datum );
    break;
  case DIST_MIXTURE:
    prob=new double[num_states];

    /* Computing posterior probabilities */
    sum=0.0;
    for( b=0; b<num_states; b++ )
      {
	prob[b]=state_prob[b]*subdist[b]->prob( datum, prev_datum );
	sum+=prob[b];
      }
    for( b=0; b<num_states; b++ )
      prob[b]/=sum;

    /* Picking a mixture component */
    i=generateBernoulli( prob, num_states );

    /* Simulating a data point from the component */
    subdist[i]->SimulateMissingEntries( datum, prev_datum );

    delete prob;

    break;
  case DIST_CLMIXTURE:
    /* Tree-structured mixture */
    /* Computing the envelopes */

    /* Allocating the array of probabilities for the current value of the latent variable */
    prob=new double[num_states];

    /* Allocating additional variables */
    mult=new double*[dim];
    for( i=0; i<dim; i++ )
      mult[i]=new double[num_states];

    sim_value=new long[dim];

    /* Precomputing probability values for components */
    for( i=0; i<dim; i++ )
      for( j=0; j<num_states; j++ )
	mult[i][j]=subdist2[i][j]->prob( datum, prev_datum );

    for( env=0; env<num_envelopes; env++ )
      /* For each envelope */
      if( envelope[env]->num_nodes==1 )
	{ /* Envelope contains exactly one variable. */
	  m1=envelope[env]->node[0];
	  sum=0.0;
	  for( b=0; b<num_states; b++ )
	    { /* Summing over the latent variable */
	      prob[b]=ind_prob[m1][b]*mult[m1][b];
	      sum+=prob[b];
	    } /* Summing over the latent variable */

	  /* Normalizing the probability */
	  for( b=0; b<num_states; b++ )
	    prob[b]/=sum;

	  /* Simulating the value for the latent variable */
	  sim_value[m1]=generateBernoulli( prob, num_states );

	  /* Simulating missing values for the appropriate component */
	  subdist2[m1][sim_value[m1]]->SimulateMissingEntries( datum, prev_datum );
	} /* Envelope contains exactly one missing variable. */
      else
	{ /* Envelope contains more than one variable. */
	  /* Propagating probabilities in a backward pass */
	  for( e=envelope[env]->num_edges-1; e>=0; e-- )
	    {
	      ei=envelope[env]->edge[e];
	      i1=edge[ei][0];
	      i2=edge[ei][1];
		
	      if( envelope[env]->node[e+1]==i2 )
		{ /* i1 is the parent of i2 */
		  for( value1=0; value1<num_states; value1++ )
		    { /* Computing the contribution to i1 with value value1 from child i2 */
		      sum=0.0;
		      for( value2=0; value2<num_states; value2++ )
			sum+=mult[i2][value2]*edge_prob[ei][value1][value2];
		      mult[i1][value1]*=sum/ind_prob[i1][value1];
		    } /* Computing the contribution to i1 with value value1 from child i2 */
		} /* i1 is the parent of i2 */
	      else
		{ /* i2 is the parent of i1 */
		  for( value2=0; value2<num_states; value2++ )
		    { /* Computing the contribution to i2 with value value2 from child i1 */
		      sum=0.0;
		      for( value1=0; value1<num_states; value1++ )
			sum+=mult[i1][value1]*edge_prob[ei][value1][value2];
		      mult[i2][value2]*=sum/ind_prob[i2][value2];
		    } /* Computing the contribution to i2 with value value2 from child i1 */
		} /* i2 is the parent of i1 */		
	    }

	  /* Summing over the root of the envelope */
	  m1=envelope[env]->node[0];
	  sum=0.0;
	  for( b=0; b<num_states; b++ )
	    { /* Summing over the latent variable */
	      prob[b]=mult[m1][b]*ind_prob[m1][b];
	      sum+=prob[b];
	    } /* Summing over the latent variable */

	  /* Normalizing the probability */
	  for( b=0; b<num_states; b++ )
	    prob[b]/=sum;
	  
	  /* Simulating the value for the latent variable */
	  sim_value[m1]=generateBernoulli( prob, num_states );

	  /* Simulating missing values for the appropriate component */
	  subdist2[m1][sim_value[m1]]->SimulateMissingEntries( datum, prev_datum );

	  /* Backward pass */
	  for( e=0; e<envelope[env]->num_edges; e++ )
	    {
	      ei=envelope[env]->edge[e];
	      i1=edge[ei][0];
	      i2=edge[ei][1];
	      
	      if( envelope[env]->node[e+1]==i2 )
		{ /* i1 is the parent of i2 */
		  sum=0.0;
		  for( b=0; b<num_states; b++ )
		    {
		      prob[b]=mult[i2][b]*edge_prob[ei][sim_value[i1]][b];
		      sum+=prob[b];
		    }

		  /* Normalizing */
		  for( b=0; b<num_states; b++ )
		    prob[b]/=sum;

		  /* Simulating value for a latent variable */
		  sim_value[i2]=generateBernoulli( prob, num_states );

		  /* Simulating missing values for the appropriate component */
		  subdist2[i2][sim_value[i2]]->SimulateMissingEntries( datum, prev_datum );
		} /* i1 is the parent of i2 */
	      else
		{ /* i2 is the parent of i1 */
		  sum=0.0;
		  for( b=0; b<num_states; b++ )
		    {
		      prob[b]=mult[i1][b]*edge_prob[ei][b][sim_value[i2]];
		      sum+=prob[b];
		    }

		  /* Normalizing */
		  for( b=0; b<num_states; b++ )
		    prob[b]/=sum;

		  /* Simulating value for a latent variable */
		  sim_value[i1]=generateBernoulli( prob, num_states );

		  /* Simulating missing values for the appropriate component */
		  subdist2[i1][sim_value[i1]]->SimulateMissingEntries( datum, prev_datum );
		} /* i2 is the parent of i1 */
	    }

	} /* Envelope contains more than one variable. */
    
    /* Deallocating */
    delete sim_value;

    for( i=0; i<dim; i++ )
      delete mult[i];
    delete mult;
    
    delete prob;

    break;
  case DIST_BERNOULLI:
    if( dim==1 )
      if( is_missing( datum->ddata[dim_index[0]] ) )
	datum->ddata[dim_index[0]]=generateBernoulli( state_prob, num_states );
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    if( dim==1 )
      if( is_missing( datum->ddata[dim_index[0]] ) )
	{
	  if( prev_datum )
	    /* Not the first sequence entry */
	    datum->ddata[dim_index[0]]=generateBernoulli( cond_state_prob[prev_datum->ddata[dim_index[0]]], num_states );
	  else
	    /* First sequence entry */
	    datum->ddata[dim_index[0]]=generateBernoulli( state_prob, num_states );
	}
    break;
  case DIST_UNICONDMVME:
    /* !!! Do not know what to do with this yet !!! */
    break;
  case DIST_CHOWLIU:
  case DIST_CHOWLIU_MDL:
    /* Allocating the array of probabilities for the current value of the latent variable */
    prob=new double[num_states];

    /* Computing the envelopes */
    node_index=new long*[dim];
    for( i=0; i<dim; i++ )
      node_index[i]=new long[2];
    envelope=new Envelope*[dim];

    num_envelopes=compute_envelopes( datum, node_index );

    /* Allocating additional variables */
    mult=new double*[dim];
    for( i=0; i<dim; i++ )
      mult[i]=new double[num_states];

    for( env=0; env<num_envelopes; env++ )
      /* For each envelope */
      if( is_missing(datum->ddata[dim_index[envelope[env]->node[0]]]) )
	{ /* Envelope contains missing values */
	  if( envelope[env]->num_nodes==1 )
	    { /* Envelope contains exactly one missing variable. */
	      m1=envelope[env]->node[0];
	      for( b=0; b<num_states; b++ )
		prob[b]=ind_prob[m1][b];
	      sum=0.0;
	      for( b=0; b<num_states; b++ )
		{/* Summing over the missing variable */
		  for( e=0; e<envelope[env]->num_edges; e++ )
		    {
		      ei=envelope[env]->edge[e];
		      i1=edge[ei][0];
		      i2=edge[ei][1];
		      
		      if( m1==i1 )
			prob[b]*=edge_prob[ei][b][datum->ddata[dim_index[i2]]]/ind_prob[i1][b];
		      else
			prob[b]*=edge_prob[ei][datum->ddata[dim_index[i1]]][b]/ind_prob[i1][b];
		    }
		  sum+=prob[b];
		} /* Summing over the missing variable */
	      
	      /* Normalizing the probability */
	      for( b=0; b<num_states; b++ )
		prob[b]/=sum;

	      /* Simulating the value */
	      datum->ddata[dim_index[m1]]=generateBernoulli( prob, num_states );
	    } /* Envelope contains exactly one missing variable. */
	  else
	    { /* Envelope contains more than one missing variable. */
	      for( i=0; i<envelope[env]->num_nodes; i++ )
		for( b=0; b<num_states; b++ )
		  mult[envelope[env]->node[i]][b]=1.0;

	      /* Propagating probabilities in a backward pass */
	      for( e=envelope[env]->num_edges-1; e>=0; e-- )
		{
		  ei=envelope[env]->edge[e];
		  i1=edge[ei][0];
		  i2=edge[ei][1];
		
		  if( !is_missing(datum->ddata[dim_index[i1]] ) )
		    /* Second node of the edge is missing. */
		    for( b=0; b<num_states; b++ )
		      mult[i2][b]*=edge_prob[ei][datum->ddata[dim_index[i1]]][b]/ind_prob[i2][b];
		  else
		    if( !is_missing( datum->ddata[dim_index[i2]] ) )
		      /* First node of the edge is missing. */
		      for( b=0; b<num_states; b++ )
			mult[i1][b]*=edge_prob[ei][b][datum->ddata[dim_index[i2]]]/ind_prob[i1][b];
		    else
		      /* Both nodes are missing. */
		      if( node_index[i1][1]<node_index[i2][1] )
			{ /* i1 is the parent of i2 */
			  for( value1=0; value1<num_states; value1++ )
			    { /* Computing the contribution to i1 with value value1 from child i2 */
			      sum=0.0;
			      for( value2=0; value2<num_states; value2++ )
				sum+=mult[i2][value2]*edge_prob[ei][value1][value2];
			      mult[i1][value1]*=sum/ind_prob[i1][value1];
			    } /* Computing the contribution to i1 with value value1 from child i2 */
			} /* i1 is the parent of i2 */
		      else
			{ /* i2 is the parent of i1 */
			  for( value2=0; value2<num_states; value2++ )
			    { /* Computing the contribution to i2 with value value2 from child i1 */
			      sum=0.0;
			      for( value1=0; value1<num_states; value1++ )
				sum+=mult[i1][value1]*edge_prob[ei][value1][value2];
			      mult[i2][value2]*=sum/ind_prob[i2][value2];
			    } /* Computing the contribution to i2 with value value2 from child i1 */
			} /* i2 is the parent of i1 */		
		}
	      
	      /* Summing over the root of the envelope */
	      m1=envelope[env]->node[0];
	      sum=0.0;
	      for( b=0; b<num_states; b++ )
		{
		  prob[b]=mult[m1][b]*ind_prob[m1][b];
		  sum+=prob[b];
		}
	      /* Normalizing */
	      for( b=0; b<num_states; b++ )
		prob[b]/=sum;

	      /* Simulating the value */
	      datum->ddata[dim_index[m1]]=generateBernoulli( prob, num_states );

	      /* Backward pass */
	      for( e=0; e<envelope[env]->num_edges; e++ )
		{
		  ei=envelope[env]->edge[e];
		  i1=edge[ei][0];
		  i2=edge[ei][1];

		  if( is_missing( datum->ddata[i1] ) )
		    { /* First node is missing */
		      sum=0.0;
		      for( b=0; b<num_states; b++ )
			{
			  prob[b]=mult[i1][b]*edge_prob[ei][b][datum->ddata[dim_index[i2]]];
			  sum+=prob[b];
			}
		      for( b=0; b<num_states; b++ )
			prob[b]/=sum;

		      datum->ddata[dim_index[i1]]=generateBernoulli( prob, num_states );
		    } /* First node is missing */
		  else if( is_missing( datum->ddata[i2] ) )
		    { /* Second node is missing */
		      sum=0.0;
		      for( b=0; b<num_states; b++ )
			{
			  prob[b]=mult[i2][b]*edge_prob[ei][b][datum->ddata[dim_index[i1]]];
			  sum+=prob[b];
			}
		      for( b=0; b<num_states; b++ )
			prob[b]/=sum;

		      datum->ddata[dim_index[i2]]=generateBernoulli( prob, num_states );
		    } /* Second node is missing */
		}
	  } /* Envelope contains more than one missing variable. */
	} /* Envelope contains missing values */

    /* Deallocating */
    for( i=0; i<dim; i++ )
      delete mult[i];
    delete mult;
    for( i=0; i<dim; i++ )
      delete node_index[i];
    delete node_index;
    for( env=0; env<num_envelopes; env++ )
      delete envelope[env];
    delete envelope;

    /* Deallocating temporary array of probabilities */
    delete prob;

    break;

  case DIST_CONDCHOWLIU:
    /* Do not know how to do that yet */
    break;
#ifdef BLAH
    /* When generating the model, assuming that postprocessing step is run as well.
       Generating the nodes in the order in which they appear in the edges. */

    ddata=new long[dim];

    generated=new long[2*dim];
    for( i=0; i<dim; i++ )
      {
	generated[i]=1;
	generated[i+dim]=0;
      }

    prob_vector=new double[num_states];

    for( e=0; e<num_edges; e++ )
      {
	/* Two possibilities:
	   both nodes need to be simulated, or just one */
	if( !generated[edge[e][0]] && !generated[edge[e][1]] )
	  { /* Both nodes need to be simulated */
	    /* Simulating the first node of the edge */
	    ddata[edge[e][0]-dim]=generateBernoulli( ind_prob[edge[e][0]], num_states );
	    generated[edge[e][0]]=1;

	    /* Simulating the second node of the edge */
	    /* Creating the probability distribution first */
	    for( i=0; i<num_states; i++ )
	      prob_vector[i]=edge_prob[e][ddata[edge[e][0]-dim]][i]/
		ind_prob[edge[e][0]-dim][ddata[edge[e][0]-dim]];

	    ddata[edge[e][1]-dim]=generateBernoulli( prob_vector, num_states );
	    generated[edge[e][1]]=1;
	  } /* Both nodes need to be simulated */
	else
	  { /* Only one node needs to be simulated */
	    if( !generated[edge[e][0]] )
	      { /* Simulating the first node of the edge given the second */
		/* Creating the probability distribution first */
		if( edge[e][1]<dim )
		  if( prev_datum )
		    temp_value=prev_datum->ddata[dim_index[edge[e][1]]];
		  else
		    temp_value=missing_value( (long)0 );
		else
		  temp_value=ddata[edge[e][1]-dim];

		if( !is_missing( temp_value ) )
		  for( i=0; i<num_states; i++ )
		    prob_vector[i]=edge_prob[e][i][temp_value]/ind_prob[edge[e][1]][temp_value];
		else
		  for( i=0; i<num_states; i++ )
		    prob_vector[i]=ind_prob[edge[e][0]][i];

		ddata[edge[e][0]-dim]=generateBernoulli( prob_vector, num_states );
		generated[edge[e][0]]=1;
	      } /* Simulating the first node of the edge given the second */
	    
	    if( !generated[edge[e][1]] )
	      { /* Simulating the second node of the edge given the first */
		/* Creating the probability distribution first */
		if( edge[e][0]<dim )
		  if( prev_datum )		    
		    temp_value=prev_datum->ddata[dim_index[edge[e][0]]];
		  else
		    temp_value=missing_value( (long)0 );
		else
		  temp_value=ddata[edge[e][0]-dim];

		if( !is_missing( temp_value ) )
		  for( i=0; i<num_states; i++ )
		    prob_vector[i]=edge_prob[e][temp_value][i]/ind_prob[edge[e][0]][temp_value];
		else
		  for( i=0; i<num_states; i++ )
		    prob_vector[i]=ind_prob[edge[e][1]][i];

		ddata[edge[e][1]-dim]=generateBernoulli( prob_vector, num_states );
		generated[edge[e][1]]=1;
	      } /* Simulating the second node of the edge given the first */
	  } /* Only one node needs to be simulated */    
      }

    /* Making sure all nodes have been simulated */
    for( i=dim; i<2*dim; i++ )
      if( !generated[i] )
	/* Simulating the node form individual probabilities */
	ddata[i-dim]=generateBernoulli( ind_prob[i], num_states );
      
    delete prob_vector;

    delete generated;

    for( i=0; i<dim; i++ )
      sim->ddata[dim_index[i]]=ddata[i];
    delete ddata;
    
    break;
#endif
  case DIST_ME_BIVAR:
    /* Not yet implemented */
    break;
#ifdef BLAH
    /* Simulation by Gibbs sampling */
    if( is_sim_initialized )
      /* Already performed burn-off */
      num_gibbs_sims=GIBBS_LAG;
    else
      { /* Need to perform the burn-off first */
	is_sim_initialized=1;
	last_sim=new long[dim];

	/* Initializing at random */
	for( i=0; i<dim; i++ )
	  if( drand48()>0.5 )
	    last_sim[i]=1;
	  else
	    last_sim[i]=0;

	num_gibbs_sims=GIBBS_BURNIN;
      } /* Need to perform the burn-off first */

    for( sim_index=0; sim_index<num_gibbs_sims; sim_index++ )
      /* Performing Gibbs sampling */
      for( i=0; i<dim; i++ )
	{ /* Resampling component i */
	  /* Calculating probability of outcome 0: 1/(exp(sum)+1) */
	  sum=0.0;
	  for( j=0; j<i; j++ )
	    if( last_sim[j]==1 )
	      sum+=sigma[i][j];
	  sum+=sigma[i][i];
	  for( j=i+1; j<dim; j++ )
	    if( last_sim[j]==1 )
	      sum+=sigma[j][i];
	  
	  /* Sampling */
	  if( drand48()<1.0/(exp(sum)+1.0) )
	    last_sim[i]=0;
	  else
	    last_sim[i]=1;
	} /* Resampling component i */

    /* Copying the last sample into the output */
    for( i=0; i<dim; i++ )
      sim->ddata[dim_index[i]]=last_sim[i];
#endif
  case DIST_BN_ME_BIVAR:
    /* Not yet implemented */
    break;
#ifdef BLAH
    for( i=0; i<dim; i++ )
      {
	/* Computing the probability of outcome 0: 1/(exp(sum)+1) */
	sum=sigma[sim_order[i]][0];
	for( j=1; j<num_features[sim_order[i]]; j++ )
	  if( sim->ddata[dim_index[feature_index[sim_order[i]][j]]]==1 )
	    sum+=sigma[sim_order[i]][j];
	
	/* Sampling */
	if( drand48()<1.0/(1.0+exp(sum)) )
	  sim->ddata[dim_index[sim_order[i]]]=0;
	else
	  sim->ddata[dim_index[sim_order[i]]]=1;
      }
    break;
#endif
  case DIST_BN_CME_BIVAR:
    /* Do not know how to perform this */
    break;
#ifdef BLAH
    for( i=0; i<dim; i++ )
      if( !prev_datum )
	{ /* First entry in a sequence -- using state_prob */
	  if( drand48()<state_prob[i] )
	    sim->ddata[dim_index[i]]=1;
	  else
	    sim->ddata[dim_index[i]]=0;
	} /* First entry in a sequence -- using state_prob */
      else
	{
	  /* Computing the probability of outcome 0: 1/(exp(sum)+1) */
	  sum=sigma[sim_order[i]][0];
	  for( j=1; j<num_features[sim_order[i]]; j++ )
	    if( feature_index[sim_order[i]][j]<dim )
	      {
		if( sim->ddata[dim_index[feature_index[sim_order[i]][j]]]==1 )
		  sum+=sigma[sim_order[i]][j];
	      }
	    else
	      {
		if( prev_datum->ddata[dim_index[feature_index[sim_order[i]][j]-dim]]==1 )
		  sum+=sigma[sim_order[i]][j];
	      }
	      
	  /* Sampling */
	  if( drand48()<1.0/(1.0+exp(sum)) )
	    sim->ddata[dim_index[sim_order[i]]]=0;
	  else
	    sim->ddata[dim_index[sim_order[i]]]=1;
	}
    break;
  case DIST_DELTAEXP:
    i=generateBernoulli( mix_prob, num_states );
    if( i==0 )
      sim->rdata[dim_index[0]]=0.0;
    else
      sim->rdata[dim_index[0]]=generateExponential( exp_param[i-1] );
    break;
  case DIST_DELTAGAMMA:
    i=generateBernoulli( mix_prob, num_states );
    if( i==0 )
      sim->rdata[dim_index[0]]=0.0;
    else
      sim->rdata[dim_index[0]]=generateGamma( gamma_param1[i-1], gamma_param2[i-1] );
    break;
  case DIST_DIRACDELTA:
    sim->rdata[dim_index[0]]=delta_value;
    break;
  case DIST_EXP:
    sim->rdata[dim_index[0]]=generateExponential( exp_param1 );
    break;
  case DIST_GAMMA:
    sim->rdata[dim_index[0]]=generateGamma( gamma1, gamma2 );
    break;
  case DIST_NORMAL:
    /* If Cholesky decomposition is not computed yet, doing so now */
    if( !chol_sigma )
      {
	chol_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  chol_sigma[i]=new double[dim];

	cholesky( sigma, dim, chol_sigma );
      }

    rdata=new double[dim];
    SimulateNormal( mu, chol_sigma, rdata, dim );
    for( i=0; i<dim; i++ )
      sim->rdata[dim_index[i]]=rdata[i];
    delete rdata;
    break;
  case DIST_NORMALCHAIN:
    /* If Cholesky decomposition is not computed yet, doing so now */
    if( !chol_sigma )
      {
	chol_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  chol_sigma[i]=new double[dim];

	cholesky( sigma, dim, chol_sigma );
      }
    if( !chol_first_sigma )
      {
	chol_first_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  chol_first_sigma[i]=new double[dim];

	cholesky( first_sigma, dim, chol_first_sigma );
      }

    rdata=new double[dim];

    if( prev_datum )
      { /* Not the first observation in the sequence */

	/* Allocating mean vector */
	prob_vector=new double[dim];

	/* Calculating the value for the mean vector */
	for( i=0; i<dim; i++ )
	  {
	    prob_vector[i]=mu[i];
	    for( e=0; e<dim; e++ )
	      prob_vector[i]+=W[i][e]*prev_datum->rdata[dim_index[e]];
	  }

	SimulateNormal( prob_vector, chol_sigma, rdata, dim );

	delete prob_vector;
      } /* Not the first observation in the sequence */
    else
      /* First observation in the sequence */
      SimulateNormal( first_mu, chol_first_sigma, rdata, dim );

    for( i=0; i<dim; i++ )
      sim->rdata[dim_index[i]]=rdata[i];
    delete rdata;

    break;
  case DIST_NORMALCL:
    /* !!! May be replaced by more efficient sampling !!! */
    /* If Cholesky decomposition is not computed yet, doing so now */
    if( !chol_sigma )
      {
	chol_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  chol_sigma[i]=new double[dim];

	cholesky( sigma, dim, chol_sigma );
      }

    rdata=new double[dim];
    SimulateNormal( mu, chol_sigma, rdata, dim );
    for( i=0; i<dim; i++ )
      sim->rdata[dim_index[i]]=rdata[i];
    delete rdata;
    break;
  case DIST_LOGISTIC:
  case DIST_TRANSLOGISTIC:
    fprintf( stderr, "Simulation from the logistic data is not yet implemented!\n" );
    exit( -1 );
    break;
#endif
  default:;
  }

  return;
}
