/*
 * Copyright (c) 2002-2006, Sergey Kirshner and the Regents of 
 * the University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* viterbi.cpp */

#include "viterbi.h"

/* Routines for the structures of state sequences */
StateSequence::StateSequence( long sl, long def ) {
  /* Index variable(s) */

  seq_length=sl;
  default_num_best=def;
  if( def>MAXBESTSEQ )
    def=MAXBESTSEQ;
  num_best=0;
}

StateSequence::~StateSequence() {
  /* Index variable(s) */
  long i;

  for( i=0; i<num_best; i++ )
    delete state[i];
}

void StateSequence::AddSequence( long* new_seq, double new_ll ) {
  /* Adding a sequence to the list or discards (deallocates) it */
  
  /* Index variable(s) */
  long i;

  long new_index;
  
  /* !!! Need to insert hooks not to go over MAXBESTSEQ sequences !!! */
  if( new_ll>NEG_INF )
    {
      new_index=find_insert_index( ll, new_ll, num_best );
      if( new_index<default_num_best || (num_best>0 && new_ll>=ll[num_best-1]-VITERBI_EPSILON) )
	{ /* Adding the entry */
	  
	  /* Shifting the tail */
	  if( new_index<num_best )
	    { /* Shifting is needed */
	      if( !(num_best<MAXBESTSEQ) )
		{ /* !!! !!! */
		  exit( -1 );
		} /* !!! !!! */
	     
	      for( i=num_best; i>new_index; i-- )
		{
		  state[i]=state[i-1];
		  ll[i]=ll[i-1];
		}
	    }
	  
	  /* Inserting new sequence */
	  state[new_index]=new_seq;
	  ll[new_index]=new_ll;
	  num_best++;
	  
	  /* Chopping the tail */
	  if( num_best>default_num_best && ll[default_num_best-1]-VITERBI_EPSILON>ll[default_num_best] )
	    { /* Need to chop the tail */
	      for( i=default_num_best; i<num_best; i++ )
		delete state[i];
	      num_best=default_num_best;
	    } /* Need to chop the tail */
	} /* Adding the entry */
      else
	/* Deallocating the entry */
	delete new_seq;
    }
  else
    /* Deallocating the entry */
    delete new_seq;

  return;
}

void StateSequence::WriteToFile( FILE* output ) {
  /* Index variable(s) */
  long i,j;

  /* Displaying all best sequences */
  for( j=0; j<num_best; j++ )
    {
      for( i=0; i<seq_length; i++ )
	fprintf( output, " %ld", state[j][i] );
      /* Every sequences ends with a new line */
      fprintf( output, "\n" );
    }

  fflush( output );

  return;
}

StateData::StateData( long ns ) {
  /* Index variable(s) */
  
  /* Index variable(s) */
  long i;

  num_seqs=ns;
  sequence=new StateSequence*[num_seqs];
  for( i=0; i<num_seqs; i++ )
    sequence[i]=NULL;
  
}

StateData::~StateData() {
  /* Index variable(s) */
  long i;

  for( i=0; i<num_seqs; i++ )
    if( sequence[i] )
      delete sequence[i];
  delete sequence;
}

void StateData::WriteToFile( FILE* output ) {
  /* Index variable(s) */
  long i;

  for( i=0; i<num_seqs; i++ )
    {
      sequence[i]->WriteToFile( output );
      /* Each set of sequences is separated from the next by a line */
      fprintf( output, "\n" );
    }
}

StateData* HMM::viterbi( Data* data, Data* input, long num_best_seqs ) {
  /* Performing Viterbi algorithm to find most likely sequence(s) of states given the data
     and the model. */

  /* Variables */
  StateData* states;            // Output structure

  double* log_first;            // Logs of the probabilities of the first state
  double** log_trans;           // Logs of the transition matrix entries
  double** log_pb;              // Logs of the P(O_t|S_i)
  
  StateSequence** best_seq;     // Structure holding best sequences for a given final state
  StateSequence** new_best_seq; // Structure holding best sequences during the update

  /* Index variable(s) */
  long i,j,seq_index,t,k,p;

  /* Temporary variable(s) */
  long* long_array_temp;
  StateSequence** temp_seq;

  /* Allocating the structure for the state sequences */
  states=new StateData( data->num_seqs );

  /* Allocating the structures for the log-values */
  log_first=new double[num_states[0]];
  log_trans=new double*[num_states[0]];
  for( i=0; i<num_states[0]; i++ )
    log_trans[i]=new double[num_states[0]];

  if( transition->type==DIST_CONDBERNOULLI || transition->type==DIST_BERNOULLI )
    /* Calculating values of the logs of the first state or mixing parameters */
    for( i=0; i<num_states[0]; i++ )
      log_first[i]=log(transition->state_prob[i]);

  if( transition->type==DIST_CONDBERNOULLI )
    /* Calculating the values of the logs of the transition matrix entries */ 
    for( i=0; i<num_states[0]; i++ )
      for( j=0; j<num_states[0]; j++ )
	log_trans[i][j]=log(transition->cond_state_prob[i][j]);

  if( transition->type==DIST_BERNOULLI )
    /* Setting log transition values for the mixture model */
    for( i=0; i<num_states[0]; i++ )
      for( j=0; j<num_states[0]; j++ )
	log_trans[i][j]=log_first[j];

  /* Allocating log_pb */
  log_pb=new double*[num_states[0]];

  /* Allocating best_seq */
  best_seq=new StateSequence*[num_states[0]];

  /* Allocating new_best_seq */
  new_best_seq=new StateSequence*[num_states[0]];

  /* Starting the loop on each of the sequences */
  for( seq_index=0; seq_index<states->num_seqs; seq_index++ )
    { /* Generating best missing value sequences for each of the input sequences */

#ifdef VITERBI_VERBOSE
      printf( "Analyzing best state(s) for sequence #%ld\n", seq_index+1 );
      fflush( stdout );
#endif
    
      /* Allocating states[i] */
      states->sequence[seq_index]=new StateSequence( data->sequence[seq_index]->seq_length, num_best_seqs );

      /* Calculating log of P(O_t|S_t) */
      for( i=0; i<num_states[0]; i++ )
	{
	  log_pb[i]=new double[data->sequence[seq_index]->seq_length];
	  for( t=0; t<data->sequence[seq_index]->seq_length; t++ )
	    if( t==0 )
	      log_pb[i][t]=log(emission[0][i]->prob( data->sequence[seq_index]->entry[t], NULL ));
	    else
	      log_pb[i][t]=log(emission[0][i]->prob( data->sequence[seq_index]->entry[t],
						  data->sequence[seq_index]->entry[t-1] ));
	}

      for( i=0; i<num_states[0]; i++ )
	best_seq[i]=new StateSequence( 1, num_best_seqs );
    
      /* Initializing best state sequences and corresponding log-likelihoods */
      /* Considering sequences of length 1 */
      if( transition->type==DIST_TRANSLOGISTIC || transition->type==DIST_LOGISTIC )
	{
	  transition->probLogistic( input->sequence[seq_index]->entry[0], -1, log_first );
	  for( i=0; i<num_states[0]; i++ )
	    log_first[i]=log(log_first[i]);
	}
      
      for( i=0; i<num_states[0]; i++ )
	{
	  long_array_temp=new long[1];
	  long_array_temp[0]=i;
	  best_seq[i]->AddSequence( long_array_temp, log_first[i]+log_pb[i][0] );
	}
    
      for( t=1; t<data->sequence[seq_index]->seq_length; t++ )
	{ /* Considering subsequences of length t+1 */

	  /* Allocating new structures for new_best_seq for each final state */
	  for( i=0; i<num_states[0]; i++ )
	    new_best_seq[i]=new StateSequence( t+1, num_best_seqs );
	  
	  for( i=0; i<num_states[0]; i++ )
	    if( transition->type==DIST_TRANSLOGISTIC || transition->type==DIST_LOGISTIC )
	      {
		transition->probLogistic( input->sequence[seq_index]->entry[t], i, log_trans[i] );
		for( j=0; j<num_states[0]; j++ )
		  log_trans[i][j]=log(log_trans[i][j]);
	      }

	  for( j=0; j<num_states[0]; j++ )
	    { /* State of S_t+1 is j */
	      for( i=0; i<num_states[0]; i++ )
		{ /* State of S_t is i */
		  for( k=0; k<best_seq[i]->num_best; k++ )
		    { /* For each of the best sequences ending in state j */
		      
		      /* Generating new sequences */
		      long_array_temp=new long[t+1];
		      
		      /* Copying the prefix */
		      for( p=0; p<t; p++ )
			long_array_temp[p]=best_seq[i]->state[k][p];
		      
		      /* Adding the latest state (j) */
		      long_array_temp[t]=j;
		      
		      /* Trying to add the sequence to the list of the best */
		      new_best_seq[j]->AddSequence( long_array_temp, 
						    best_seq[i]->ll[k]+log_trans[i][j]+log_pb[j][t] );
		    } /* For each of the best sequences ending in state j */
		} /* State of S_t is i */
	    } /* State of S_t+1 is j */
	  
	  /* Moving new_best_seq into best_seq */
	  temp_seq=best_seq;
	  best_seq=new_best_seq;
	  new_best_seq=temp_seq;

	  /* Deallocating new structures for new_best_seq for each final state */
	  for( i=0; i<num_states[0]; i++ )
	    delete new_best_seq[i];
	} /* Considering subsequences of length t+1 */	  
  
      /* Finding the best sequences and likelihoods */
      for( j=0; j<num_states[0]; j++ )
	/* State of S_T is j */
	   
	for( k=0; k<best_seq[j]->num_best; k++ )
	  { /* k-th best sequence of ending in state j of length T */

	    /* Copying the sequence into a temp variable */
	    long_array_temp=new long[data->sequence[seq_index]->seq_length];
	    for( p=0; p<data->sequence[seq_index]->seq_length; p++ )
	      long_array_temp[p]=best_seq[j]->state[k][p];

	    /* Trying to add the sequence to the list of the best */
	    states->sequence[seq_index]->AddSequence( long_array_temp, best_seq[j]->ll[k] );
	  } /* k-th best sequence of ending in state j of length T */

      /* Deallocating best_seq */
      for( i=0; i<num_states[0]; i++ )
	delete best_seq[i];
      
      /* Deallocating log_pb */
      for( i=0; i<num_states[0]; i++ )
	delete log_pb[i];

#ifdef VITERBI_VERBOSE
      printf( "Optimal state(s) for sequence #%ld calculated\n", seq_index+1 );
      fflush( stdout );
#endif

    } /* Generating best missing value sequences for each of the input sequences */

  /* Deallocating variables */
  delete new_best_seq;
  
  delete best_seq;

  delete log_pb;

  for( i=0; i<num_states[0]; i++ )
    delete log_trans[i];
  delete log_trans;

  delete log_first;

  return( states );
}

long find_insert_index( double* ll_array, double ll, long ll_length ) {
  /* Returns the index indicating where the new entry should go in the decreasing-order array */
  /* Uses binary insertion */

  long high_index;
  long low_index;
  long mid_index;

  high_index=ll_length;
  low_index=0;
  mid_index=(long)floor(0.5*(double)(high_index+low_index));
  while( low_index<high_index )
    {    
      if( ll>ll_array[mid_index] )
	high_index=mid_index;
      else
	low_index=mid_index+1;
      
      mid_index=(long)floor(0.5*(double)(high_index+low_index));
    }

  return( mid_index );
}
