/*
 * Copyright (c) 2002-2006, Sergey Kirshner and the Regents of the
 * University of California
 * All rights reserved.
 *
 * This software is open-source under the BSD license; see "LICENSE"
*/

/* gammaln, digamma, and trigammma code by Thomas Minka */
/* http://www.stat.cmu.edu/~minka/papers/dirichlet/minka-digamma.zip */


/* distribution.cpp */

/* Includes */
#include "struct.h"

extern long SINGULAR_FAIL;
extern double cg_epsilon;

Distribution::Distribution( long d_type, long d_num_states, long d_dim ) {
  /* Index variable(s) */
  long i,j,i1;

  type=d_type;
  if( !(d_dim>0) )
    {
      fprintf( stderr, "Dimensionality of any distribution must be at least 1.  Aborting.\n" );
      exit( -1 );
    }
  dim=d_dim;
  num_states=d_num_states;

  /* Initializing parameter structures */
  switch( type ) {
  case DIST_FACTOR:
    subdist=new Distribution*[dim];
    for( i=0; i<dim; i++ )
      subdist[i]=NULL;
    break;
  case DIST_MIXTURE:
    subdist=new Distribution*[num_states];
    for( i=0; i<num_states; i++ )
      subdist[i]=NULL;
    state_prob=new double[num_states];
    for( i=0; i<num_states; i++ )
      state_prob[i]=0.0;    
    break;
  case DIST_CLMIXTURE:
    subdist2=new Distribution**[dim];
    for( i=0; i<dim; i++ )
      {
	subdist2[i]=new Distribution*[num_states];
	for( j=0; j<num_states; j++ )
	  subdist2[i][j]=NULL;
      }

    num_edges=0;

    /* Allocating the array of marginal probabilities for each dimension */
    ind_prob=new double*[dim];
    for( i=0; i<dim; i++ )
      ind_prob[i]=new double[num_states];

    for( i=0; i<dim; i++ )
      for( i1=0; i1<num_states; i1++ )
	ind_prob[i][i1]=0.0;

    /* Allocating the arrays of edges */
    /* There can be no more edges than the number of nodes-1 */
    edge=new long*[dim-1];
    for( i=0; i<dim-1; i++ )
      /* Each edge connects of a pair of nodes */
      edge[i]=new long[2];

    edge_prob=new double**[dim-1];
    for( i=0; i<dim-1; i++ )
      {
	/* Probability table for each edge consists of a 
	   num_states x num_states matrix */
	edge_prob[i]=new double*[num_states];
	for( i1=0; i1<num_states; i1++ )
	  edge_prob[i][i1]=new double[num_states];
      }

    /* Allocating array of flags for the nodes */
    node_used=new long[dim];

    /* Allocating the arrays of ranks and MIs for edges */
    edge_rank=new long[dim-1];
    edge_MI=new double[dim-1];

    for( i=0; i<dim-1; i++ )
      {
	edge_rank[i]=0;
	edge_MI[i]=0.0;
      }

    /* Initializing the envelopes */
    num_envelopes=0;
    envelope=new Envelope*[dim];

    break;
  case DIST_BERNOULLI:
    dim_index=new long[dim];
    for( i=0; i<dim; i++ )
      dim_index[i]=i;
    state_prob=new double[num_states];

    /* Initializing */
    for( i=0; i<num_states; i++ )
      state_prob[i]=0.0;
    
    /* Posterior univariate probabilities (for updates) */
    uni_prob=new double***[1];
    uni_prob[0]=new double**[num_states];
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    /* First statesequence entry probabilities */
    dim_index=new long[dim];
    for( i=0; i<dim; i++ )
      dim_index[i]=i;
    state_prob=new double[num_states];

    /* Conditional probabilities */
    cond_state_prob=new double*[num_states];
    for( i=0; i<num_states; i++ )
      cond_state_prob[i]=new double[num_states];

    /* Initializing */
    for( i=0; i<num_states; i++ )
      state_prob[i]=0.0;

    for( i=0; i<num_states; i++ )
      for( i1=0; i1<num_states; i1++ )
	cond_state_prob[i][i1]=0.0;

    /* Posterior univariate probabilities (for updates) */
    uni_prob=new double***[1];
    uni_prob[0]=new double**[num_states];

    /* Joint probabilities (for updates) */
    joint_prob=new double***[num_states];
    for( i=0; i<num_states; i++ )
      joint_prob[i]=new double**[num_states];

    /* Scaled forward pass updates */
    fwd_update=new double**[num_states];

    /* Scaled backward pass updates */
    bkd_update=new double**[num_states];

    break;
  case DIST_UNICONDMVME:
    /* Multinomial (polytomous) MaxEnt (logistic) distribution */
    
    /* num_states indicates the number of possible values */
    /* dim indicates the number of parameters */

    /* No need for dim_index -- the value in question is always indexed by 0 */

    /* Allocating the array of parameters */
    lambda=new double[MAXNUMMAXENTFEATURES];

    /* Initializing the number of features for each function */
    num_features=new long[MAXNUMMAXENTFEATURES];

    /* Initializing the indices for the functions */
    feature_index=new long*[MAXNUMMAXENTFEATURES];

    /* Initializing the values for the functions */
    feature_value=new long*[MAXNUMMAXENTFEATURES];

    break;
  case DIST_CHOWLIU:
    /* Allocating structures for tree-structured distribution */
    dim_index=new long[dim];
    for( i=0; i<dim; i++ )
      dim_index[i]=i;
    num_edges=0;

    /* Allocating the array of marginal probabilities for each dimension */
    ind_prob=new double*[dim];
    for( i=0; i<dim; i++ )
      ind_prob[i]=new double[num_states];

    for( i=0; i<dim; i++ )
      for( i1=0; i1<num_states; i1++ )
	ind_prob[i][i1]=0.0;

    /* Allocating the arrays of edges */
    /* There can be no more edges than the number of nodes-1 */
    edge=new long*[dim-1];
    for( i=0; i<dim-1; i++ )
      /* Each edge connects of a pair of nodes */
      edge[i]=new long[2];

    edge_prob=new double**[dim-1];
    for( i=0; i<dim-1; i++ )
      {
	/* Probability table for each edge consists of a 
	   num_states x num_states matrix */
	edge_prob[i]=new double*[num_states];
	for( i1=0; i1<num_states; i1++ )
	  edge_prob[i][i1]=new double[num_states];
      }

    /* Allocating array of flags for the nodes */
    node_used=new long[dim];

    /* Allocating the arrays of ranks and MIs for edges */
    edge_rank=new long[dim-1];
    edge_MI=new double[dim-1];

    for( i=0; i<dim-1; i++ )
      {
	edge_rank[i]=0;
	edge_MI[i]=0.0;
      }

    break;
  case DIST_CONDCHOWLIU:
    dim_index=new long[dim];
    for( i=0; i<dim; i++ )
      dim_index[i]=i;
    num_edges=0;

    /* Allocating the array of marginal probabilities for each dimension (previous + current) */
    ind_prob=new double*[2*dim];
    for( i=0; i<2*dim; i++ )
      ind_prob[i]=new double[num_states];

    /* Initializing the probabilities */
    for( i=0; i<2*dim; i++ )
      for( i1=0; i1<num_states; i1++ )
	ind_prob[i][i1]=0.0;

    /* Allocating the arrays of edges */
    /* There can be no more edges than the number of nodes in the predictive space */
    edge=new long*[dim];
    for( i=0; i<dim; i++ )
      /* Each edge connects of a pair of nodes */
      edge[i]=new long[2];

    edge_prob=new double**[dim];
    for( i=0; i<dim; i++ )
      {
	/* Probability table for each edge consists of a 
	   num_states x num_states matrix */
	edge_prob[i]=new double*[num_states];
	for( i1=0; i1<num_states; i1++ )
	  edge_prob[i][i1]=new double[num_states];
      }

    /* Allocating array of flags for the nodes */
    node_used=new long[2*dim];

    /* Allocating the arrays of ranks and MIs for edges */
    edge_rank=new long[dim];
    edge_MI=new double[dim];

    for( i=0; i<dim; i++ )
      {
	edge_rank[i]=0;
	edge_MI[i]=0.0;
      }

    break;   
  case DIST_ME_BIVAR:
    dim_index=new long[dim];
    for( i=0; i<dim; i++ )
      dim_index[i]=i;
    sigma=new double*[dim];
    for( i=0; i<dim; i++ )
      sigma[i]=new double[i+1];
    for( i=0; i<dim; i++ )
      for( i1=0; i1<i+1; i1++ )
	sigma[i][i1]=0.0;
    det=(double)dim*CONST_LN2;
    is_sim_initialized=0;
    last_sim=NULL;
    break;
  case DIST_BN_ME_BIVAR:
    dim_index=new long[dim];
    for( i=0; i<dim; i++ )
      dim_index[i]=i;
    num_features=new long[dim];
    sigma=new double*[dim];
    for( i=0; i<dim; i++ )
      sigma[i]=new double[dim];    
    feature_index=new long*[dim];
    for( i=0; i<dim; i++ )
      feature_index[i]=new long[dim];
    for( i=0; i<dim; i++ )
      {
	num_features[i]=1;
	feature_index[i][0]=i;
	sigma[i][0]=0.0;
	for( i1=1; i1<dim; i1++ )
	  {
	    feature_index[i][i1]=0;
	    sigma[i][i1]=0.0;
	  }
      }
    sim_order=new long[dim];
    for( i=0; i<dim; i++ )
      sim_order[i]=i;

    break;
  case DIST_BN_CME_BIVAR:
    dim_index=new long[dim];
    for( i=0; i<dim; i++ )
      dim_index[i]=i;
    num_features=new long[dim];
    sigma=new double*[dim];
    for( i=0; i<dim; i++ )
      sigma[i]=new double[2*dim];    
    feature_index=new long*[dim];
    for( i=0; i<dim; i++ )
      feature_index[i]=new long[2*dim];
    for( i=0; i<dim; i++ )
      {
	num_features[i]=1;
	feature_index[i][0]=i;
	sigma[i][0]=0.0;
	for( i1=1; i1<2*dim; i1++ )
	  {
	    feature_index[i][i1]=0;
	    sigma[i][i1]=0.0;
	  }
      }
    sim_order=new long[dim];
    for( i=0; i<dim; i++ )
      sim_order[i]=i;

    /* First state probabilities */
    state_prob=new double[dim];
    for( i=0; i<dim; i++ )
      state_prob[i]=0.0;

    break;
  case DIST_DELTAEXP:
    dim_index=new long[dim];
    for( i=0; i<dim; i++ )
      dim_index[i]=i;
    mix_prob=new double[num_states];
    for( i=0; i<num_states; i++ )
      mix_prob[i]=0.0;

    exp_param=new double[num_states-1];
    for( i=0; i<num_states-1; i++ )
      exp_param[i]=0.0;
    break;
  case DIST_DELTAGAMMA:
    dim_index=new long[dim];
    for( i=0; i<dim; i++ )
      dim_index[i]=i;
    mix_prob=new double[num_states];
    for( i=0; i<num_states; i++ )
      mix_prob[i]=0.0;

    gamma_param1=new double[num_states-1];
    gamma_param2=new double[num_states-1];
    for( i=0; i<num_states-1; i++ )
      {
	gamma_param1[i]=0.0;
	gamma_param2[i]=0.0;
      }
    break;
  case DIST_DIRACDELTA:
    /* Dirac delta */
    dim_index=new long[1];
    dim_index[0]=0;

    /* By default, the value is assumed to be 0 */
    delta_value=0.0;
    break;
  case DIST_EXP:
    /* Geometric */
    dim_index=new long[1];
    dim_index[0]=0;
    exp_param1=0.0;
    break;
  case DIST_GAMMA:
    /* Gamma distribution */
    dim_index=new long[1];
    dim_index[0]=0;
    gamma1=0.0;
    gamma2=0.0;
    break;
  case DIST_LOGNORMAL:
    /* Log-normal distribution */
    dim_index=new long[1];
    dim_index[0]=0;
    log_normal1=0.0;
    log_normal2=0.0;
    break;
  case DIST_NORMAL:
    dim_index=new long[dim];
    for( i=0; i<dim; i++ )
      dim_index[i]=i;
    mu=new double[dim];
    sigma=new double*[dim];
    for( i=0; i<dim; i++ )
      sigma[i]=new double[dim];
    inv_sigma=new double*[dim];
    for( i=0; i<dim; i++ )
      inv_sigma[i]=new double[dim];
    chol_sigma=NULL;
      
    /* Initializing */
    for( i=0; i<dim; i++ )
      mu[i]=0.0;
    for( i=0; i<dim; i++ )
      for( i1=0; i1<dim; i1++ )
	sigma[i][i1]=0.0;
    for( i=0; i<dim; i++ )
      for( i1=0; i1<dim; i1++ )
	inv_sigma[i][i1]=0.0;
    det=0.0;
    break;
  case DIST_NORMALCHAIN:
    dim_index=new long[dim];
    for( i=0; i<dim; i++ )
      dim_index[i]=i;
    first_mu=new double[dim];    
    first_sigma=new double*[dim];
    for( i=0; i<dim; i++ )
      first_sigma[i]=new double[dim];
    inv_first_sigma=new double*[dim];
    for( i=0; i<dim; i++ )
      inv_first_sigma[i]=new double[dim];
    chol_first_sigma=NULL;

    mu=new double[dim];
    sigma=new double*[dim];
    for( i=0; i<dim; i++ )
      sigma[i]=new double[dim];
    inv_sigma=new double*[dim];
    for( i=0; i<dim; i++ )
      inv_sigma[i]=new double[dim];      
    chol_sigma=NULL;
    W=new double*[dim];
    for( i=0; i<dim; i++ )
      W[i]=new double[dim];

    /* Initializing */
    for( i=0; i<dim; i++ )
      first_mu[i]=0.0;
    for( i=0; i<dim; i++ )
      for( i1=0; i1<dim; i1++ )
	first_sigma[i][i1]=0.0;
    for( i=0; i<dim; i++ )
      for( i1=0; i1<dim; i1++ )
	inv_first_sigma[i][i1]=0.0;
    first_det=0.0;

    for( i=0; i<dim; i++ )
      mu[i]=0.0;
    for( i=0; i<dim; i++ )
      for( i1=0; i1<dim; i1++ )
	sigma[i][i1]=0.0;
    for( i=0; i<dim; i++ )
      for( i1=0; i1<dim; i1++ )
	inv_sigma[i][i1]=0.0;
    for( i=0; i<dim; i++ )
      for( i1=0; i1<dim; i1++ )
	W[i][i1]=0.0;
    det=0.0;
    break;
  case DIST_NORMALCL:
    dim_index=new long[dim];
    for( i=0; i<dim; i++ )
      dim_index[i]=i;
    mu=new double[dim];
    sigma=new double*[dim];
    for( i=0; i<dim; i++ )
      sigma[i]=new double[dim];
    inv_sigma=new double*[dim];
    for( i=0; i<dim; i++ )
      inv_sigma[i]=new double[dim];
    chol_sigma=NULL;
      
    /* Initializing */
    for( i=0; i<dim; i++ )
      mu[i]=0.0;
    for( i=0; i<dim; i++ )
      for( i1=0; i1<dim; i1++ )
	sigma[i][i1]=0.0;
    for( i=0; i<dim; i++ )
      for( i1=0; i1<dim; i1++ )
	inv_sigma[i][i1]=0.0;
    det=0.0;

    num_edges=0;
    
    /* Allocating the arrays of edges */
    /* There can be no more edges than the number of nodes-1 */
    edge=new long*[dim-1];
    for( i=0; i<dim-1; i++ )
      /* Each edge connects of a pair of nodes */
      edge[i]=new long[2];

    /* Allocating array of flags for the nodes */
    node_used=new long[dim];

    /* Allocating the arrays of ranks and MIs for edges */
    edge_rank=new long[dim-1];
    edge_MI=new double[dim-1];

    for( i=0; i<dim-1; i++ )
      {
	edge_rank[i]=0;
	edge_MI[i]=0.0;
      }

    break;
  case DIST_LOGISTIC:
    lambda=new double[num_states];
    rho=new double*[num_states];
    for( i=0; i<num_states; i++ )
      rho[i]=new double[dim];

    /* Initializing */
    for( i=0; i<num_states; i++ )
      lambda[i]=0.0;
    for( i=0; i<num_states; i++ )
      for( i1=0; i1<dim; i1++ )
	rho[i][i1]=0.0;

    /* Posterior univariate probabilities (for updates) */
    uni_prob=new double***[1];
    uni_prob[0]=new double**[num_states];

    joint_prob=NULL;

    /* Arrays for storing probability values */
    p_tr=new double***[num_states];
    for( i=0; i<num_states; i++ )
      p_tr[i]=new double**[1];

    log_un_prob=new double***[num_states];
    for( i=0; i<num_states; i++ )
      log_un_prob[i]=new double**[1];

    break;
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    lambda=new double[num_states];
    sigma=new double*[num_states];
    for( i=0; i<num_states; i++ )
      sigma[i]=new double[num_states];
    rho=new double*[num_states];
    for( i=0; i<num_states; i++ )
      rho[i]=new double[dim];

    /* Initializing */
    for( i=0; i<num_states; i++ )
      lambda[i]=0.0;
    for( i=0; i<num_states; i++ )
      for( i1=0; i1<num_states; i1++ )
	sigma[i][i1]=0.0;
    for( i=0; i<num_states; i++ )
      for( i1=0; i1<dim; i1++ )
	rho[i][i1]=0.0;

    /* Posterior univariate probabilities (for updates) */
    uni_prob=new double***[1];
    uni_prob[0]=new double**[num_states];

    /* Joint probabilities (for updates) */
    joint_prob=new double***[num_states];
    for( i=0; i<num_states; i++ )
      joint_prob[i]=new double**[num_states];

    /* Scaled forward pass updates */
    fwd_update=new double**[num_states];

    /* Scaled backward pass updates */
    bkd_update=new double**[num_states];

    /* Arrays for storing probability values */
    p_tr=new double***[num_states];
    for( i=0; i<num_states; i++ )
      p_tr[i]=new double**[num_states];

    log_un_prob=new double***[num_states];
    for( i=0; i<num_states; i++ )
      log_un_prob[i]=new double**[num_states];

    break;
  default:;
  }
}

Distribution::~Distribution() {
  /* Index variable(s) */
  long i,j,i1;

  /* Deallocating parameter structures */
  switch( type ) {
  case DIST_FACTOR:
    /* Removing sub-component distributions and deallocating the structure */
    for( i=0; i<dim; i++ )
      if( subdist[i] )
	delete subdist[i];
    delete subdist;
    break;
  case DIST_MIXTURE:
    for( i=0; i<num_states; i++ )
      if( subdist[i] )
	delete subdist[i];
    delete subdist;
    delete state_prob;
    break;
  case DIST_CLMIXTURE:
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<num_states; j++ )
	  if( subdist2[i][j] )
	    delete subdist2[i][j];
	delete subdist2[i];
      }
    delete subdist2;

    /* Deallocating the array of marginal probabilities */
    for( i=0; i<dim; i++ )
      delete ind_prob[i];
    delete ind_prob;

    /* Deallocating the edges */
    for( i=0; i<dim-1; i++ )
      delete edge[i];
    delete edge;

    for( i=0; i<dim-1; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  delete edge_prob[i][i1];
	delete edge_prob[i];
      }
    delete edge_prob;

    /* Deallocating the array of flags */
    delete node_used;

    /* Deallocating the arrays of edge ranks and strengths */
    delete edge_rank;
    delete edge_MI;    

    /* Deallocating the envelopes */
    for( i=0; i<num_envelopes; i++ )
      delete envelope[i];
    delete envelope;

    break;
  case DIST_BERNOULLI:
    delete dim_index;
    delete state_prob;

    /* Deallocating the array of univariate posterior probabilities */
    delete uni_prob[0];
    delete uni_prob;
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    /* Deallocating the array of first sequence entry probabilities */
    delete dim_index;
    delete state_prob;

    /* Deallocating the array of conditional probabilities */
    for( i=0; i<num_states; i++ )
      delete cond_state_prob[i];
    delete cond_state_prob;

    /* Deallocating the array of univariate posterior probabilities */
    delete uni_prob[0];
    delete uni_prob;

    /* Deallocating the array of joint probabilities */
    for( i=0; i<num_states; i++ )
      delete joint_prob[i];
    delete joint_prob;

    /* Deallocating scaled forward and backward pass updates */
    delete fwd_update;
    delete bkd_update;

    break;
  case DIST_UNICONDMVME:
    /* Deallocating the array of parameters */
    delete lambda;

    /* Deallocating the indices for the functions */
    for( i=0; i<dim; i++ )
      delete feature_index[i];
    delete feature_index;

    /* Deallocating the values for the functions */
    for( i=0; i<dim; i++ )
      delete feature_value[i];
    delete feature_value;

    /* Deallocating the number of features for each function */
    delete num_features;

    break;
  case DIST_CHOWLIU:
    delete dim_index;
    /* Deallocating the array of marginal probabilities */
    for( i=0; i<dim; i++ )
      delete ind_prob[i];
    delete ind_prob;

    /* Deallocating the edges */
    for( i=0; i<dim-1; i++ )
      delete edge[i];
    delete edge;

    for( i=0; i<dim-1; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  delete edge_prob[i][i1];
	delete edge_prob[i];
      }
    delete edge_prob;

    /* Deallocating the array of flags */
    delete node_used;

    /* Deallocating the arrays of edge ranks and strengths */
    delete edge_rank;
    delete edge_MI;

    break;
  case DIST_CONDCHOWLIU:
    delete dim_index;
    /* Deallocating the array of marginal probabilities */
    for( i=0; i<2*dim; i++ )
      delete ind_prob[i];
    delete ind_prob;

    /* Deallocating the edges */
    for( i=0; i<dim; i++ )
      delete edge[i];
    delete edge;

    for( i=0; i<dim; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  delete edge_prob[i][i1];
	delete edge_prob[i];
      }
    delete edge_prob;

    /* Deallocating the array of flags */
    delete node_used;

    /* Deallocating the arrays of edge ranks and strengths */
    delete edge_rank;
    delete edge_MI;

    break;
  case DIST_ME_BIVAR:
    delete dim_index;
    for( i=0; i<dim; i++ )
      delete sigma[i];
    delete sigma;
    if( last_sim )
      delete last_sim;
    break;
  case DIST_BN_ME_BIVAR:
    delete dim_index;
    delete num_features;
    for( i=0; i<dim; i++ )
      delete sigma[i];
    delete sigma;
    for( i=0; i<dim; i++ )
      delete feature_index[i];
    delete feature_index;
    delete sim_order;
    break;
  case DIST_BN_CME_BIVAR:
    delete dim_index;
    delete num_features;
    for( i=0; i<dim; i++ )
      delete sigma[i];
    delete sigma;
    for( i=0; i<dim; i++ )
      delete feature_index[i];
    delete feature_index;
    delete sim_order;
    delete state_prob;
    break;
  case DIST_DELTAEXP:
    delete dim_index;
    delete mix_prob;
    delete exp_param;
    break;
  case DIST_DELTAGAMMA:
    delete dim_index;
    delete mix_prob;
    delete gamma_param1;
    delete gamma_param2;
    break;
  case DIST_DIRACDELTA:
  case DIST_EXP:
  case DIST_GAMMA:
  case DIST_LOGNORMAL:
    delete dim_index;
    break;
  case DIST_NORMAL:
    delete dim_index;
    delete mu;
    for( i=0; i<dim; i++ )
      delete sigma[i];
    delete sigma;
    for( i=0; i<dim; i++ )
      delete inv_sigma[i];
    delete inv_sigma;
    if( chol_sigma )
      {
	for( i=0; i<dim; i++ )
	  delete chol_sigma[i];
	delete chol_sigma;
      }
    break;
  case DIST_NORMALCHAIN:
    delete dim_index;
    delete first_mu;
    for( i=0; i<dim; i++ )
      delete first_sigma[i];
    delete first_sigma;
    for( i=0; i<dim; i++ )
      delete inv_first_sigma[i];
    delete inv_first_sigma;
    if( chol_first_sigma )
      {
	for( i=0; i<dim; i++ )
	  delete chol_first_sigma[i];
	delete chol_first_sigma;
      }

    delete mu;
    for( i=0; i<dim; i++ )
      delete sigma[i];
    delete sigma;
    for( i=0; i<dim; i++ )
      delete inv_sigma[i];
    delete inv_sigma;
    if( chol_sigma )
      {
	for( i=0; i<dim; i++ )
	  delete chol_sigma[i];
	delete chol_sigma;
      }

    for( i=0; i<dim; i++ )
      delete W[i];
    delete W;
    break;
  case DIST_NORMALCL:
    delete dim_index;
    delete mu;
    for( i=0; i<dim; i++ )
      delete sigma[i];
    delete sigma;
    for( i=0; i<dim; i++ )
      delete inv_sigma[i];
    delete inv_sigma;
    if( chol_sigma )
      {
	for( i=0; i<dim; i++ )
	  delete chol_sigma[i];
	delete chol_sigma;
      }
 
    for( i=0; i<dim-1; i++ )
      delete edge[i];
    delete edge;

    delete node_used;

    delete edge_rank;
    delete edge_MI;

    break;
  case DIST_LOGISTIC:
    delete lambda;
    for( i=0; i<num_states; i++ )
      delete rho[i];
    delete rho;

    /* Deallocating the array of univariate posterior probabilities */
    delete uni_prob[0];
    delete uni_prob;

    /* Deallocating arrays of probability values */
    for( i=0; i<num_states; i++ )
      delete p_tr[i];
    delete p_tr;

    for( i=0; i<num_states; i++ )
      delete log_un_prob[i];
    delete log_un_prob;

    break;
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    delete lambda;
    for( i=0; i<num_states; i++ )
      delete sigma[i];
    delete sigma;
    for( i=0; i<num_states; i++ )
      delete rho[i];
    delete rho;

    /* Deallocating the array of joint probabilities */
    for( i=0; i<num_states; i++ )
      delete joint_prob[i];
    delete joint_prob;

    /* Deallocating scaled forward and backward pass updates */
    delete fwd_update;
    delete bkd_update;

    /* Deallocating the array of univariate posterior probabilities */
    delete uni_prob[0];
    delete uni_prob;

    /* Deallocating arrays of probability values */
    for( i=0; i<num_states; i++ )
      delete p_tr[i];
    delete p_tr;

    for( i=0; i<num_states; i++ )
      delete log_un_prob[i];
    delete log_un_prob;

    break;
  default:;
  }
}

Distribution* Distribution::copy() {
  /* Copy constructor */

  Distribution* dist;

  /* Index variable(s) */
  long i,j,i1,i2;

  dist=new Distribution( type, num_states, dim );

  switch( type ) {
  case DIST_FACTOR:
    for( i=0; i<dim; i++ )
      dist->subdist[i]=subdist[i]->copy();
    break;
  case DIST_MIXTURE:
    for( i=0; i<num_states; i++ )
      dist->state_prob[i]=state_prob[i];
    for( i=0; i<num_states; i++ )
      dist->subdist[i]=subdist[i]->copy();
    break;
  case DIST_CLMIXTURE:
    for( i=0; i<dim; i++ )
      for( j=0; j<num_states; j++ )
	dist->subdist2[i][j]=subdist2[i][j]->copy();

    /* Copying marginal probabilities */
    for( i=0; i<dim; i++ )
      for( i1=0; i1<num_states; i1++ )
	dist->ind_prob[i][i1]=ind_prob[i][i1];

    /* Copying edge information */
    dist->num_edges=num_edges;
    
    for( i=0; i<num_edges; i++ )
      {
	dist->edge[i][0]=edge[i][0];
	dist->edge[i][1]=edge[i][1];
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    dist->edge_prob[i][i1][i2]=edge_prob[i][i1][i2];
      }

    for( i=0; i<num_edges; i++ )
      {
	dist->edge_rank[i]=edge_rank[i];
	dist->edge_MI[i]=edge_MI[i];
      }
    
    /* Computing the envelopes */
    dist->num_envelopes=dist->compute_envelopes_full();

    dist->mdl_beta=mdl_beta;
    break;
  case DIST_BERNOULLI:
    for( i=0; i<dim; i++ )
      dist->dim_index[i]=dim_index[i];
    for( i=0; i<num_states; i++ )
      dist->state_prob[i]=state_prob[i];
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    for( i=0; i<dim; i++ )
      dist->dim_index[i]=dim_index[i];

    /* First sequence entry probabilities */
    for( i=0; i<num_states; i++ )
      dist->state_prob[i]=state_prob[i];

    /* Conditional probabilities */
    for( i=0; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	dist->cond_state_prob[i][j]=cond_state_prob[i][j];
    break;
  case DIST_UNICONDMVME:
    for( i=0; i<dim; i++ )
      dist->lambda[i]=lambda[i];

    for( i=0; i<dim; i++ )
      dist->num_features[i]=num_features[i];

    for( i=0; i<dim; i++ )
      {
	feature_index[i]=new long[num_features[i]];
	feature_value[i]=new long[num_features[i]];
	for( i1=0; i1<num_features[i]; i++ )
	  {	  
	    dist->feature_index[i][i1]=feature_index[i][i1];
	    dist->feature_value[i][i1]=feature_value[i][i1];
	  }
      }
    break;
  case DIST_CHOWLIU:
    for( i=0; i<dim; i++ )
      dist->dim_index[i]=dim_index[i];
    /* Copying marginal probabilities */
    for( i=0; i<dim; i++ )
      for( i1=0; i1<num_states; i1++ )
	dist->ind_prob[i][i1]=ind_prob[i][i1];

    /* Copying edge information */
    dist->num_edges=num_edges;
    
    for( i=0; i<num_edges; i++ )
      {
	dist->edge[i][0]=edge[i][0];
	dist->edge[i][1]=edge[i][1];
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    dist->edge_prob[i][i1][i2]=edge_prob[i][i1][i2];
      }

    for( i=0; i<num_edges; i++ )
      {
	dist->edge_rank[i]=edge_rank[i];
	dist->edge_MI[i]=edge_MI[i];
      }
    
    dist->mdl_beta=mdl_beta;
    break;
  case DIST_CONDCHOWLIU:
    for( i=0; i<dim; i++ )
      dist->dim_index[i]=dim_index[i];
    /* Copying marginal probabilities */
    for( i=0; i<dim; i++ )
      for( i1=0; i1<num_states; i1++ )
	dist->ind_prob[i][i1]=ind_prob[i][i1];

    /* Copying edge information */
    dist->num_edges=num_edges;
    
    for( i=0; i<num_edges; i++ )
      {
	dist->edge[i][0]=edge[i][0];
	dist->edge[i][1]=edge[i][1];
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    dist->edge_prob[i][i1][i2]=edge_prob[i][i1][i2];
      }

    for( i=0; i<num_edges; i++ )
      {
	dist->edge_rank[i]=edge_rank[i];
	dist->edge_MI[i]=edge_MI[i];
      }

    dist->mdl_beta=mdl_beta;
    break;
  case DIST_ME_BIVAR:
    for( i=0; i<dim; i++ )
      dist->dim_index[i]=dim_index[i];
    for( i=0; i<dim; i++ )
      for( j=0; j<i+1; j++ )
	dist->sigma[i][j]=sigma[i][j];
    dist->det=det;
    break;
  case DIST_BN_ME_BIVAR:
    for( i=0; i<dim; i++ )
      dist->dim_index[i]=dim_index[i];
    for( i=0; i<dim; i++ )
      dist->num_features[i]=num_features[i];
    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	{
	  dist->sigma[i][j]=sigma[i][j];
	  dist->feature_index[i][j]=feature_index[i][j];
	}
    for( i=0; i<dim; i++ )
      dist->sim_order[i]=sim_order[i];

    dist->mdl_beta=mdl_beta;
    break;
  case DIST_BN_CME_BIVAR:
    for( i=0; i<dim; i++ )
      dist->dim_index[i]=dim_index[i];
    for( i=0; i<dim; i++ )
      dist->num_features[i]=num_features[i];
    for( i=0; i<dim; i++ )
      for( j=0; j<num_features[i]; j++ )
	{
	  dist->sigma[i][j]=sigma[i][j];
	  dist->feature_index[i][j]=feature_index[i][j];
	}
    for( i=0; i<dim; i++ )
      dist->sim_order[i]=sim_order[i];
    for( i=0; i<dim; i++ )
      dist->state_prob[i]=state_prob[i];

    dist->mdl_beta=mdl_beta;
    break;
  case DIST_DELTAEXP:
    for( i=0; i<dim; i++ )
      dist->dim_index[i]=dim_index[i];
    /* Copying mixing probabilities */
    for( i=0; i<num_states; i++ )
      dist->mix_prob[i]=mix_prob[i];
    for( i=0; i<num_states-1; i++ )
      dist->exp_param[i]=exp_param[i];
    break;
  case DIST_DELTAGAMMA:
    for( i=0; i<dim; i++ )
      dist->dim_index[i]=dim_index[i];
    /* Copying mixing probabilities */
    for( i=0; i<num_states; i++ )
      dist->mix_prob[i]=mix_prob[i];
    for( i=0; i<num_states-1; i++ )
      {
	dist->gamma_param1[i]=gamma_param1[i];
	dist->gamma_param2[i]=gamma_param2[i];
      }
    break;
  case DIST_DIRACDELTA:
    dist->dim_index[0]=dim_index[0];
    break;
  case DIST_EXP:
    dist->dim_index[0]=dim_index[0];
    dist->exp_param1=exp_param1;
    break;
  case DIST_GAMMA:
    dist->dim_index[0]=dim_index[0];
    dist->gamma1=gamma1;
    dist->gamma2=gamma2;
    break;
  case DIST_LOGNORMAL:
    dist->dim_index[0]=dim_index[0];
    dist->log_normal1=log_normal1;
    dist->log_normal2=log_normal2;
   break;
  case DIST_NORMAL:
    for( i=0; i<dim; i++ )
      dist->dim_index[i]=dim_index[i];
    for( i=0; i<dim; i++ )
      dist->mu[i]=mu[i];
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	dist->sigma[i][j]=sigma[i][j];
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	dist->inv_sigma[i][j]=inv_sigma[i][j];
    if( chol_sigma )
      {
	dist->chol_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  {
	    dist->chol_sigma[i]=new double[dim];
	    for( j=0; j<dim; j++ )
	      dist->chol_sigma[i][j]=chol_sigma[i][j];
	  }
      }
    dist->det=det;
    break;
  case DIST_NORMALCHAIN:
    for( i=0; i<dim; i++ )
      dist->dim_index[i]=dim_index[i];
    for( i=0; i<dim; i++ )
      dist->first_mu[i]=first_mu[i];
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	dist->first_sigma[i][j]=first_sigma[i][j];
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	dist->inv_first_sigma[i][j]=inv_first_sigma[i][j];
    if( chol_first_sigma )
      {
	dist->chol_first_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  {
	    dist->chol_first_sigma[i]=new double[dim];
	    for( j=0; j<dim; j++ )
	      dist->chol_first_sigma[i][j]=chol_first_sigma[i][j];
	  }
      }

    for( i=0; i<dim; i++ )
      dist->mu[i]=mu[i];
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	dist->sigma[i][j]=sigma[i][j];
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	dist->inv_sigma[i][j]=inv_sigma[i][j];
    if( chol_sigma )
      {
	dist->chol_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  {
	    dist->chol_sigma[i]=new double[dim];
	    for( j=0; j<dim; j++ )
	      dist->chol_sigma[i][j]=chol_sigma[i][j];
	  }
      }
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	dist->W[i][j]=W[i][j];
    dist->det=det;
    break;
  case DIST_NORMALCL:
    for( i=0; i<dim; i++ )
      dist->dim_index[i]=dim_index[i];
    for( i=0; i<dim; i++ )
      dist->mu[i]=mu[i];
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	dist->sigma[i][j]=sigma[i][j];
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	dist->inv_sigma[i][j]=inv_sigma[i][j];
    if( chol_sigma )
      {
	dist->chol_sigma=new double*[dim];
	for( i=0; i<dim; i++ )
	  {
	    dist->chol_sigma[i]=new double[dim];
	    for( j=0; j<dim; j++ )
	      dist->chol_sigma[i][j]=chol_sigma[i][j];
	  }
      }
    dist->det=det;

    /* Copying edge information */
    dist->num_edges=num_edges;
    
    for( i=0; i<num_edges; i++ )
      {
	dist->edge[i][0]=edge[i][0];
	dist->edge[i][1]=edge[i][1];
      }

    for( i=0; i<num_edges; i++ )
      {
	dist->edge_rank[i]=edge_rank[i];
	dist->edge_MI[i]=edge_MI[i];
      }
    
    break;
  case DIST_LOGISTIC:
    for( i=0; i<num_states; i++ )
      dist->lambda[i]=lambda[i];
    for( i=0; i<num_states; i++ )
      for( i1=0; i1<dim; i1++ )
	dist->rho[i][i1]=rho[i][i1];
    break;
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    for( i=0; i<num_states; i++ )
      dist->lambda[i]=lambda[i];
    for( i=0; i<num_states; i++ )
      for( j=0; j<num_states; j++ )
	dist->sigma[i][j]=sigma[i][j];
    for( i=0; i<num_states; i++ )
      for( i1=0; i1<dim; i1++ )
	dist->rho[i][i1]=rho[i][i1];
    break;
  default:;
  }
	
  return( dist );
}

void Distribution::Initialize( Data* data ) {
  long num_entries;
  double m, sd;         // Mean and standard deviation
  double lower_bound;   // Lower and upper bounds for the logistic parameters

  /* Variables used with Chow-Liu */
  double**** pair_prob;         // Projection on the pair of the variables
  double** MI;                  // Mutual information
  long** sorted_MI;             // Indices of the sorted mutual information entries
  long** connected;             // Connectivity matrix
  long max_num_edges;           // Maximum number of edges allowed for Chow-Liu distribution
  long num_selected_edges;      // Number of edges selected to find a forest from

  /* Index variable(s) */
  long i,j,i1,i2;
  long n,t;

  /* Temporary variable */
  double sum;
  long b;

  switch( type ) {
  case DIST_FACTOR:
    /* Initializing all sub-components */
    for( i=0; i<dim; i++ )
      subdist[i]->Initialize( data );
    break;
  case DIST_MIXTURE:
    /* Initializing mixing probabilities */
    sum=0.0;
    for( i=0; i<num_states; i++ )
      {
	state_prob[i]=drand48();
	sum+=state_prob[i];
      }
    for( i=0; i<num_states; i++ )
      state_prob[i]/=sum;

    /* Initializing mixing components */
    for( i=0; i<num_states; i++ )
      subdist[i]->Initialize( data );
    break;
  case DIST_CLMIXTURE:
    /* Initializing mixing probabilities as conditionally independent */
    num_edges=0;

    for( i=0; i<dim; i++ )
      {
	sum=0.0;
	for( i1=0; i1<num_states; i1++ )
	  {
	    ind_prob[i][i1]=drand48();
	    sum+=ind_prob[i][i1];
	  }
	
	for( i1=0; i1<num_states; i1++ )
	  ind_prob[i][i1]/=sum;
      }    

    /* Initializing mixing components */
    for( i=0; i<dim; i++ )
      for( j=0; j<num_states; j++ )
	subdist2[i][j]->Initialize( data );

    /* Computing the envelopes */
    for( i=0; i<num_envelopes; i++ )
      delete envelope[i];

    num_envelopes=compute_envelopes_full();

    break;
  case DIST_BERNOULLI:
    /* Initializing Bernoulli parameters */
    sum=0.0;
    for( i=0; i<num_states; i++ )
      {
	state_prob[i]=drand48();
	sum+=state_prob[i];
      }

    /* Renormalizing */
    for( i=0; i<num_states; i++ )
      state_prob[i]/=sum;
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    /* Initializing conditional probabilities */
    for( i=0; i<num_states; i++ )
      {
	sum=0.0;
	for( j=0; j<num_states; j++ )
	  {
	    cond_state_prob[i][j]=drand48();
	    sum+=cond_state_prob[i][j];
	  }
	/* Renormalizing */
	for( j=0; j<num_states; j++ )
	  cond_state_prob[i][j]/=sum;
      }
    /* Initializing first sequence entry probabilities */
    sum=0.0;
    for( i=0; i<num_states; i++ )
      {
	state_prob[i]=drand48();
	sum+=state_prob[i];
      }
    /* Renormalizing */
    for( i=0; i<num_states; i++ )
      state_prob[i]/=sum;

    break;
  case DIST_UNICONDMVME:
    /* !!! Weights are initialized arbitrarily !!! */
    for( i=0; i<dim; i++ )
      lambda[i]=drand48();
    break;
  case DIST_CHOWLIU:
    /* Initializing with conditional independence assumption */

    /* Assuming there are no edges -- need to initialize only 
       marginal probabilities */
    num_edges=0;

    for( i=0; i<dim; i++ )
      {
	sum=0.0;
	for( i1=0; i1<num_states; i1++ )
	  {
	    ind_prob[i][i1]=drand48();
	    sum+=ind_prob[i][i1];
	  }
	
	for( i1=0; i1<num_states; i1++ )
	  ind_prob[i][i1]/=sum;
      }

    break;
  case DIST_CONDCHOWLIU:
    /* !!! Initializing with conditional independence assumption !!! */
    
    /* Assuming there are no edges -- need to initialize only 
       marginal probabilities of the unobserved variables */
    num_edges=0;

    for( i=dim; i<2*dim; i++ )
      {
	sum=0.0;
	for( i1=0; i1<num_states; i1++ )
	  {
	    ind_prob[i][i1]=drand48();
	    sum+=ind_prob[i][i1];
	  }
	
	for( i1=0; i1<num_states; i1++ )
	  ind_prob[i][i1]/=sum;
      }

    break;
  case DIST_ME_BIVAR:
    /* Exponent of the normalization constant */
    det=0.0;

    /* Initializing only the diagonal */
    for( i=0; i<dim; i++ )
      {
	sigma[i][i]=generateUnitNormal();
	det+=log(exp(sigma[i][i])+1.0);
      }

    for( i=0; i<dim; i++ )
      for( j=0; j<i; j++ )
	sigma[i][j]=0.0;
    break;
  case DIST_BN_ME_BIVAR:
    /* Assuming conditional independence */
    /* Initializing the coefficients as normally distributed around zero */
    for( i=0; i<dim; i++ )
      {
	num_features[i]=1;
	feature_index[i][0]=i;
      }

    for( i=0; i<dim; i++ )
      sigma[i][0]=generateUnitNormal();
    break;
  case DIST_BN_CME_BIVAR:
    /* Assuming conditional independence */
    /* Initializing the coefficients as normally distributed around zero */
    for( i=0; i<dim; i++ )
      {
	num_features[i]=1;
	feature_index[i][0]=i;
      }

    for( i=0; i<dim; i++ )
      sigma[i][0]=generateUnitNormal();

    /* First entry probabilities */
    for( i=0; i<dim; i++ )
      state_prob[i]=drand48();
    break;
  case DIST_DELTAEXP:
    /* f(x)=l*exp(-l*x) */
    /* Initializing mixing probabilities */
    sum=0.0;
    for( i=0; i<num_states; i++ )
      {
	mix_prob[i]=drand48();
	sum+=mix_prob[i];
      }
    for( i=0; i<num_states; i++ )
      mix_prob[i]/=sum;

    /* !!! Choosing the parameter between 0.01 and 1 !!! */
    for( i=0; i<num_states-1; i++ )
      exp_param[i]=1.0/(1.0+99.0*drand48());
    break;
  case DIST_DELTAGAMMA:
    /* f(x)=x^(param1-1)*exp(-x*param2)*param2^param1/Gamma(param1) */
    /* Initializing mixing probabilities */
    sum=0.0;
    for( i=0; i<num_states; i++ )
      {
	mix_prob[i]=drand48();
	sum+=mix_prob[i];
      }
    for( i=0; i<num_states; i++ )
      mix_prob[i]/=sum;

    /* Choosing first parameter between 0.1 and 1 and second between 0.01 and 1 */
    for( i=0; i<num_states-1; i++ )
      {
	gamma_param1[i]=1.0/(1.0+9.0*drand48());
	gamma_param2[i]=1.0/(1.0+99.0*drand48());
      }
    break;
  case DIST_DIRACDELTA:
    break;
  case DIST_EXP:
    /* f(x)=l*exp(-l*x) */
    /* !!! Choosing the parameter between 0.01 and 1 !!! */
    exp_param1=1.0/(1.0+99.0*drand48());
    break;
  case DIST_GAMMA:
    /* !!! Choosing first parameter between 0.1 and 1 and second between 0.01 and 1 !!! */
    gamma1=1.0/(1.0+9.0*drand48());
    gamma2=1.0/(1.0+99.0*drand48());
    break;
  case DIST_LOGNORMAL:
    /* !!! Choosing parameters between 0.1 and 1 !!! */
    log_normal1=1.0/(1.0+9.0*drand48());
    log_normal2=1.0/(1.0+9.0*drand48());
    break;
  case DIST_NORMAL:
    /* Tricky */

    /* Taking the covariance of the data as the covariance matrix */
    if( data )
      {
	/* !!! Better data retrieval needed !!! */

	num_entries=0;

	/* Displaying data */

	/* Zeroing out the mean */
	for( i=0; i<dim; i++ )
	  mu[i]=0.0;

	/* Zeroing out the covariance matrix */
	for( i=0; i<dim; i++ )
	  for( j=0; j<dim; j++ )
	    sigma[i][j]=0.0;

	/* Looping over all input entries while calculating the mean */
	for( n=0; n<data->num_seqs; n++ )
	  {
	    for( t=0; t<data->sequence[n]->seq_length; t++ )
	      for( i=0; i<dim; i++ )
		mu[i]+=data->sequence[n]->entry[t]->rdata[dim_index[i]];

	    num_entries+=data->sequence[n]->seq_length;
	  }

	for( i=0; i<dim; i++ )
	  mu[i]/=(double)num_entries;

	/* Calculating covariance */
	for( i=0; i<dim; i++ )
	  for( j=0; j<=i; j++ )
	    for( n=0; n<data->num_seqs; n++ )
	      for( t=0; t<data->sequence[n]->seq_length; t++ )
		{
		  sigma[i][j]+=((data->sequence[n]->entry[t]->rdata[dim_index[i]]-mu[i])*
				(data->sequence[n]->entry[t]->rdata[dim_index[j]]-mu[j]));
		  sigma[j][i]=sigma[i][j];
		}

	/* Unbiased estimate of the covariance */
	for( i=0; i<dim; i++ )
	  for( j=0; j<dim; j++ )
	    sigma[i][j]/=(double)(num_entries-1);
      }
    else
      { /* !!! Bad initialization !!! */
	for( i=0; i<dim; i++ )
	  for( j=0; j<=i; j++ )
	    {
	      sigma[i][j]=drand48()-0.5;
	      sigma[j][i]=sigma[i][j];
	    }
      } /* !!! Bad initialization !!! */
    
    /* !!! Has to be changed !!! */
    for( i=0; i<dim; i++ )
      mu[i]=drand48()-0.5;

    /* Finding inverse and determinant */
    det=find_det( sigma, dim );
    if( SINGULAR_FAIL )
      /* Data is highly correlated */
      {
	fprintf( stderr, "Data is highly correlated -- initialization failed.\n" );
	exit( -3 );
      }


    find_inv( sigma, dim, inv_sigma );
    break;
  case DIST_NORMALCHAIN:
    /* !!! For right now, covariance matrices are initialized to unit matrices, and means
       are initialized to zero vectors. !!! */

    /* First state */
    for( i=0; i<dim; i++ )
      /* !!! Uniformly distributed between -0.2 and 0.2 !!! */
      first_mu[i]=0.4*drand48()-0.2;
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	if( i==j )
	  first_sigma[i][j]=1E-00;
	else
	  first_sigma[i][j]=0.0;
    first_det=find_det( first_sigma, dim );
    find_inv( first_sigma, dim, inv_first_sigma );


    /* Transition */
    for( i=0; i<dim; i++ )
      /* !!! Uniformly distributed between -0.2 and 0.2 !!! */
#ifdef BLAH
      mu[i]=0.4*drand48()-0.2;
#endif
    /* Not going to use it */
    mu[i]=0.0;

    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	if( i==j )
	  sigma[i][j]=1E-00;
	else
	  sigma[i][j]=0.0;
    det=find_det( sigma, dim );
    find_inv( sigma, dim, inv_sigma );

    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	if( i==j )
	  W[i][j]=1.0;
	else
	  W[i][j]=0.0;

    break;
  case DIST_NORMALCL:
    /* Tricky */

    /* Taking the covariance of the data as the covariance matrix */
    if( data )
      {
	/* !!! Better data retrieval needed !!! */
	num_entries=0;

	/* Zeroing out the mean */
	for( i=0; i<dim; i++ )
	  mu[i]=0.0;

	/* Zeroing out the covariance matrix */
	for( i=0; i<dim; i++ )
	  for( j=0; j<dim; j++ )
	    sigma[i][j]=0.0;

	/* Looping over all input entries while calculating the mean */
	for( n=0; n<data->num_seqs; n++ )
	  {
	    for( t=0; t<data->sequence[n]->seq_length; t++ )
	      for( i=0; i<dim; i++ )
		mu[i]+=data->sequence[n]->entry[t]->rdata[dim_index[i]];

	    num_entries+=data->sequence[n]->seq_length;
	  }

	for( i=0; i<dim; i++ )
	  mu[i]/=(double)num_entries;

	/* Calculating covariance */
	for( i=0; i<dim; i++ )
	  for( j=0; j<=i; j++ )
	    for( n=0; n<data->num_seqs; n++ )
	      for( t=0; t<data->sequence[n]->seq_length; t++ )
		{
		  sigma[i][j]+=((data->sequence[n]->entry[t]->rdata[dim_index[i]]-mu[i])*
				(data->sequence[n]->entry[t]->rdata[dim_index[j]]-mu[j]));
		  sigma[j][i]=sigma[i][j];
		}

	/* Unbiased estimate of the covariance */
	for( i=0; i<dim; i++ )
	  for( j=0; j<dim; j++ )
	    sigma[i][j]/=(double)(num_entries-1);
      }
    else
      { /* !!! Bad initialization !!! */
	for( i=0; i<dim; i++ )
	  for( j=0; j<=i; j++ )
	    {
	      sigma[i][j]=drand48()-0.5;
	      sigma[j][i]=sigma[i][j];
	    }
      } /* !!! Bad initialization !!! */
    
    /* !!! Has to be changed !!! */
    for( i=0; i<dim; i++ )
      mu[i]=drand48()-0.5;

    /* Finding inverse and determinant */
    det=find_det( sigma, dim );
    if( SINGULAR_FAIL )
      /* Data is highly correlated */
      {
	fprintf( stderr, "Data is highly correlated -- initialization failed.\n" );
	exit( -3 );
      }


    find_inv( sigma, dim, inv_sigma );
    break;
  case DIST_LOGISTIC:
    /* Logistic distribution */

    /* Initializing */
    
    /* Since the parameters for each state are rescaled in the end,
       can fix the set of parameters for one of the state (state 0) */

    for( i=1; i<num_states; i++ )
      lambda[i]=0.1+0.8*drand48();
    for( i1=0; i1<dim; i1++ )
      { /* Initializing parameters for the linear terms */

	/* Calculating mean and standard deviation for the i1-th component */
	m=0.0;
	num_entries=0;
	for( n=0; n<data->num_seqs; n++ )
	  {
	    num_entries+=data->num_seqs;
	    for( t=0; t<data->sequence[n]->seq_length; t++ )
	      m+=data->sequence[n]->entry[t]->rdata[i1];
	  }

	m/=(double)num_entries;

	sd=0.0;
	for( n=0; n<data->num_seqs; n++ )
	  for( t=0; t<data->sequence[n]->seq_length; t++ )
	    sd+=(m-data->sequence[n]->entry[t]->rdata[i1])*(m-data->sequence[n]->entry[t]->rdata[i1]);
	sd=sqrt(sd/(double)num_entries);

	/* Calculating lower and upper bound for the parameter (mean +- standard deviation) */
	lower_bound=m-sd;

	for( i=1; i<num_states; i++ )
	  {
	    rho[i][i1]=lower_bound+2*sd*drand48();
	    if( fabs( rho[i][i1] )<INIT_EPSILON )
	      i--;
	    else
	      rho[i][i1]=1.0/rho[i][i1];
	  }

      } /* Initializing parameters for the linear terms */
    break;
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    /* Logistic distribution with transition terms */

    /* Initializing */
    
    /* Since the parameters for each state are rescaled in the end,
       can fix the set of parameters for one of the state (state 0) */

    for( i=1; i<num_states; i++ )
      lambda[i]=0.1+0.8*drand48();
    for( i=1; i<num_states; i++ )
      for( i1=0; i1<num_states; i1++ )
	sigma[i1][i]=0.1+0.8*drand48();
    for( i1=0; i1<dim; i1++ )
      { /* Initializing parameters for the linear terms */

	/* Calculating mean and standard deviation for the i1-th component */
	m=0.0;
	num_entries=0;
	for( n=0; n<data->num_seqs; n++ )
	  {
	    num_entries+=data->num_seqs;
	    for( t=0; t<data->sequence[n]->seq_length; t++ )
	      m+=data->sequence[n]->entry[t]->rdata[i1];
	  }

	m/=(double)num_entries;

	sd=0.0;
	for( n=0; n<data->num_seqs; n++ )
	  for( t=0; t<data->sequence[n]->seq_length; t++ )
	    sd+=(m-data->sequence[n]->entry[t]->rdata[i1])*(m-data->sequence[n]->entry[t]->rdata[i1]);

	sd=sqrt(sd/(double)num_entries);

	/* Calculating lower and upper bound for the parameter (mean +- standard deviation) */
	lower_bound=m-sd;

	for( i=1; i<num_states; i++ )
	  {
	    rho[i][i1]=lower_bound+2*sd*drand48();
	    if( fabs( rho[i][i1] )<INIT_EPSILON )
	      i--;
	    else
	      rho[i][i1]=1.0/rho[i][i1];
	  }

      } /* Initializing parameters for the linear terms */
    break;
  default:;
  } 

  return;
}

double Distribution::prob( DataPoint* datum, DataPoint* prev_datum ) {
  /* Finding the probability of the datum according to the distribution */
  double p;
  double n_exp=0.0;
  
  double* xminusmu;
  /* Chow-Liu trees */
  long** node_index;
  double** mult;

  /* Index variable(s) */
  long i,j,env,e,b;
  long value1,value2;

  /* Temporary variable(s) */
  double sum,prod;
  long i1,i2,ei,m1;
  double* exp_sum;
  double temp;

  switch( type ) {
  case DIST_FACTOR:
    p=1.0;
    for( i=0; i<dim; i++ )
      p*=subdist[i]->prob( datum, prev_datum );
    break;
  case DIST_MIXTURE:
    p=0.0;
    for( i=0; i<num_states; i++ )
      p+=state_prob[i]*subdist[i]->prob( datum, prev_datum );
    break;
  case DIST_CLMIXTURE:
    /* Computing the envelopes */

    /* Allocating additional variables */
    mult=new double*[dim];
    for( i=0; i<dim; i++ )
      mult[i]=new double[num_states];

    /* Precomputing probability values for components */
    for( i=0; i<dim; i++ )
      for( j=0; j<num_states; j++ )
	mult[i][j]=subdist2[i][j]->prob( datum, prev_datum );

    p=1.0;

    for( env=0; env<num_envelopes; env++ )
      /* For each envelope */
      if( envelope[env]->num_nodes==1 )
	{ /* Envelope contains exactly one variable. */
	  m1=envelope[env]->node[0];
	  sum=0.0;
	  for( b=0; b<num_states; b++ )
	    /* Summing over the missing variable */
	    sum+=ind_prob[m1][b]*mult[m1][b];
	  
	  /* Updating the probability */
	  p*=sum;
	} /* Envelope contains exactly one missing variable. */
      else
	{ /* Envelope contains more than one variable. */
	  /* Propagating probabilities in a backward pass */
	  for( e=envelope[env]->num_edges-1; e>=0; e-- )
	    {
	      ei=envelope[env]->edge[e];
	      i1=edge[ei][0];
	      i2=edge[ei][1];
		
	      if( envelope[env]->node[e+1]==i2 )
		{ /* i1 is the parent of i2 */
		  /* Scaling P(i2->)/P(i2) */
		  sum=0.0;
		  for( value2=0; value2<num_states; value2++ )
		    sum+=mult[i2][value2];
		  for( value2=0; value2<num_states; value2++ )
		    mult[i2][value2]/=sum;
		  p*=sum;

		  for( value1=0; value1<num_states; value1++ )
		    { /* Computing the contribution to i1 with value value1 from child i2 */
		      sum=0.0;
		      for( value2=0; value2<num_states; value2++ )
			sum+=mult[i2][value2]*edge_prob[ei][value1][value2];
		      mult[i1][value1]*=sum/ind_prob[i1][value1];
		    } /* Computing the contribution to i1 with value value1 from child i2 */
		} /* i1 is the parent of i2 */
	      else
		{ /* i2 is the parent of i1 */
		  /* Scaling P(i1->)/P(i1) */
		  sum=0.0;
		  for( value1=0; value1<num_states; value1++ )
		    sum+=mult[i1][value1];
		  for( value1=0; value1<num_states; value1++ )
		    mult[i1][value1]/=sum;
		  p*=sum;

		  for( value2=0; value2<num_states; value2++ )
		    { /* Computing the contribution to i2 with value value2 from child i1 */
		      sum=0.0;
		      for( value1=0; value1<num_states; value1++ )
			sum+=mult[i1][value1]*edge_prob[ei][value1][value2];
		      mult[i2][value2]*=sum/ind_prob[i2][value2];
		    } /* Computing the contribution to i2 with value value2 from child i1 */
		} /* i2 is the parent of i1 */		
	    }

	  /* Summing over the root of the envelope */
	  m1=envelope[env]->node[0];
	  sum=0.0;
	  for( b=0; b<num_states; b++ )
	    sum+=mult[m1][b]*ind_prob[m1][b];
	  /* Updating the probability */
	  p*=sum;
	} /* Envelope contains more than one variable. */
    
    /* Deallocating */
    for( i=0; i<dim; i++ )
      delete mult[i];
    delete mult;

    break;
  case DIST_BERNOULLI:
    if( dim==1 )
      if( is_missing( datum->ddata[dim_index[0]] ) )
	p=1.0;
      else
	p=state_prob[datum->ddata[dim_index[0]]];
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    /* !!! Do not know how to handle missing data !!! */
    if( prev_datum )
      {
	/* Not the first entry in the sequence */
	if( !is_missing( prev_datum->ddata[dim_index[0]] ) && 
	    !is_missing( datum->ddata[dim_index[0]] ) )
	  p=cond_state_prob[prev_datum->ddata[dim_index[0]]][datum->ddata[dim_index[0]]];
      }
    else
      /* First entry in the sequence */
      if( !is_missing( datum->ddata[dim_index[0]] ) )
	p=state_prob[datum->ddata[dim_index[0]]];    
      else
	p=1.0;
    break;
  case DIST_UNICONDMVME:
    /* !!! Do not know how to handle missing data !!! */

    /* datum->ddata contains the value for which the probability is to be computed (index 0),
       and the covariates are in the other components */
    /* prev_datum is irrelevant */

    /* First feature index is always missing as it corresponds to the variable in question */
    /* Feature values for real-valued variables are also not specified */

    /* Initializing the exponent sums for state values */
    exp_sum=new double[num_states];
    for( b=0; b<num_states; b++ )
      exp_sum[b]=0.0;

    /* Computing the normalization constant and the exponent for the state value in question */
    
    /* Also computing the exponential sum for all values of the variable */
    for( i=0; i<dim; i++ )
      { /* Considering function i */
	temp=lambda[i];
	for( j=1; j<num_features[i]; j++ )
	  if( feature_index[i][j]<datum->ddim )
	    { /* Categorical feature */
	      if( feature_value[i][j]!=datum->ddata[feature_index[i][j]] )
		{
		  temp=0.0;
		  /* No need to look at other features for this function */
		  j=num_features[i];
		}
	    } /* Categorical feature */
	  else
	    /* Real-valued feature */
	    temp*=datum->rdata[feature_index[i][j]-datum->ddim];

	/* Updating the exp-sum for the appropriate value */
	exp_sum[feature_value[i][0]]+=temp;
      } /* Considering function i */

    /* Computing the normalization constant */
    sum=0.0;
    for( b=0; b<num_states; b++ )
      sum+=exp(exp_sum[b]);

    p=exp(exp_sum[datum->ddata[0]])/sum;

    delete exp_sum;
      
    break;
  case DIST_CHOWLIU:
    /* Computing univariate contributions for instantiated variables */
    p=1.0;
    for( i=0; i<dim; i++ )
      if( !is_missing( datum->ddata[dim_index[i]] ) )
	p*=ind_prob[i][datum->ddata[dim_index[i]]];

    /* Computing the envelopes */
    node_index=new long*[dim];
    for( i=0; i<dim; i++ )
      node_index[i]=new long[2];
    envelope=new Envelope*[dim];

    num_envelopes=compute_envelopes( datum, node_index );

    /* Allocating additional variables */
    mult=new double*[dim];
    for( i=0; i<dim; i++ )
      mult[i]=new double[num_states];

    for( env=0; env<num_envelopes; env++ )
      /* For each envelope */
      if( !is_missing(datum->ddata[dim_index[envelope[env]->node[0]]]) )
	/* Envelope contains no missing values */
	for( e=0; e<envelope[env]->num_edges; e++ )
	  {
	    ei=envelope[env]->edge[e];
	    i1=edge[ei][0];
	    i2=edge[ei][1];
	    p*=edge_prob[ei][datum->ddata[dim_index[i1]]][datum->ddata[dim_index[i2]]]
	      /(ind_prob[i1][datum->ddata[dim_index[i1]]]*ind_prob[i2][datum->ddata[dim_index[i2]]]);
	  }
      else
	if( envelope[env]->num_nodes==1 )
	  { /* Envelope contains exactly one missing variable. */
	    m1=envelope[env]->node[0];
	    for( b=0; b<num_states; b++ )
	      mult[m1][b]=ind_prob[m1][b];
	    sum=0.0;
	    for( b=0; b<num_states; b++ )
	      {/* Summing over the missing variable */
		for( e=0; e<envelope[env]->num_edges; e++ )
		  {
		    ei=envelope[env]->edge[e];
		    i1=edge[ei][0];
		    i2=edge[ei][1];
		    
		    if( m1==i1 )
		      mult[m1][b]*=edge_prob[ei][b][datum->ddata[dim_index[i2]]]/
			(ind_prob[i1][b]*ind_prob[i2][datum->ddata[dim_index[i2]]]);
		    else
		      mult[m1][b]*=edge_prob[ei][datum->ddata[dim_index[i1]]][b]/
			(ind_prob[i1][datum->ddata[dim_index[i1]]]*ind_prob[i2][b]);		    
		  }
		sum+=mult[m1][b];
	      } /* Summing over the missing variable */

	    /* Updating the probability */
	    p*=sum;

	    //	    fprintf( stdout, "Single variable missing in an envelope.  Contribution: %le\n", sum );
	  } /* Envelope contains exactly one missing variable. */
	else
	  { /* Envelope contains more than one missing variable. */
	    for( i=0; i<envelope[env]->num_nodes; i++ )
	      for( b=0; b<num_states; b++ )
		mult[envelope[env]->node[i]][b]=1.0;
	    /*
	      fprintf( stdout, "Number of edges: %ld\n", envelope[env]->num_edges );
	      fprintf( stdout, "Missing variables:" );
	    
	      for( i=0; i<envelope[env]->num_nodes; i++ )
	      fprintf( stdout, "\t%ld", envelope[env]->node[i] );
	      fprintf( stdout, "\n" );
	    */
	    /* Propagating probabilities in a backward pass */
	    for( e=envelope[env]->num_edges-1; e>=0; e-- )
	      {
		ei=envelope[env]->edge[e];
		i1=edge[ei][0];
		i2=edge[ei][1];
		//	fprintf( stdout, "Edge: %ld %ld\n", i1, i2 );
		
		if( !is_missing(datum->ddata[dim_index[i1]] ) )
		  /* Second node of the edge is missing. */
		  for( b=0; b<num_states; b++ )
		    mult[i2][b]*=edge_prob[ei][datum->ddata[dim_index[i1]]][b]/
		      (ind_prob[i1][datum->ddata[dim_index[i1]]]*ind_prob[i2][b]);
		else
		  if( !is_missing( datum->ddata[dim_index[i2]] ) )
		    /* First node of the edge is missing. */
		    for( b=0; b<num_states; b++ )
		      mult[i1][b]*=edge_prob[ei][b][datum->ddata[dim_index[i2]]]/
			(ind_prob[i1][b]*ind_prob[i2][datum->ddata[dim_index[i2]]]);
		  else
		    /* Both nodes are missing. */
		    if( node_index[i1][1]<node_index[i2][1] )
		      { /* i1 is the parent of i2 */
			/* Scaling P(i2->)/P(i2) */
			sum=0.0;
			for( value2=0; value2<num_states; value2++ )
			  sum+=mult[i2][value2];
			for( value2=0; value2<num_states; value2++ )
			  mult[i2][value2]/=sum;
			p*=sum;

			for( value1=0; value1<num_states; value1++ )
			  { /* Computing the contribution to i1 with value value1 from child i2 */
			    sum=0.0;
			    for( value2=0; value2<num_states; value2++ )
			      sum+=mult[i2][value2]*edge_prob[ei][value1][value2];
			    mult[i1][value1]*=sum/ind_prob[i1][value1];
			  } /* Computing the contribution to i1 with value value1 from child i2 */
		      } /* i1 is the parent of i2 */
		    else
		      { /* i2 is the parent of i1 */
			sum=0.0;
			for( value1=0; value1<num_states; value1++ )
			  sum+=mult[i1][value1];
			for( value1=0; value1<num_states; value1++ )
			  mult[i1][value1]/=sum;
			p*=sum;

			for( value2=0; value2<num_states; value2++ )
			  { /* Computing the contribution to i2 with value value2 from child i1 */
			    sum=0.0;
			    for( value1=0; value1<num_states; value1++ )
			      sum+=mult[i1][value1]*edge_prob[ei][value1][value2];
			    mult[i2][value2]*=sum/ind_prob[i2][value2];
			  } /* Computing the contribution to i2 with value value2 from child i1 */
		      } /* i2 is the parent of i1 */		
	      }

	    /* Summing over the root of the envelope */
	    m1=envelope[env]->node[0];
	    sum=0.0;
	    for( b=0; b<num_states; b++ )
	      sum+=mult[m1][b]*ind_prob[m1][b];
	    /* Updating the probability */
	    p*=sum;

	    // fprintf( stdout, "Multiple variables missing in an envelope.  Contribution: %le\n", sum );
	  } /* Envelope contains more than one missing variable. */

    /* Deallocating */
    for( i=0; i<dim; i++ )
      delete mult[i];
    delete mult;
    for( i=0; i<dim; i++ )
      delete node_index[i];
    delete node_index;
    for( env=0; env<num_envelopes; env++ )
      delete envelope[env];
    delete envelope;

    break;
  case DIST_CONDCHOWLIU:
    /* !!! Does not handle missing data !!! */

    /* Initializing the array of flags */
    for( i=0; i<dim; i++ )
      node_used[i]=1;
    for( i=dim; i<2*dim; i++ )
      node_used[i]=0;

    p=1.0;

    for( i=0; i<num_edges; i++ )
      { /* For each edge */
	
	/* value1 is the value for the first node; value2 is the value for the second node */
	if( edge[i][0]<dim )
	  /* Observable */
	  if( prev_datum )
	    value1=prev_datum->ddata[dim_index[edge[i][0]]];
	  else
	    value1=missing_value( (long)0 );
	else
	  /* Non-observable */
	  value1=datum->ddata[dim_index[edge[i][0]-dim]];

	if( edge[i][1]<dim )
	  /* Observable */
	  if( prev_datum )
	    value2=prev_datum->ddata[dim_index[edge[i][1]]];
	  else
	    value2=missing_value( (long)0 );
	else
	  /* Non-observable */
	  value2=datum->ddata[dim_index[edge[i][1]-dim]];

	if( is_missing( value1 ) )
	  {
	    if( !is_missing( value2 ) && !node_used[edge[i][1]] )
	      p*=ind_prob[edge[i][1]][value2];
	  }
	else
	  if( is_missing( value2 ) )
	    {
	      if( !node_used[edge[i][0]] )
		p*=ind_prob[edge[i][0]][value1];
	    }
	  else
	    {
	      p*=edge_prob[i][value1][value2];
	      
	      /* Checking whether need to divide by the probability of the node */
	      if( node_used[edge[i][0]] )
		p/=ind_prob[edge[i][0]][value1];
	      
	      if( node_used[edge[i][1]] )
		p/=ind_prob[edge[i][1]][value2];
	    }

	node_used[edge[i][0]]=1;
	node_used[edge[i][1]]=1;
      } /* For each edge */

    /* Making sure that all nodes have been used */
    for( i=dim; i<2*dim; i++ )
      if( !node_used[i] && !is_missing( datum->ddata[dim_index[i-dim]] ) )
	/* Updating the probability taking the unused node into the account */
	p*=ind_prob[i][datum->ddata[dim_index[i-dim]]];

    break;
  case DIST_ME_BIVAR:
    /* !!! Can't handle missing data yet !!! */
    sum=0.0;
    for( i=0; i<dim; i++ )
      if( datum->ddata[dim_index[i]]==1 )
	for( j=0; j<i+1; j++ )
	  if( datum->ddata[dim_index[j]]==1 )
	    sum+=sigma[i][j];
    p=exp(sum-det);
    break;
  case DIST_BN_ME_BIVAR:
    /* !!! Can't handle missing data !!! */
    p=0.0;  // log-probability
    for( i=0; i<dim; i++ )
      {
	sum=sigma[i][0];
	for( j=1; j<num_features[i]; j++ )
	  if( datum->ddata[dim_index[feature_index[i][j]]]==1 )
	    sum+=sigma[i][j];
	p-=log(1.0+exp(sum));  // Subtracting the normalization constant
	if( datum->ddata[dim_index[i]]==1 )
	  p+=sum;
      }
    p=exp(p);
    break;
  case DIST_BN_CME_BIVAR:
    /* !!! Can't handle missing data !!! */
    if( prev_datum )
      {
	p=0.0;  // log-probability
	for( i=0; i<dim; i++ )
	  {
	    sum=sigma[i][0];
	    for( j=1; j<num_features[i]; j++ )
	      if( feature_index[i][j]<dim )
		{
		  if( datum->ddata[dim_index[feature_index[i][j]]]==1 )
		    sum+=sigma[i][j];
		}
	      else
		{
		  if( prev_datum->ddata[dim_index[feature_index[i][j]-dim]]==1 )
		    sum+=sigma[i][j];
		}
	    p-=log(1.0+exp(sum));  // Subtracting the normalization constant
	    if( datum->ddata[dim_index[i]]==1 )
	      p+=sum;
	  }
	p=exp(p);
      }
    else
      {
	p=1.0;
	for( i=0; i<dim; i++ )
	  if( datum->ddata[dim_index[i]]==1 )
	    p*=state_prob[i];
	  else
	    p*=(1.0-state_prob[i]);
      }
    break;
  case DIST_DELTAEXP:
    /* f(x)=l*exp(-l*x) */    
    if( is_missing( datum->rdata[dim_index[0]] ) )
      p=1.0;
    else
      {
	if( datum->rdata[dim_index[0]]==0.0 )
	  /* First (delta) component */
	  p=mix_prob[0];
	else
	  { /* Mix of exponential components */
	    p=0.0;
	    for( i=0; i<num_states-1; i++ )
	      p+=mix_prob[i+1]*exp_dist(datum->rdata[dim_index[0]], exp_param[i]);
	  } /* Mix of exponential components */
      }
    break;
  case DIST_DELTAGAMMA:
    /* f(x)=x^(param1-1)*exp(-x*param2)*param2^param1/Gamma(param1) */
    if( is_missing( datum->rdata[dim_index[0]] ) )
      p=1.0;
    else
      {
	if( datum->rdata[dim_index[0]]==0.0 )
	  /* First (delta) component */
	  p=mix_prob[0];
	else
	  { /* Mix of gamma components */
	    p=0.0;
	    for( i=0; i<num_states-1; i++ )
	      p+=mix_prob[i+1]*
		gamma_dist(datum->rdata[dim_index[0]],gamma_param1[i],gamma_param2[i]);
	  } /* Mix of gamma components */
      }
    break;
  case DIST_DIRACDELTA:
    if( is_missing( datum->rdata[dim_index[0]] ) )
      p=1.0;
    else
      if( datum->rdata[dim_index[0]]==0.0 )
	p=1.0;
      else
	p=0.0;
    break;
  case DIST_EXP:
    if( is_missing( datum->rdata[dim_index[0]] ) )
      p=1.0;
    else
      if( datum->rdata[dim_index[0]]>0.0 )
	p=exp_dist( datum->rdata[dim_index[0]], exp_param1 );
      else
	p=0.0;
    break;
  case DIST_GAMMA:
    if( is_missing( datum->rdata[dim_index[0]] ) )
      p=1.0;
    else
      if( datum->rdata[dim_index[0]]>0.0 )
	p=gamma_dist( datum->rdata[dim_index[0]], gamma1, gamma2 );
      else
	p=0.0;
    break;
  case DIST_LOGNORMAL:
    if( is_missing( datum->rdata[dim_index[0]] ) )
      p=1.0;
    else
      if( datum->rdata[dim_index[0]]>0.0 )
	p=lognormal_dist( datum->rdata[dim_index[0]], log_normal1, log_normal2 );
      else
	p=0.0;
    break;
  case DIST_NORMAL:
    xminusmu=new double[dim];

    /* Calculating x-mu */
    for( i=0; i<dim; i++ )
      xminusmu[i]=datum->rdata[dim_index[i]]-mu[i];

    /* Calculating (x-mu)'*sigma^(-1)*(x-mu) */
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	n_exp+=xminusmu[i]*inv_sigma[i][j]*xminusmu[j];

    delete xminusmu;
    
    p=exp(-0.5*n_exp)/(pow(2*M_PI,0.5*(double)dim)*sqrt(det));
    break;
  case DIST_NORMALCHAIN:
    xminusmu=new double[dim];

    if( prev_datum )
      {
	/* Calculating W*x_{t-1}+mu-x_t */
	for( i=0; i<dim; i++ )
	  {
	    xminusmu[i]=mu[i]-datum->rdata[dim_index[i]];
	    for( j=0; j<dim; j++ )
	      xminusmu[i]+=W[i][j]*prev_datum->rdata[dim_index[j]];
	  }
	
	/* Calculating (W*x_{t-1}+mu-x_t)'*sigma^(-1)*(W*x_{t-1}+mu-x_t) */
	for( i=0; i<dim; i++ )
	  for( j=0; j<dim; j++ )
	    n_exp+=xminusmu[i]*inv_sigma[i][j]*xminusmu[j];

	p=exp(-0.5*n_exp)/(pow(2*M_PI,0.5*(double)dim)*sqrt(det));
      }
    else
      {
	/* Calculating x-mu */
	for( i=0; i<dim; i++ )
	  xminusmu[i]=datum->rdata[dim_index[i]]-first_mu[i];
	
	/* Calculating (x-mu)'*sigma^(-1)*(x-mu) */
	for( i=0; i<dim; i++ )
	  for( j=0; j<dim; j++ )
	    n_exp+=xminusmu[i]*inv_first_sigma[i][j]*xminusmu[j];

	p=exp(-0.5*n_exp)/(pow(2*M_PI,0.5*(double)dim)*sqrt(first_det));
      }

    delete xminusmu;
    
    break;
  case DIST_NORMALCL:
    xminusmu=new double[dim];

    /* Calculating x-mu */
    for( i=0; i<dim; i++ )
      xminusmu[i]=datum->rdata[dim_index[i]]-mu[i];

    /* Calculating (x-mu)'*sigma^(-1)*(x-mu) */
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	n_exp+=xminusmu[i]*inv_sigma[i][j]*xminusmu[j];

    delete xminusmu;
    
    p=exp(-0.5*n_exp)/(pow(2*M_PI,0.5*(double)dim)*sqrt(det));
    break;
  case DIST_LOGISTIC:
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    /* Implemented in probLogistic */
    break;
  default:
    p=1.0;
  }

  return( p );
}

void Distribution::probLogistic( DataPoint* datum, long prev_state, double* output ) {
  /* Calculates the probabilities of all states given the previous state and X */

  /* Index variables */
  long i,i1;
  
  switch( type ) {
  case DIST_LOGISTIC:
    for( i=0; i<num_states; i++ )
      {
	output[i]=lambda[i];
	for( i1=0; i1<dim; i1++ )
	  output[i]+=rho[i][i1]*datum->rdata[i1];
      }
    normalize_logistic( output, output, num_states, 0 );
    break;
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    for( i=0; i<num_states; i++ )
      {
	if( prev_state<0 )
	  output[i]=lambda[i];
	else
	  output[i]=sigma[prev_state][i];

	for( i1=0; i1<dim; i1++ )
	  output[i]+=rho[i][i1]*datum->rdata[i1];
      }
    normalize_logistic( output, output, num_states, 0 );
    break;
  default:;
  }
  
  return;
}

double Distribution::log_prior() {
  /* Log-probability of the prior distribution */
  double lp=0.0;

  /* Index variable(s) */
  long i,j,v;

  switch( type ) {
  case DIST_FACTOR:
  case DIST_MIXTURE:
    for( i=0; i<dim; i++ )
      lp+=subdist[i]->log_prior();
    break;
  case DIST_CLMIXTURE:
    for( i=0; i<dim; i++ )
      for( j=0; j<num_states; j++ )
	lp+=subdist2[i][j]->log_prior();
    lp-=(double)num_edges*((double)num_states-1.0)*((double)num_states-1.0)*mdl_beta;
    break;
  case DIST_CHOWLIU:
  case DIST_CONDCHOWLIU:
    lp-=(double)num_edges*((double)num_states-1.0)*((double)num_states-1.0)*mdl_beta;
    break;
  case DIST_BN_ME_BIVAR:
  case DIST_BN_CME_BIVAR:
    for( i=0; i<dim; i++ )
      lp-=mdl_beta*((double)num_features[i]-1);
    break;
    break;
  case DIST_BERNOULLI:
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
  case DIST_UNICONDMVME:
  case DIST_ME_BIVAR:
  case DIST_DELTAEXP:
  case DIST_DELTAGAMMA:
  case DIST_DIRACDELTA:
  case DIST_EXP:
  case DIST_GAMMA:
  case DIST_LOGNORMAL:
  case DIST_NORMAL:
  case DIST_NORMALCHAIN:
  case DIST_NORMALCL:
  case DIST_LOGISTIC:
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    break;
  default:;
  }

  return( lp );
}

double Distribution::log_un_probLogistic( DataPoint* datum, long cur_state, long prev_state ) {
  /* Calculates the log of unnormalized probabilities for the current state given the previsous state and X */

  double output;

  /* Index variables */
  long i1;

  switch( type ) {
  case DIST_LOGISTIC:
    output=lambda[cur_state];
    for( i1=0; i1<dim; i1++ )
      output+=rho[cur_state][i1]*datum->rdata[i1];
    break;
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    if( prev_state<0 )
      output=lambda[cur_state];
    else
      output=sigma[prev_state][cur_state];

    for( i1=0; i1<dim; i1++ )
      output+=rho[cur_state][i1]*datum->rdata[i1];
    break;
  default:;
  }

  return( output );  
}

void Distribution::WriteToFile( FILE* output ) {
  /* Writing distribution to a file */

  /* Index variable(s) */
  long i,j,i1,i2,s;

  /* Types of distribution are not written -- only the parameters */
  switch( type ) {
  case DIST_FACTOR:
    for( i=0; i<dim; i++ )
      subdist[i]->WriteToFile( output );
    break;
  case DIST_MIXTURE:
    fprintf( output, "%c %ld-component mixture\n", COMMENT_SYMBOL, num_states );
    fprintf( output, "%c Mixing probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states; i++ )
      fprintf( output, "\t%0.12le", state_prob[i] );
    fprintf( output, "\n" );
    for( i=0; i<num_states; i++ )
      {
	fprintf( output, "%c Mixture component %ld:\n", COMMENT_SYMBOL, i+1 );
	subdist[i]->WriteToFile( output );
      }
    break;
  case DIST_CLMIXTURE:
    fprintf( output, "%c %ld-variable %ld-component tree-structured mixture\n",
	     COMMENT_SYMBOL, dim, num_states );
    fprintf( output, "%c Distribution for mixing variables:\n", COMMENT_SYMBOL );
    fprintf( output, "%c Marginal probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  fprintf( output, "\t%0.12le", ind_prob[i][i1] );
	fprintf( output, "\n" );
      }

    fprintf( output, "%c Number of edges:\n", COMMENT_SYMBOL );
    fprintf( output, "%ld\n", num_edges );
    
    fprintf( output, "%c Edges:\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      { 
	fprintf( output, "%ld %ld", edge[i][0], edge[i][1] );
	fprintf( output, "\t%c Rank: %ld\t MI: %le\n",  
		 COMMENT_SYMBOL, edge_rank[i], edge_MI[i] );
      }
    fprintf( output, "%c Edge probability tables\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    fprintf( output, "\t%0.12le", edge_prob[i][i1][i2] );
	fprintf( output, "\n" );
      }
    for( i=0; i<dim; i++ )
      for( j=0; j<num_states; j++ )
	{
	  fprintf( output, "%c Mixture component for variable %ld value %ld\n",
		   COMMENT_SYMBOL, i+1, j );
	  subdist2[i][j]->WriteToFile( output );
	}
    break;
  case DIST_BERNOULLI:
    // fprintf( output, "%c Mapping of vector dimensions:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", dim_index[i] );
    fprintf( output, "\n" );
    for( i=0; i<num_states; i++ )
      fprintf( output, "\t%0.12le", state_prob[i] );
    fprintf( output, "\n" );
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    // fprintf( output, "%c Mapping of vector dimensions:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", dim_index[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c First entry probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states; i++ )
      fprintf( output, "\t%0.12le", state_prob[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Conditional probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states; i++ )
      {
	for( j=0; j<num_states; j++ )
	  fprintf( output, "\t%0.12le", cond_state_prob[i][j] );
	fprintf( output, "\n" );
      }
    break;
  case DIST_UNICONDMVME:
    fprintf( output, "%c Number of functions:\n", COMMENT_SYMBOL );
    fprintf( output, "%ld\n", dim );

    fprintf( output, "%c Number of features for each function:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", num_features[i] );
    fprintf( output, "\n" );

    fprintf( output, "%c Function feature indices:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	/* First feature index should be missing */
	fprintf( output, "%le ", missing_value( 0.0 ) );
	for( j=1; j<num_features[i]; j++ )
	  fprintf( output, "%ld ", feature_index[i][j] );
	fprintf( output, "\n" );
      }

    fprintf( output, "%c Corresponding function feature values:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<num_features[i]; j++ )
	  if( is_missing( feature_value[i][j] ) )
	    /* Real-valued features do not have corresponding values */
	    fprintf( output, "%le ", missing_value( 0.0 ) );
	  else
	    fprintf( output, "%ld ", feature_value[i][j] );
      }

    fprintf( output, "%c Weights:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%\t0.12le", lambda[i] );
    fprintf( output, "\n" );
    break;
  case DIST_CHOWLIU:
    // fprintf( output, "%c Mapping of vector dimensions:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", dim_index[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Marginal probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  fprintf( output, "\t%0.12le", ind_prob[i][i1] );
	fprintf( output, "\n" );
      }

    fprintf( output, "%c Number of edges:\n", COMMENT_SYMBOL );
    fprintf( output, "%ld\n", num_edges );
    
    fprintf( output, "%c Edges:\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      { 
	fprintf( output, "%ld %ld", edge[i][0], edge[i][1] );
	fprintf( output, "\t%c Rank: %ld\t MI: %le\n",  
		 COMMENT_SYMBOL, edge_rank[i], edge_MI[i] );
      }
    fprintf( output, "%c Edge probability tables\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    fprintf( output, "\t%0.12le", edge_prob[i][i1][i2] );
	fprintf( output, "\n" );
      }
    break;
  case DIST_CONDCHOWLIU:
    // fprintf( output, "%c Mapping of vector dimensions:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", dim_index[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Marginal probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<2*dim; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  fprintf( output, "\t%0.12le", ind_prob[i][i1] );
	fprintf( output, "\n" );
      }

    fprintf( output, "%c Number of edges:\n", COMMENT_SYMBOL );
    fprintf( output, "%ld\n", num_edges );
    
    fprintf( output, "%c Edges:\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      { 
	fprintf( output, "%ld %ld", edge[i][0], edge[i][1] );
	fprintf( output, "\t%c Rank: %ld\t MI: %le\n",  
		 COMMENT_SYMBOL, edge_rank[i], edge_MI[i] );
      }
    fprintf( output, "%c Edge probability tables\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    fprintf( output, "\t%0.12le", edge_prob[i][i1][i2] );
	fprintf( output, "\n" );
      }
    break;
  case DIST_ME_BIVAR:
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", dim_index[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Coefficients:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<i+1; j++ )
	  fprintf( output, "\t%0.12le", sigma[i][j] );
	fprintf( output, "\n" );
      }
    fprintf( output, "%c Exponent of the normalization constant:\n", COMMENT_SYMBOL );
    fprintf( output, "%0.12le\n", det );
    break;
  case DIST_BN_ME_BIVAR:
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", dim_index[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Numbers of features:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", num_features[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Features:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<num_features[i]; j++ ) 
	  fprintf( output, "%ld ", feature_index[i][j] );
	fprintf( output, "\n" );
      }
    fprintf( output, "%c Parameter values:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<num_features[i]; j++ )
	  fprintf( output, "\t%0.12le", sigma[i][j] );
	fprintf( output, "\n" );
      }
    fprintf( output, "%c Simulation order:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", sim_order[i] );
    fprintf( output, "\n" );
    break;
   case DIST_BN_CME_BIVAR:
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", dim_index[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c First entry probabilities of 1:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "\t%0.12le", state_prob[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Numbers of features:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", num_features[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Features:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<num_features[i]; j++ ) 
	  fprintf( output, "%ld ", feature_index[i][j] );
	fprintf( output, "\n" );
      }
    fprintf( output, "%c Parameter values:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<num_features[i]; j++ )
	  fprintf( output, "\t%0.12le", sigma[i][j] );
	fprintf( output, "\n" );
      }
    fprintf( output, "%c Simulation order:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", sim_order[i] );
    fprintf( output, "\n" );
    break;
 case DIST_DELTAEXP:
    // fprintf( output, "%c Mapping of vector dimensions:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", dim_index[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Mixture component probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states; i++ )
      fprintf( output, "\t%0.12le", mix_prob[i] );
    fprintf( output, "\n" );

    fprintf( output, "%c Parameters for exponential components:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states-1; i++ )
      fprintf( output, "%0.12le\n", exp_param[i] ); 
    break;
  case DIST_DELTAGAMMA:
    // fprintf( output, "%c Mapping of vector dimensions:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", dim_index[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Mixture component probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states; i++ )
      fprintf( output, "\t%0.12le", mix_prob[i] );
    fprintf( output, "\n" );

    fprintf( output, "%c Parameters for gamma components:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states-1; i++ )
      fprintf( output, "%0.12le\t%0.12le\n", gamma_param1[i], gamma_param2[i] ); 
    break;
  case DIST_DIRACDELTA:
    fprintf( output, "%c Dirac delta function at %0.12le\n", COMMENT_SYMBOL, delta_value );
    fprintf( output, "%ld\n", dim_index[0] );
    break;
  case DIST_EXP:
    fprintf( output, "%ld\n", dim_index[0] );
    fprintf( output, "%c Parameter for geometric distribution:\n", COMMENT_SYMBOL );
    fprintf( output, "%0.12le\n", exp_param1 ); 
    break;
  case DIST_GAMMA:
    fprintf( output, "%ld\n", dim_index[0] );
    fprintf( output, "%c Gamma distribution:\n", COMMENT_SYMBOL );
    fprintf( output, "%0.12le\t%0.12le\n", gamma1, gamma2 );
    break;
  case DIST_LOGNORMAL:
    fprintf( output, "%ld\n", dim_index[0] );
    fprintf( output, "%c Log-normal distribution:\n", COMMENT_SYMBOL );
    fprintf( output, "%0.12le\t%0.12le\n", log_normal1, log_normal2 );
    break;
  case DIST_NORMAL:
    // fprintf( output, "%c Mapping of vector dimensions:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", dim_index[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Mean vector\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "\t%0.12le", mu[i] );
    fprintf( output, "\n\n" );
    fprintf( output, "%c Covariance matrix\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<dim; j++ )
	  fprintf( output, "\t%0.12le", sigma[i][j] );
	fprintf( output, "\n" );
      }
    break;
  case DIST_NORMALCHAIN:
    // fprintf( output, "%c Mapping of vector dimensions:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", dim_index[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c First state mean:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "\t%0.12le", first_mu[i] );
    fprintf( output, "\n\n" );
    fprintf( output, "%c First state covariance matrix\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<dim; j++ )
	  fprintf( output, "\t%0.12le", first_sigma[i][j] );
	fprintf( output, "\n" );
       }
    fprintf( output, "\n" );
    
    fprintf( output, "%c Linear transformation:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<dim; j++ )
	  fprintf( output, "\t%0.12le", W[i][j] );
	fprintf( output, "\n" );
      }
    fprintf( output, "\n" );

    fprintf( output, "%c Translation vector:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "\t%0.12le", mu[i] );
    fprintf( output, "\n" );

    fprintf( output, "%c Noise covariance:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<dim; j++ )
	  fprintf( output, "\t%0.12le", sigma[i][j] );
	fprintf( output, "\n" );
      }

    break;
  case DIST_NORMALCL:
    // fprintf( output, "%c Mapping of vector dimensions:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", dim_index[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Mean vector\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "\t%0.12le", mu[i] );
    fprintf( output, "\n\n" );
    fprintf( output, "%c Covariance matrix\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<dim; j++ )
	  fprintf( output, "\t%0.12le", sigma[i][j] );
	fprintf( output, "\n" );
      }

    fprintf( output, "%c Number of edges:\n", COMMENT_SYMBOL );
    fprintf( output, "%ld\n", num_edges );
    
    fprintf( output, "%c Edges:\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      { 
	fprintf( output, "%ld %ld", edge[i][0], edge[i][1] );
	fprintf( output, "\t%c Rank: %ld\t MI: %le\n",  
		 COMMENT_SYMBOL, edge_rank[i], edge_MI[i] );
      }

    break;
  case DIST_LOGISTIC:
    for( s=0; s<num_states; s++ )
      {
	fprintf( output, "%c State %ld\n", COMMENT_SYMBOL, s+1 );

	fprintf( output, "%c Constant term\n", COMMENT_SYMBOL );
	fprintf( output, "%0.12le\n\n", lambda[s] );

	fprintf( output, "%c Linear terms\n", COMMENT_SYMBOL );
	for( i=0; i<dim; i++ )
	  fprintf( output, "\t%0.12le", rho[s][i] );
	fprintf( output, "\n\n" );
      }
    break;
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    for( s=0; s<num_states; s++ )
      {
	fprintf( output, "%c State %ld\n", COMMENT_SYMBOL, s+1 );

	fprintf( output, "%c Constant term\n", COMMENT_SYMBOL );
	fprintf( output, "%0.12le\n\n", lambda[s] );

	fprintf( output, "%c Transition terms\n", COMMENT_SYMBOL );
	for( i=0; i<num_states; i++ )
	  fprintf( output, "\t%0.12le", sigma[i][s] );
	fprintf( output, "\n\n" );

	fprintf( output, "%c Linear terms\n", COMMENT_SYMBOL );
	for( i=0; i<dim; i++ )
	  fprintf( output, "\t%0.12le", rho[s][i] );
	fprintf( output, "\n\n" );
      }
    break;
  default:;
  }

  return;
}

void Distribution::WriteToFile2( FILE* output ) {
  /* Writing distribution to a file WITHOUT dimension indices */

  /* Index variable(s) */
  long i,j,i1,i2,s;

  /* Types of distribution are not written -- only the parameters */
  switch( type ) {
  case DIST_FACTOR:
    for( i=0; i<dim; i++ )
      subdist[i]->WriteToFile2( output );
    break;
  case DIST_MIXTURE:
    fprintf( output, "%c %ld-component mixture\n", COMMENT_SYMBOL, num_states );
    fprintf( output, "%c Mixing probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states; i++ )
      fprintf( output, "\t%0.12le", state_prob[i] );
    fprintf( output, "\n" );
    for( i=0; i<num_states; i++ )
      {
	fprintf( output, "%c Mixture component %ld:\n", COMMENT_SYMBOL, i+1 );
	subdist[i]->WriteToFile2( output );
      }
    break;
  case DIST_CLMIXTURE:
    fprintf( output, "%c %ld-variable %ld-component tree-structured mixture\n",
	     COMMENT_SYMBOL, dim, num_states );
    fprintf( output, "%c Distribution for mixing variables:\n", COMMENT_SYMBOL );
    fprintf( output, "%c Marginal probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  fprintf( output, "\t%0.12le", ind_prob[i][i1] );
	fprintf( output, "\n" );
      }

    fprintf( output, "%c Number of edges:\n", COMMENT_SYMBOL );
    fprintf( output, "%ld\n", num_edges );
    
    fprintf( output, "%c Edges:\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      { 
	fprintf( output, "%ld %ld", edge[i][0], edge[i][1] );
	fprintf( output, "\t%c Rank: %ld\t MI: %le\n",  
		 COMMENT_SYMBOL, edge_rank[i], edge_MI[i] );
      }
    fprintf( output, "%c Edge probability tables\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    fprintf( output, "\t%0.12le", edge_prob[i][i1][i2] );
	fprintf( output, "\n" );
      }
    for( i=0; i<dim; i++ )
      for( j=0; j<num_states; j++ )
	{
	  fprintf( output, "%c Mixture component for variable %ld value %ld\n",
		   COMMENT_SYMBOL, i+1, j );
	  subdist2[i][j]->WriteToFile2( output );
	}
    break;
  case DIST_BERNOULLI:
    for( i=0; i<num_states; i++ )
      fprintf( output, "\t%0.12le", state_prob[i] );
    fprintf( output, "\n" );
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    fprintf( output, "%c First entry probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states; i++ )
      fprintf( output, "\t%0.12le", state_prob[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Conditional probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states; i++ )
      {
	for( j=0; j<num_states; j++ )
	  fprintf( output, "\t%0.12le", cond_state_prob[i][j] );
	fprintf( output, "\n" );
      }
    break;
  case DIST_UNICONDMVME:
    fprintf( output, "%c Number of functions:\n", COMMENT_SYMBOL );
    fprintf( output, "%ld\n", dim );

    fprintf( output, "%c Number of features for each function:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", num_features[i] );
    fprintf( output, "\n" );

    fprintf( output, "%c Function feature indices:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	/* First feature index should be missing */
	fprintf( output, "%le ", missing_value( 0.0 ) );
	for( j=1; j<num_features[i]; j++ )
	  fprintf( output, "%ld ", feature_index[i][j] );
	fprintf( output, "\n" );
      }

    fprintf( output, "%c Corresponding function feature values:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<num_features[i]; j++ )
	  if( is_missing( feature_value[i][j] ) )
	    /* Real-valued features do not have corresponding values */
	    fprintf( output, "%le ", missing_value( 0.0 ) );
	  else
	    fprintf( output, "%ld ", feature_value[i][j] );
      }

    fprintf( output, "%c Weights:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%\t0.12le", lambda[i] );
    fprintf( output, "\n" );
    break;
  case DIST_CHOWLIU:
    fprintf( output, "%c Marginal probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  fprintf( output, "\t%0.12le", ind_prob[i][i1] );
	fprintf( output, "\n" );
      }

    fprintf( output, "%c Number of edges:\n", COMMENT_SYMBOL );
    fprintf( output, "%ld\n", num_edges );
    
    fprintf( output, "%c Edges:\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      { 
	fprintf( output, "%ld %ld", edge[i][0], edge[i][1] );
	fprintf( output, "\t%c Rank: %ld\t MI: %le\n",  
		 COMMENT_SYMBOL, edge_rank[i], edge_MI[i] );
      }
    fprintf( output, "%c Edge probability tables\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    fprintf( output, "\t%0.12le", edge_prob[i][i1][i2] );
	fprintf( output, "\n" );
      }
    break;
  case DIST_CONDCHOWLIU:
    fprintf( output, "%c Marginal probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<2*dim; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  fprintf( output, "\t%0.12le", ind_prob[i][i1] );
	fprintf( output, "\n" );
      }

    fprintf( output, "%c Number of edges:\n", COMMENT_SYMBOL );
    fprintf( output, "%ld\n", num_edges );
    
    fprintf( output, "%c Edges:\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      { 
	fprintf( output, "%ld %ld", edge[i][0], edge[i][1] );
	fprintf( output, "\t%c Rank: %ld\t MI: %le\n",  
		 COMMENT_SYMBOL, edge_rank[i], edge_MI[i] );
      }
    fprintf( output, "%c Edge probability tables\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      {
	for( i1=0; i1<num_states; i1++ )
	  for( i2=0; i2<num_states; i2++ )
	    fprintf( output, "\t%0.12le", edge_prob[i][i1][i2] );
	fprintf( output, "\n" );
      }
    break;
  case DIST_ME_BIVAR:
    fprintf( output, "%c Coefficients:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<i+1; j++ )
	  fprintf( output, "\t%0.12le", sigma[i][j] );
	fprintf( output, "\n" );
      }
    fprintf( output, "%c Exponent of the normalization constant:\n", COMMENT_SYMBOL );
    fprintf( output, "%0.12le\n", det );
    break;
  case DIST_BN_ME_BIVAR:
    fprintf( output, "%c Numbers of features:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", num_features[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Features:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<num_features[i]; j++ ) 
	  fprintf( output, "%ld ", feature_index[i][j] );
	fprintf( output, "\n" );
      }
    fprintf( output, "%c Parameter values:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<num_features[i]; j++ )
	  fprintf( output, "\t%0.12le", sigma[i][j] );
	fprintf( output, "\n" );
      }
    fprintf( output, "%c Simulation order:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", sim_order[i] );
    fprintf( output, "\n" );
    break;
  case DIST_BN_CME_BIVAR:
    fprintf( output, "%c First entry probabilities of 1:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "\t%0.12le", state_prob[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Numbers of features:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", num_features[i] );
    fprintf( output, "\n" );
    fprintf( output, "%c Features:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<num_features[i]; j++ ) 
	  fprintf( output, "%ld ", feature_index[i][j] );
	fprintf( output, "\n" );
      }
    fprintf( output, "%c Parameter values:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<num_features[i]; j++ )
	  fprintf( output, "\t%0.12le", sigma[i][j] );
	fprintf( output, "\n" );
      }
    fprintf( output, "%c Simulation order:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "%ld ", sim_order[i] );
    fprintf( output, "\n" );
    break;
 case DIST_DELTAEXP:
    fprintf( output, "%c Mixture component probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states; i++ )
      fprintf( output, "\t%0.12le", mix_prob[i] );
    fprintf( output, "\n" );

    fprintf( output, "%c Parameters for exponential components:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states-1; i++ )
      fprintf( output, "%0.12le\n", exp_param[i] ); 
    break;
  case DIST_DELTAGAMMA:
    fprintf( output, "%c Mixture component probabilities:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states; i++ )
      fprintf( output, "\t%0.12le", mix_prob[i] );
    fprintf( output, "\n" );

    fprintf( output, "%c Parameters for gamma components:\n", COMMENT_SYMBOL );
    for( i=0; i<num_states-1; i++ )
      fprintf( output, "%0.12le\t%0.12le\n", gamma_param1[i], gamma_param2[i] ); 
    break;
  case DIST_DIRACDELTA:
    fprintf( output, "%c Dirac delta function at %0.12le\n", COMMENT_SYMBOL, delta_value );
    break;
  case DIST_EXP:
    fprintf( output, "%c Parameter for geometric distribution:\n", COMMENT_SYMBOL );
    fprintf( output, "%0.12le\n", exp_param1 ); 
    break;
  case DIST_GAMMA:
    fprintf( output, "%c Gamma distribution:\n", COMMENT_SYMBOL );
    fprintf( output, "%0.12le\t%0.12le\n", gamma1, gamma2 );
    break;
  case DIST_LOGNORMAL:
    fprintf( output, "%c Log-normal distribution:\n", COMMENT_SYMBOL );
    fprintf( output, "%0.12le\t%0.12le\n", log_normal1, log_normal2 );
    break;
  case DIST_NORMAL:
    fprintf( output, "%c Mean vector\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "\t%0.12le", mu[i] );
    fprintf( output, "\n\n" );
    fprintf( output, "%c Covariance matrix\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<dim; j++ )
	  fprintf( output, "\t%0.12le", sigma[i][j] );
	fprintf( output, "\n" );
      }
    break;
  case DIST_NORMALCHAIN:
    fprintf( output, "%c First state mean:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "\t%0.12le", first_mu[i] );
    fprintf( output, "\n\n" );
    fprintf( output, "%c First state covariance matrix\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<dim; j++ )
	  fprintf( output, "\t%0.12le", first_sigma[i][j] );
	fprintf( output, "\n" );
       }
    fprintf( output, "\n" );
    
    fprintf( output, "%c Linear transformation:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<dim; j++ )
	  fprintf( output, "\t%0.12le", W[i][j] );
	fprintf( output, "\n" );
      }
    fprintf( output, "\n" );

    fprintf( output, "%c Translation vector:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "\t%0.12le", mu[i] );
    fprintf( output, "\n" );

    fprintf( output, "%c Noise covariance:\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<dim; j++ )
	  fprintf( output, "\t%0.12le", sigma[i][j] );
	fprintf( output, "\n" );
      }

    break;
  case DIST_NORMALCL:
    fprintf( output, "%c Mean vector\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      fprintf( output, "\t%0.12le", mu[i] );
    fprintf( output, "\n\n" );
    fprintf( output, "%c Covariance matrix\n", COMMENT_SYMBOL );
    for( i=0; i<dim; i++ )
      {
	for( j=0; j<dim; j++ )
	  fprintf( output, "\t%0.12le", sigma[i][j] );
	fprintf( output, "\n" );
      }

    fprintf( output, "%c Number of edges:\n", COMMENT_SYMBOL );
    fprintf( output, "%ld\n", num_edges );
    
    fprintf( output, "%c Edges:\n", COMMENT_SYMBOL );
    for( i=0; i<num_edges; i++ )
      { 
	fprintf( output, "%ld %ld", edge[i][0], edge[i][1] );
	fprintf( output, "\t%c Rank: %ld\t MI: %le\n",  
		 COMMENT_SYMBOL, edge_rank[i], edge_MI[i] );
      }

    break;
  case DIST_LOGISTIC:
    for( s=0; s<num_states; s++ )
      {
	fprintf( output, "%c State %ld\n", COMMENT_SYMBOL, s+1 );

	fprintf( output, "%c Constant term\n", COMMENT_SYMBOL );
	fprintf( output, "%0.12le\n\n", lambda[s] );

	fprintf( output, "%c Linear terms\n", COMMENT_SYMBOL );
	for( i=0; i<dim; i++ )
	  fprintf( output, "\t%0.12le", rho[s][i] );
	fprintf( output, "\n\n" );
      }
    break;
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    for( s=0; s<num_states; s++ )
      {
	fprintf( output, "%c State %ld\n", COMMENT_SYMBOL, s+1 );

	fprintf( output, "%c Constant term\n", COMMENT_SYMBOL );
	fprintf( output, "%0.12le\n\n", lambda[s] );

	fprintf( output, "%c Transition terms\n", COMMENT_SYMBOL );
	for( i=0; i<num_states; i++ )
	  fprintf( output, "\t%0.12le", sigma[i][s] );
	fprintf( output, "\n\n" );

	fprintf( output, "%c Linear terms\n", COMMENT_SYMBOL );
	for( i=0; i<dim; i++ )
	  fprintf( output, "\t%0.12le", rho[s][i] );
	fprintf( output, "\n\n" );
      }
    break;
  default:;
  }

  return;
}

long Distribution::num_params() {
  /* Number of free parameters in the distribution */
  long num=0;
 
  /* Index variables */
  long i,j;

  switch( type ) {
  case DIST_FACTOR:
    for( i=0; i<dim; i++ )
      num+=subdist[i]->num_params();
    break;
  case DIST_MIXTURE:
    /* Mixing probabilities */
    num+=num_states-1;
    /* Mixing components */
    for( i=0; i<num_states; i++ )
      num+=subdist[i]->num_params();
    break;
  case DIST_CLMIXTURE:
    num+=dim*(num_states-1);
    num+=num_edges*(num_states-1)*(num_states-1);
    for( i=0; i<dim; i++ )
      for( j=0; j<num_states; j++ )
	num+=subdist2[i][j]->num_params();
    break;
  case DIST_BERNOULLI:
    num+=num_states-1;
    break;
  case DIST_CONDBERNOULLI:
  case DIST_CONDBERNOULLIG:
    /* First sequence entry probabilities */
    num+=num_states-1;
    /* Conditional probabilities */
    num+=num_states*(num_states-1);
    break;
  case DIST_UNICONDMVME:
    /* !!! Only counting weights !!! */
    num+=dim;
    break;
  case DIST_CHOWLIU:
    /* Quite tricky */

    /* 
       Let nc = number of outcomes, assuming all nodes have the same number of outcomes.
       Each node contributes nc-1 parameters corresponding to probability values.
       Each edge contributes nc^2-1-2*(nc-1)=(nc-1)^2 parameters corresponding to
       probability values.

       BIC implemented here does not count the parameters indicating what edges are
       selected.  At this point in time, it is not clear how to include structure
       specification parameters into BIC.
    */

    /* !!! Assuming all nodes have the same number of outcomes !!! */
    /* Counting node contributions */
    num+=dim*(num_states-1);

    /* Counting edge contributions */
    num+=num_edges*(num_states-1)*(num_states-1);

    break;
  case DIST_CONDCHOWLIU:
    /* Quite tricky */

    /* 
       Let nc = number of outcomes, assuming all nodes have the same number of outcomes.
       Each node contributes nc-1 parameters corresponding to probability values.
       Each edge contributes nc^2-1-2*(nc-1)=(nc-1)^2 parameters corresponding to
       probability values.

       We need to include nodes for both time t and t-1.

       BIC implemented here does not count the parameters indicating what edges are
       selected.  At this point in time, it is not clear how to include structure
       specification parameters into BIC.
    */

    /* !!! Assuming all nodes have the same number of outcomes !!! */
    /* Counting node contributions */
    num+=2*dim*(num_states-1);

    /* Counting edge contributions */
    num+=num_edges*(num_states-1)*(num_states-1);

    break;
  case DIST_ME_BIVAR:
    num+=dim*(dim+1)/2;
    break;
  case DIST_BN_ME_BIVAR:
    num=0;
    for( i=0; i<dim; i++ )
      num+=num_features[i];
    /* !!! Not counting indices indicating which features are selected !!! */
    break;
  case DIST_BN_CME_BIVAR:
    num=dim;  // First entry probabilities of 1
    for( i=0; i<dim; i++ )
      num+=num_features[i];
    /* !!! Not counting indices indicating which features are selected !!! */
    break;
 case DIST_DELTAEXP:
    /* Mixing probabilities */
    num+=num_states-1;
    /* Exponential parameters */
    num+=num_states-1;
    break;
  case DIST_DELTAGAMMA:
    /* Mixing probabilities */
    num+=num_states-1;
    /* Gamma parameters */
    num+=2*(num_states-1);
    break;
  case DIST_DIRACDELTA:
    break;
  case DIST_EXP:
    num+=1;
    break;
  case DIST_GAMMA:
    num+=2;
    break;
  case DIST_LOGNORMAL:
    num+=2;
    break;
  case DIST_NORMAL:
    /* Mean vector */
    num+=dim;
    /* Covariance matrix */
    num+=dim*(dim+1)/2;
    break;
  case DIST_NORMALCHAIN:
    /* First state mean vector */
    num+=dim;
    /* First state covariance matrix */
    num+=dim*(dim+1)/2;
    /* Linear transformation matrix */
    num+=dim^2;
    /* Translation vector */
    num+=dim;
    /* Noise covariance matrix */
    num+=dim*(dim+1)/2;
    break;
  case DIST_NORMALCL:
    /* Mean vector */
    num+=dim;
    /* Tree-structured covariance matrix */
    num+=(dim+num_edges);
    break;
  case DIST_LOGISTIC:
    /* First state */
    num+=num_states-1;
    /* Linear terms */
    num+=dim*(num_states-1);
    break;
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    /* First state */
    num+=num_states-1;
    /* Transition */
    num+=num_states*(num_states-1);
    /* Linear terms */
    num+=dim*(num_states-1);
    break;
  default:;
  }

  return( num );
}

void Distribution::UpdateEmissionParameters( Data* data, double** prob_s, double norm_const ) {
  /* data -- the output data
     prob_s -- p_old(.|R_nt,X_nt)
     norm_const -- sum_n sum_{t_n} p_old(.|R_nt,X_nt)
  */

  /* Index variable(s) */
  long i;

  switch( type ) {
  case DIST_FACTOR:
    /* Conditionally independent components */

    /* Updating parameters of the sub-components */
    for( i=0; i<dim; i++ )
      subdist[i]->UpdateEmissionParameters( data, prob_s, norm_const );
    break;
  case DIST_MIXTURE:
    /* Mixture */
    UpdateEmissionMixture( data, prob_s );
    break;
  case DIST_CLMIXTURE:
    /* Tree-dependent mixture */
    UpdateEmissionCLMixture( data, prob_s );
    break;
  case DIST_BERNOULLI:
    /* Univariate Bernoulli */
    UpdateEmissionBernoulli( data, prob_s );
    break;
  case DIST_CONDBERNOULLI:
    /* Univariate conditional Bernoulli */
    UpdateEmissionConditionalBernoulli( data, prob_s, 0 );
    break;
  case DIST_CONDBERNOULLIG:
    /* Univariate conditional Bernoulli with averaged first entry probabilities */
    UpdateEmissionConditionalBernoulli( data, prob_s, 1 );
  case DIST_UNICONDMVME:
    /* Polytomous logistic distribution */
    
    /* !!! Updated in a different manner !!! */

    break;
  case DIST_CHOWLIU:
    /* Chow-Liu tree */
    UpdateEmissionChowLiu( data, prob_s, norm_const );
    break;
  case DIST_CONDCHOWLIU:
    /* Conditional Chow-Liu tree */
    UpdateEmissionConditionalChowLiu( data, prob_s, norm_const );
    break;
  case DIST_ME_BIVAR:
    /* Full bivariate MaxEnt */
    UpdateEmissionFullBivariateMaxEnt( data, prob_s, norm_const );
    break;
  case DIST_BN_ME_BIVAR:
    /* PUC-MaxEnt */
    UpdateEmissionPUCMaxEnt( data, prob_s, norm_const );
    break;
  case DIST_BN_CME_BIVAR:
    /* Conditional PUC-MaxEnt */
    UpdateEmissionConditionalPUCMaxEnt( data, prob_s, norm_const );
    break;
  case DIST_DELTAEXP:
    /* Mixture of one delta function and multiple exponential functions */
    UpdateEmissionDeltaExponential( data, prob_s, norm_const );
    break;
  case DIST_DELTAGAMMA:
    /* Mixture of one gamma function and multiple gamma functions */
    UpdateEmissionDeltaGamma( data, prob_s, norm_const );
    break;
  case DIST_DIRACDELTA:
    /* Dirac's delta */
    break;
  case DIST_EXP:
    /* Exponential (geometric) distribution */
    UpdateEmissionExponential( data, prob_s );    
    break;
  case DIST_GAMMA:
    /* Gamma distribution */
    UpdateEmissionGamma( data, prob_s );
    break;
  case DIST_LOGNORMAL:
    /* Log-normal distribution */
    UpdateEmissionLognormal( data, prob_s );
    break;
  case DIST_NORMAL:
    /* Multivariate Gaussian distribution */
    UpdateEmissionGaussian( data, prob_s, norm_const );
    break;
  case DIST_NORMALCHAIN:
    /* Auto-regressive Gaussian */
    UpdateEmissionARGaussian( data, prob_s, norm_const );
    break;
  case DIST_NORMALCL:
    /* Gaussian with tree structured covariance matrix */
    UpdateEmissionGaussianChowLiu( data, prob_s, norm_const );
    break;
  case DIST_LOGISTIC:
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    fprintf( stderr, "Cannot have logistic distribution as an emission!\n" );
    exit( -1 );
    break;
  default:;
  }
  
  return;
}

double Distribution::weighted_ll( Data* data, double** prob_s ) {
  /* The total expected log-likelihood of the data under the distribution
     and the probabilities of the data points */

  double ll=0.0;
  
  /* Index variable(s) */
  long n,t;

  for( n=0; n<data->num_seqs; n++ )
    for( t=0; t<data->sequence[n]->seq_length; t++ )
      if( t==0 )
	ll+=prob_s[n][t]*log(prob( data->sequence[n]->entry[t], NULL ));
      else
	ll+=prob_s[n][t]*log(prob( data->sequence[n]->entry[t], data->sequence[n]->entry[t-1] ));

  return( ll );
}

void Distribution::UpdateTransitionParameters( Data* data, 
					       Data* input,
					       long iteration ) {
  
  /* Updating the parameters of the input distribution */
  
  /* Since the logistic dunction does not have a closed-form update rule,
     need to perform the update iteratively */

  long num_grad_runs;    // Number of need to choose a new direction vector

  Distribution* g;       // Current gradient
  Distribution* g_prev;  // Previous gradient
  Distribution* h;       // Current direction vector

  double a,b;
  
  /* Index variable(s) */
  long i,j,n,t,index;

  /* Temporary variable(s) */
  double sum;

  switch( type ) {
  case DIST_LOGISTIC:
  case DIST_TRANSLOGISTIC:
  case DIST_TRANSLOGISTICG:
    if( iteration>NUM_BURNOFF3_ITER )
      num_grad_runs=4;
    else if( iteration>NUM_BURNOFF2_ITER )
      num_grad_runs=3;
    else if( iteration>NUM_BURNOFF1_ITER )
      num_grad_runs=2;
    else
      num_grad_runs=1;
    
    /* Calculating the initial gradient vector */
    g=logistic_gradient( input, uni_prob[0], joint_prob, log_un_prob );
    
    /* Maximizing the objective function in the direction of the gradient */
    h=g->copy();
    
    linear_search_logistic_NR( input, uni_prob[0], joint_prob, log_un_prob, h, iteration );
    
    for( index=1; index<num_grad_runs; index++ )
      {
	/* Performing an iteration of Polak-Ribiere conjugate gradient algorithm */
	g_prev=g;
	g=logistic_gradient( input, uni_prob[0], joint_prob, log_un_prob );
	
	/* Calculating the adjustment coefficient */
	
	/* Denominator */
	b=0.0;
	for( i=1; i<num_states; i++ )
	  {
	    b+=g_prev->lambda[i]*g_prev->lambda[i];
	    if( type==DIST_TRANSLOGISTIC )
	      for( j=0; j<num_states; j++ )
		b+=g_prev->sigma[j][i]*g_prev->sigma[j][i];
	    for( j=0; j<dim; j++ )
	      b+=g_prev->rho[i][j]*g_prev->rho[i][j];
	  }
	
	/* Numerator */
	a=0.0;
	for( i=1; i<num_states; i++ )
	  {
	    a+=g->lambda[i]*(g->lambda[i]-g_prev->lambda[i]);
	    if( type==DIST_TRANSLOGISTIC )
	      for( j=0; j<num_states; j++ )
		a+=g->sigma[j][i]*(g->sigma[j][i]-g_prev->sigma[j][i]);
	    for( j=0; j<dim; j++ )
	      a+=g->rho[i][j]*(g->rho[i][j]-g_prev->rho[i][j]);
	  }
	
	/* Deallocating the previous gradient */
	delete g_prev;
	
	a/=b;
	
	/* New vector h=g+a*h */
	for( i=1; i<num_states; i++ )
	  {
	    h->lambda[i]=g->lambda[i]+a*h->lambda[i];
	    if( type==DIST_TRANSLOGISTIC )
	      for( j=0; j<num_states; j++ )
		h->sigma[j][i]=g->sigma[j][i]+a*h->sigma[j][i];
	    for( j=0; j<dim; j++ )
	      h->rho[i][j]=g->rho[i][j]+a*h->rho[i][j];
	  }
	
	linear_search_logistic_NR( input, uni_prob[0], joint_prob, log_un_prob, h, iteration );
      }
    
    /* Deallocating the current gradient */
    delete g;
    
    /* Deallocating the direction vector */
    delete h;
    
    break;
  case DIST_BERNOULLI:
    /* State probabilities */
    for( i=0; i<num_states; i++ )
      {
	state_prob[i]=0.0;
	for( n=0; n<data->num_seqs; n++ )
	  for( t=0; t<data->sequence[n]->seq_length; t++ )
	    state_prob[i]+=uni_prob[0][i][n][t];
      }

    sum=0.0;
    for( i=0; i<num_states; i++ )
      sum+=state_prob[i];
    
    /* Normalizing */
    for( i=0; i<num_states; i++ )
      state_prob[i]/=sum;
    
    break;
  case DIST_CONDBERNOULLI:
    /* First state probabilities */
    sum=0.0;
    for( i=0; i<num_states; i++ )
      {
	state_prob[i]=0.0;
	for( n=0; n<data->num_seqs; n++ )
	  state_prob[i]+=uni_prob[0][i][n][0];
	sum+=state_prob[i];
      }
    
    /* Normalizing */
    for( i=0; i<num_states; i++ )
      state_prob[i]/=sum;

    /* Transition probabilities */
    for( j=0; j<num_states; j++ )
      { /* Summing over S_{t-1} */
	sum=0.0;
	for( i=0; i<num_states; i++ )
	  { /* Summing over S_t */
	    cond_state_prob[j][i]=0.0;
	    
	    for( n=0; n<data->num_seqs; n++ )
	      for( t=1; t<data->sequence[n]->seq_length; t++ )
		cond_state_prob[j][i]+=joint_prob[i][j][n][t];

	    sum+=cond_state_prob[j][i];
	  } /* Summing over S_t */
    
	/* Normalizing */
	for( i=0; i<num_states; i++ )
	  cond_state_prob[j][i]/=sum;
      } /* Summing over S_{t-1} */
    break;
  case DIST_CONDBERNOULLIG:
    /* First state probabilities */
    sum=0.0;
    for( i=0; i<num_states; i++ )
      {
	state_prob[i]=0.0;
	for( n=0; n<data->num_seqs; n++ )
	  for( t=0; t<data->sequence[n]->seq_length; t++ )
	    state_prob[i]+=uni_prob[0][i][n][t];
      }

    sum=0.0;
    for( i=0; i<num_states; i++ )
      sum+=state_prob[i];
    
    /* Normalizing */
    for( i=0; i<num_states; i++ )
      state_prob[i]/=sum;

    /* Transition probabilities */
    for( j=0; j<num_states; j++ )
      { /* Summing over S_{t-1} */
	sum=0.0;
	for( i=0; i<num_states; i++ )
	  { /* Summing over S_t */
	    cond_state_prob[j][i]=0.0;
	    
	    for( n=0; n<data->num_seqs; n++ )
	      for( t=1; t<data->sequence[n]->seq_length; t++ )
		//		cond_state_prob[j][i]+=Bntij[i][j][n][t];
		cond_state_prob[j][i]+=joint_prob[i][j][n][t];

	    sum+=cond_state_prob[j][i];
	  } /* Summing over S_t */
    
	/* Normalizing */
	for( i=0; i<num_states; i++ )
	  cond_state_prob[j][i]/=sum;
      } /* Summing over S_{t-1} */
    break;
  default:;
  }

  return;
}

Distribution* Distribution::logistic_gradient( Data* input, 
					       double*** Anti,
					       double**** Bntij,
					       double**** log_un_prob ) {
  /* Calculating the gradient vector for the logistic transition */
  
  /* Variables passed:
     input -- input data
     Anti -- the array of probabilities P(S_nt=i|R_n,X_n,Theta)
     Bntij -- the array of probabilities P(S_nt=i,S_n,t-1=j|R_n,X_n,Theta)
     log_un_prob -- unnormalized log probabilities of S_nt=i|S_n,t-1=j,X_nt
  */

  Distribution* gradient;         // The gradient of the Q function evaluated at the current set of parameters 
  double**** prob;                // Current normalized probabilities of S_nt=i|S_n,t-1=j,X_nt 

  /* Index variables */
  long i,j,n,t,d;

  /* Temporary variable(s) */
  double sum;
  double temp;
  double max;
  
  switch( type ) {
  case DIST_TRANSLOGISTIC:
    /* Allocating the structure for the gradient */
    gradient=new Distribution( type, num_states, dim );
    
    /* Allocating the arrays */
    prob=new double***[num_states];

    for( i=1; i<num_states; i++ )
      {
	prob[i]=new double**[num_states];
	for( j=0; j<num_states; j++ )
	  {
	    prob[i][j]=new double*[input->num_seqs];
	    for( n=0; n<input->num_seqs; n++ )
	      prob[i][j][n]=new double[input->sequence[n]->seq_length];
	  }
      }

    /* Calculating P(S_nt=state|X_nt, R_n,t-1,c) */
    for( n=0; n<input->num_seqs; n++ )
      {
	/* First hidden state in the sequence */
	max=0.0;
	
	for( i=1; i<num_states; i++ )
	  if( max<log_un_prob[i][0][n][0] )
	    max=log_un_prob[i][0][n][0];
      
	sum=exp(-max);
	for( i=1; i<num_states; i++ )
	  /* Calculating unnormalized P(S_n1=i|X_n1) */
	  sum+=exp(log_un_prob[i][0][n][0]-max);
      
	/* Calculating normalized P(S_n1=state|X_n1) */
	for( i=1; i<num_states; i++ )
	  prob[i][0][n][0]=exp(log_un_prob[i][0][n][0]-max)/sum;
      
	for( t=1; t<input->sequence[n]->seq_length; t++ )
	  /* Not the first state in the sequence */
	  for( j=0; j<num_states; j++ )
	    {
	      max=0.0;
	      for( i=1; i<num_states; i++ )
		if( max<log_un_prob[i][j][n][t] )
		  max=log_un_prob[i][j][n][t];
	      
	      sum=exp(-max);
	      for( i=1; i<num_states; i++ )
		/* Calculating unnormalized P(S_nt=i|S_n,t-1=j,X_nt) */
		sum+=exp(log_un_prob[i][j][n][t]-max);
	      
	      /* Calculating normalized P(S_nt=state|S_n,t-1=j,X_nt) */
	      for( i=1; i<num_states; i++ )
		prob[i][j][n][t]=exp(log_un_prob[i][j][n][t]-max)/sum;
	    }
      }

    /* Calculating the values of the gradient */
    for( i=1; i<num_states; i++ )
      { /* For each state */
	/* State 1 is ignored since its parameters are never updated */
	for( n=0; n<input->num_seqs; n++ )
	  for( t=0; t<input->sequence[n]->seq_length; t++ )
	    {
	      if( t==0 )
		{
		  gradient->lambda[i]+=(Anti[i][n][t]-prob[i][0][n][t]);
		  for( d=0; d<dim; d++ )
		    gradient->rho[i][d]+=(Anti[i][n][t]-prob[i][0][n][t])*input->sequence[n]->entry[t]->rdata[d];
		}
	      else
		{
		  temp=0.0;
		  for( j=0; j<num_states; j++ )
		    {
		      gradient->sigma[j][i]+=(Bntij[i][j][n][t]-Anti[j][n][t-1]*prob[i][j][n][t]);
		      temp+=Anti[j][n][t-1]*prob[i][j][n][t];
		    }
		  for( d=0; d<dim; d++ )
		    gradient->rho[i][d]+=(Anti[i][n][t]-temp)*input->sequence[n]->entry[t]->rdata[d];
		}
	    }
      } /* For each state */

    /* Deallocating the arrays */  
    for( i=1; i<num_states; i++ )
      {
	for( j=0; j<num_states; j++ )
	  {
	    for( n=0; n<input->num_seqs; n++ )
	      delete prob[i][j][n];
	    delete prob[i][j];
	  }
	delete prob[i];
      }
    delete prob;

    break;
  case DIST_TRANSLOGISTICG:
    /* Allocating the structure for the gradient */
    gradient=new Distribution( type, num_states, dim );
  
    /* Allocating the arrays */
    prob=new double***[num_states];
    
    for( i=1; i<num_states; i++ )
      {
	prob[i]=new double**[num_states];
	for( j=0; j<num_states; j++ )
	  {
	    prob[i][j]=new double*[input->num_seqs];
	    for( n=0; n<input->num_seqs; n++ )
	      prob[i][j][n]=new double[input->sequence[n]->seq_length];
	  }
      }

    /* Calculating P(S_nt=state|X_nt, R_n,t-1,c) */
    for( n=0; n<input->num_seqs; n++ )
      {
	/* First hidden state in the sequence */
	max=0.0;
	
	for( i=1; i<num_states; i++ )
	  if( max<log_un_prob[i][0][n][0] )
	    max=log_un_prob[i][0][n][0];
	
	sum=exp(-max);
	for( i=1; i<num_states; i++ )
	  /* Calculating unnormalized P(S_n1=i|X_n1) */
	  sum+=exp(log_un_prob[i][0][n][0]-max);
	
	/* Calculating normalized P(S_n1=state|X_n1) */
	for( i=1; i<num_states; i++ )
	  prob[i][0][n][0]=exp(log_un_prob[i][0][n][0]-max)/sum;
	
	for( t=1; t<input->sequence[n]->seq_length; t++ )
	  /* Not the first state in the sequence */
	  for( j=0; j<num_states; j++ )
	    {
	      max=0.0;
	      for( i=1; i<num_states; i++ )
		if( max<log_un_prob[i][j][n][t] )
		  max=log_un_prob[i][j][n][t];

	      sum=exp(-max);
	      for( i=1; i<num_states; i++ )
		/* Calculating unnormalized P(S_nt=i|S_n,t-1=j,X_nt) */
		sum+=exp(log_un_prob[i][j][n][t]-max);
	      
	      /* Calculating normalized P(S_nt=state|S_n,t-1=j,X_nt) */
	      for( i=1; i<num_states; i++ )
		prob[i][j][n][t]=exp(log_un_prob[i][j][n][t]-max)/sum;
	    }
      }

    /* Calculating the values of the gradient */
    for( i=1; i<num_states; i++ )
      { /* For each state */
	/* State 1 is ignored since its parameters are never updated */
	for( n=0; n<input->num_seqs; n++ )
	  for( t=0; t<input->sequence[n]->seq_length; t++ )
	    {
	      if( t==0 )
		{
		  gradient->lambda[i]+=(Anti[i][n][t]-prob[i][0][n][t]);
		  for( d=0; d<dim; d++ )
		    gradient->rho[i][d]+=(Anti[i][n][t]-prob[i][0][n][t])*input->sequence[n]->entry[t]->rdata[d];
		}
	      else
		{
		  temp=0.0;
		  for( j=0; j<num_states; j++ )
		    {
		      gradient->sigma[j][i]+=(Bntij[i][j][n][t]-Anti[j][n][t-1]*prob[i][j][n][t]);
		      temp+=Anti[j][n][t-1]*prob[i][j][n][t];
		    }
		  for( d=0; d<dim; d++ )
		    gradient->rho[i][d]+=(Anti[i][n][t]-temp)*input->sequence[n]->entry[t]->rdata[d];
		}
	    }
      } /* For each state */

    /* Deallocating the arrays */  
    for( i=1; i<num_states; i++ )
      {
	for( j=0; j<num_states; j++ )
	  {
	    for( n=0; n<input->num_seqs; n++ )
	      delete prob[i][j][n];
	    delete prob[i][j];
	  }
	delete prob[i];
      }
    delete prob;
    break;
  case DIST_LOGISTIC:
    /* Allocating the structure for the gradient */
    gradient=new Distribution( type, num_states, dim );
    
    /* Allocating the arrays */
    prob=new double***[num_states];
    
    for( i=1; i<num_states; i++ )
      {
	prob[i]=new double**[num_states];
	prob[i][0]=new double*[input->num_seqs];
	for( n=0; n<input->num_seqs; n++ )
	  prob[i][0][n]=new double[input->sequence[n]->seq_length];
      }

    /* Calculating P(S_nt=state|X_nt, R_n,t-1,c) */
    for( n=0; n<input->num_seqs; n++ )
      for( t=0; t<input->sequence[n]->seq_length; t++ )
	/* Not the first state in the sequence */
	{
	  max=0.0;
	  
	  for( i=1; i<num_states; i++ )
	    if( max<log_un_prob[i][0][n][t] )
	      max=log_un_prob[i][0][n][t];
	  
	  sum=exp(-max);
	  for( i=1; i<num_states; i++ )
	    /* Calculating unnormalized P(S_nt=i|X_nt) */
	    sum+=exp(log_un_prob[i][0][n][t]-max);
	  
	  /* Calculating normalized P(S_nt=state|X_nt) */
	  for( i=1; i<num_states; i++ )
	    prob[i][0][n][t]=exp(log_un_prob[i][0][n][t]-max)/sum;
	}

    /* Calculating the values of the gradient */
    for( i=1; i<num_states; i++ )
      { /* For each state */
	/* State 1 is ignored since its parameters are never updated */
	for( n=0; n<input->num_seqs; n++ )
	  for( t=0; t<input->sequence[n]->seq_length; t++ )
	    {
	      gradient->lambda[i]+=(Anti[i][n][t]-prob[i][0][n][t]);
	      for( d=0; d<dim; d++ )
		gradient->rho[i][d]+=(Anti[i][n][t]-prob[i][0][n][t])*input->sequence[n]->entry[t]->rdata[d];
	    }
      } /* For each state */

    /* Deallocating the arrays */  
    for( i=1; i<num_states; i++ )
      {
	for( n=0; n<input->num_seqs; n++ )
	  delete prob[i][0][n];
	delete prob[i][0];
	delete prob[i];
      }
    delete prob;
  
    break;
  default:;
  }
  
  return( gradient );
}

void Distribution::linear_search_logistic_NR( Data* input, 
					      double*** Anti,
					      double**** Bntij,
					      double**** log_un_prob,
					      Distribution* delta,
					      long iter ) {
  /* Line search method (utilizing Newton Raphson) to find the maximum of the function */
  
  /* Variables passed:
     input -- input data
     Anti -- the array of probabilities P(S_nt=i|R_n,X_n,Theta)
     Bntij -- the array of probabilities P(S_nt=i,S_n,t-1=j|R_n,X_n,Theta)
     log_un_prob -- unnormalized log probabilities of S_nt=i|S_n,t-1=j,X_nt
     delta -- the direction along which the search is performed
     iter -- current iteration of the EM algorithm (if needed)
  */

  double**** adj_fact;            // The adjustment factor (sigma+rho*x) for the direction
  
  double**** prob;                // Current normalized probabilities of S_nt=i|S_n,t-1=j,X_nt 

  double f;                       // Current value of the function (partial derivative of the log-likelihood)
  double f_prev;                  // Previous value of f

  double fprime=0.0;              // Derivative (second partial derivative of the log-likelihood)
  double factor;                  // f/fprime
  
  double gamma=0.0;               // Value of the scalar to be optimized
  double gamma_prev;              // Previous value of gamma

  double found_positive=0;
  double gamma_positive;          // Value of gamma corresponding to the positive f
  double fneg;
  double found_negative=0;
  double gamma_negative;          // Value of gamma corresponding to the negative f
  double fpos;

  long iteration;                 // Number of iterations

  double epsilon;                 // Sensitivity threshold

  /* Index variables */
  long i,j,n,t;

  /* Temporary variable(s) */
  double sum;
  double max;
  
  if( !(type==DIST_LOGISTIC || type==DIST_TRANSLOGISTIC) )
    { /* !!! !!! */
      exit( -1 );
    } /* !!! !!! */

  /* Allocating the arrays */
  adj_fact=new double***[num_states];
  for( i=1; i<num_states; i++ )
    {
      adj_fact[i]=new double**[num_states];
      switch( type ) {
      case DIST_TRANSLOGISTIC:
	for( j=0; j<num_states; j++ )
	  {
	    adj_fact[i][j]=new double*[input->num_seqs];
	    for( n=0; n<input->num_seqs; n++ )
	      adj_fact[i][j][n]=new double[input->sequence[n]->seq_length];
	  }
	break;
      case DIST_LOGISTIC:
	adj_fact[i][0]=new double*[input->num_seqs];
	for( n=0; n<input->num_seqs; n++ )
	  adj_fact[i][0][n]=new double[input->sequence[n]->seq_length];
	break;
      default:;
      }
    }

  prob=new double***[num_states];
  for( i=1; i<num_states; i++ )
    {
      prob[i]=new double**[num_states];
      switch( type ) {
      case DIST_TRANSLOGISTIC:
	for( j=0; j<num_states; j++ )
	  {
	    prob[i][j]=new double*[input->num_seqs];
	    for( n=0; n<input->num_seqs; n++ )
	      prob[i][j][n]=new double[input->sequence[n]->seq_length];
	  }
	break;
      case DIST_LOGISTIC:
	prob[i][0]=new double*[input->num_seqs];
	for( n=0; n<input->num_seqs; n++ )
	  prob[i][0][n]=new double[input->sequence[n]->seq_length];
	break;
      default:;
      }
    }

  /* Precalculating the adjustment values */
  for( i=1; i<num_states; i++ )
    for( n=0; n<input->num_seqs; n++ )
      switch( type ) {
      case DIST_TRANSLOGISTIC:
	adj_fact[i][0][n][0]=delta->log_un_probLogistic( input->sequence[n]->entry[0], i, -1 );
	for( t=1; t<input->sequence[n]->seq_length; t++ )
	  for( j=0; j<num_states; j++ )
	    adj_fact[i][j][n][t]=delta->log_un_probLogistic( input->sequence[n]->entry[t], i, j );
	break;
      case DIST_LOGISTIC:
	for( t=0; t<input->sequence[n]->seq_length; t++ )
	  adj_fact[i][0][n][t]=delta->log_un_probLogistic( input->sequence[n]->entry[t], i, -1 );
	break;
      default:;
      }

  /* Calculating P(S_nt=state|X_nt, R_n,t-1,c) */
  for( n=0; n<input->num_seqs; n++ )
    switch( type ) {
    case DIST_TRANSLOGISTIC: 
      /* First state in the sequence */
      max=0.0;
      for( i=1; i<num_states; i++ )
	if( max<log_un_prob[i][0][n][0] )
	  max=log_un_prob[i][0][n][0];

      sum=exp(-max);
      for( i=1; i<num_states; i++ )
	/* Calculating unnormalized P(S_n1=i|X_n1) */
	sum+=exp(log_un_prob[i][0][n][0]-max);
	      
      /* Calculating normalized P(S_n1=state|X_n1) */
      for( i=1; i<num_states; i++ )
	prob[i][0][n][0]=exp(log_un_prob[i][0][n][0]-max)/sum;

      for( t=1; t<input->sequence[n]->seq_length; t++ )
	/* Not the first state in the sequence */
	for( j=0; j<num_states; j++ )
	  {
	    max=0.0;
	    for( i=1; i<num_states; i++ )
	      if( max<log_un_prob[i][j][n][t] )
		max=log_un_prob[i][j][n][t];

	    sum=exp(-max);
	    for( i=1; i<num_states; i++ )
	      /* Calculating unnormalized P(S_nt=i|S_n,t-1=j,X_nt) */
	      sum+=exp(log_un_prob[i][j][n][t]-max);
	      	    
	    /* Calculating normalized P(S_nt=state|S_n,t-1=j,X_nt) */
	    for( i=1; i<num_states; i++ )
	      prob[i][j][n][t]=exp(log_un_prob[i][j][n][t]-max)/sum;
	  }
      break;
    case DIST_LOGISTIC:
      for( t=0; t<input->sequence[n]->seq_length; t++ )
	{
	  max=0.0;
	  for( i=1; i<num_states; i++ )
	    if( max<log_un_prob[i][0][n][t] )
	      max=log_un_prob[i][0][n][t];

	  sum=exp(-max);
	  for( i=1; i<num_states; i++ )
	    sum+=exp(log_un_prob[i][0][n][t]-max);

	  for( i=1; i<num_states; i++ )
	    prob[i][0][n][t]=exp(log_un_prob[i][0][n][t]-max)/sum;
	}
      break;
    default:;
    }

  /* Calculating initial value of the function (gamma=0) */
  f=0.0;
  for( i=1; i<num_states; i++ )
    for( n=0; n<input->num_seqs; n++ )
      { /* For each sequence */
	
	/* First state */
	f+=adj_fact[i][0][n][0]*(Anti[i][n][0]-prob[i][0][n][0]);

	for( t=1; t<input->sequence[n]->seq_length; t++ )
	  if( type==DIST_LOGISTIC )
	    f+=adj_fact[i][0][n][t]*(Anti[i][n][t]-prob[i][0][n][t]);
	  else
	    for( j=0; j<num_states; j++ )
	      f+=adj_fact[i][j][n][t]*(Bntij[i][j][n][t]-Anti[j][n][t-1]*prob[i][j][n][t]);
      } /* For each sequence */

  if( !found_positive && f>0.0 )
    {
      gamma_positive=gamma;
      fpos=f;
      found_positive=1;      
    }
  else if( !found_negative && f<0.0 )
    {
      gamma_negative=gamma;
      fneg=f;
      found_negative=1;
    }

#ifdef NEWTONRAPHSON_VERBOSE
  fprintf( stdout, "Initial gamma=%0.12le, f(gamma)=%0.12le\n", gamma, f );
#endif

  epsilon=MIN_NR_EPSILON/(double)iter;

  iteration=0;
  while( fabs( f )>epsilon && iteration<MAX_NR_ITERATIONS )
    {
      /* Calculating the second derivative fprime */
      fprime=0.0;
      for( i=1; i<num_states; i++ )
	for( n=0; n<input->num_seqs; n++ )
	  {
	    fprime-=adj_fact[i][0][n][0]*adj_fact[i][0][n][0]*prob[i][0][n][0]*(1.0-prob[i][0][n][0]);
	    for( t=1; t<input->sequence[n]->seq_length; t++ )
	      if( type==DIST_LOGISTIC )		
		fprime-=adj_fact[i][0][n][t]*adj_fact[i][0][n][t]*prob[i][0][n][t]*(1.0-prob[i][0][n][t]);
	      else
		for( j=0; j<num_states; j++ )
		  fprime-=Anti[j][n][t-1]*adj_fact[i][j][n][t]*adj_fact[i][j][n][t]*prob[i][j][n][t]*(1.0-prob[i][j][n][t]);
	  }
      
      if( fabs( fprime )<NR_EPSILON )
	{
	  //	  fprintf( stderr, "Newton-Raphson is stuck!  Exiting iterations\n" );
	  iteration=MAX_NR_ITERATIONS;
	}

      /* Storing the previous values */
      f_prev=f;
      gamma_prev=gamma;
      
      /* Updating the value in question */
      factor=f/fprime;
      gamma-=factor;
      
      /* Updating the unnormalized probabilities */
      for( i=1; i<num_states; i++ )
	for( n=0; n<input->num_seqs; n++ )
	  {
	    log_un_prob[i][0][n][0]-=factor*adj_fact[i][0][n][0];
	    for( t=1; t<input->sequence[n]->seq_length; t++ )
	      if( type==DIST_LOGISTIC )
		log_un_prob[i][0][n][t]-=factor*adj_fact[i][0][n][t];
	      else
		for( j=0; j<num_states; j++ )
		  log_un_prob[i][j][n][t]-=factor*adj_fact[i][j][n][t];
	  }

      /* Calculating normalized probabilities */
      for( n=0; n<input->num_seqs; n++ )
	{
	  /* First state in the sequence */
	  max=0.0;
	  for( i=1; i<num_states; i++ )
	    if( max<log_un_prob[i][0][n][0] )
	      max=log_un_prob[i][0][n][0];
	  
	  sum=exp(-max);
	  for( i=1; i<num_states; i++ )
	    /* Calculating unnormalized P(S_n1=i|X_n1) */
	    sum+=exp(log_un_prob[i][0][n][0]-max);
	  
	  /* Calculating normalized P(S_n1=state|X_n1) */
	  for( i=1; i<num_states; i++ )
	    prob[i][0][n][0]=exp(log_un_prob[i][0][n][0]-max)/sum;
	  
	  for( t=1; t<input->sequence[n]->seq_length; t++ )
	    /* Not the first state in the sequence */
	    if( type==DIST_LOGISTIC )
	      {
		max=0.0;
		for( i=1; i<num_states; i++ )
		  if( max<log_un_prob[i][0][n][t] )
		    max=log_un_prob[i][0][n][t];

		sum=exp(-max);
		for( i=1; i<num_states; i++ )
		  sum+=exp(log_un_prob[i][0][n][t]-max);
		
		for( i=1; i<num_states; i++ )
		  prob[i][0][n][t]=exp(log_un_prob[i][0][n][t]-max)/sum;
	      }
	    else
	      for( j=0; j<num_states; j++ )
		{
		  max=0.0;
		  for( i=1; i<num_states; i++ )
		    if( max<log_un_prob[i][j][n][t] )
		      max=log_un_prob[i][j][n][t];

		  sum=exp(-max);
		  for( i=1; i<num_states; i++ )
		    /* Calculating unnormalized P(S_nt=i|S_n,t-1=j,X_nt) */
		    sum+=exp(log_un_prob[i][j][n][t]-max);
		  
		  /* Calculating normalized P(S_nt=state|S_n,t-1=j,X_nt) */
		  for( i=1; i<num_states; i++ )
		    prob[i][j][n][t]=exp(log_un_prob[i][j][n][t]-max)/sum;
		}
	}

      /* Calculating the value of the function */
      f=0.0;
      for( i=1; i<num_states; i++ )
	for( n=0; n<input->num_seqs; n++ )
	  { /* For each sequence */
	    
	    /* First state */
	    f+=adj_fact[i][0][n][0]*(Anti[i][n][0]-prob[i][0][n][0]);
	    
	    for( t=1; t<input->sequence[n]->seq_length; t++ )
	      if( type==DIST_LOGISTIC )
		f+=adj_fact[i][0][n][t]*(Anti[i][n][t]-prob[i][0][n][t]);
	      else
		for( j=0; j<num_states; j++ )
		  f+=adj_fact[i][j][n][t]*(Bntij[i][j][n][t]-Anti[j][n][t-1]*prob[i][j][n][t]);
	  } /* For each sequence */
      
      if( f>0.0 )
	{
	  if( !found_positive )
	    {
	      gamma_positive=gamma;
	      fpos=f;
	      found_positive=1;
	    }
	  else if( f<fpos )
	    {
	      gamma_positive=gamma;
	      fpos=f;
	    }	    
	}
      else
	/* f<0.0 */
	if( !found_negative )
	  {
	    gamma_negative=gamma;
	    fneg=f;
	    found_negative=1;
	  }
	else if( f>fneg )
	  {
	    gamma_negative=gamma;
	    fneg=f;
	  }

      iteration++;
      
#ifdef NEWTONRAPHSON_VERBOSE
      fprintf( stdout, "Iteration %ld: new value=%0.12le, f(gamma)=%0.12le; previous f'(gamma)=%0.8le; factor=%0.8le\n", 
	       iteration, gamma, f, fprime, factor );
#endif

      if( (!finite(f) || (fabs(f)>fabs(f_prev))) && iteration<MAX_NR_ITERATIONS )
	{ /* Diverging */
#ifdef NEWTONRAPHSON_VERBOSE
	  fprintf( stdout, "Absolute value of f is increasing -- needs adjustment\n" );
#endif

	  /* Dividing factor until reach a better point */
	  do
	    {
	      factor/=2.0;
	      gamma=gamma_prev-factor;
	      f=logistic_derivative_along_vector( input, Anti, Bntij, delta, gamma );
#ifdef NEWTONRAPHSON_VERBOSE
	      fprintf( stdout, "Trying factor %0.12le; new f=%0.12le\n", factor, f );
#endif
	    }
	  while( (!finite(f) && finite(factor)) || (fabs(f)>fabs(f_prev) && fabs(f-f_prev)>COMP_EPSILON) );

	  if( !finite(f) || fabs(f-f_prev)<=COMP_EPSILON )
	    { /* Resetting to previous value of gamma and exiting */
	      gamma=gamma_prev;
	      iteration=MAX_NR_ITERATIONS;
#ifdef NEWTONRAPHSON_VERBOSE
	      fprintf( stdout, "Unable to improve on f\n" );
#endif
	    } /* Resetting to previous value of gamma and exiting */
	  else
	    {
#ifdef NEWTONRAPHSON_VERBOSE
	      fprintf( stdout, "Updated value=%0.12le, f(gamma)=%0.12le, factor=%0.8le\n",
		       gamma, f, factor );
#endif
	    }
	  
	  /* Updating log_un_prob */
	  for( i=1; i<num_states; i++ )
	    for( n=0; n<input->num_seqs; n++ )
	      {
		log_un_prob[i][0][n][0]=log_un_probLogistic( input->sequence[n]->entry[0], i, -1 ) + 
		  gamma*delta->log_un_probLogistic( input->sequence[n]->entry[0], i, -1 );
		for( t=1; t<input->sequence[n]->seq_length; t++ )
		  if( type==DIST_LOGISTIC )
		    {
		      log_un_prob[i][0][n][t]=log_un_probLogistic( input->sequence[n]->entry[t], i, -1 ) +
			gamma*delta->log_un_probLogistic( input->sequence[n]->entry[t], i, -1 );
		    }
		  else
		    for( j=0; j<num_states; j++ )
		      log_un_prob[i][j][n][t]=log_un_probLogistic( input->sequence[n]->entry[t], i, j ) +
			gamma*delta->log_un_probLogistic( input->sequence[n]->entry[t], i, j );

	      }
	 
	} /* Diverging */
    }

  if( iteration>=MAX_NR_ITERATIONS )
    { /* Newton-Raphson failed */

      if( found_negative && found_positive )
	{ /* Finding a zero by halving the interval */
#ifdef NEWTONRAPHSON_VERBOSE
	  fprintf( stdout, "Newton-Raphson failed; running halving algorithm\n" );
#endif

	  linear_search_logistic_halving( input, Anti, Bntij, log_un_prob, 
					  delta, iter, gamma_negative, gamma_positive );
	} /* Finding a zero by halving the interval */
      else
	/* Recalculating the original log-unnormalized probabilities */
	for( i=1; i<num_states; i++ )
	  for( n=0; n<input->num_seqs; n++ )
	    {
	      log_un_prob[i][0][n][0]=log_un_probLogistic( input->sequence[n]->entry[0], i, -1 );
	      for( t=1; t<input->sequence[n]->seq_length; t++ )
		if( type==DIST_LOGISTIC )
		  log_un_prob[i][0][n][t]=log_un_probLogistic( input->sequence[n]->entry[t], i, -1 );
		else
		  for( j=0; j<num_states; j++ )
		    log_un_prob[i][j][n][t]=log_un_probLogistic( input->sequence[n]->entry[t], i, j );
	    }
    } /* Newton-Raphson failed */
  else
    { /* Run finished fine */
      /* Updating the set of parameters by gamma*delta */
      if( fprime>0.0 )
	{
	  fprintf( stderr, "f'=%0.12le: Newton-Raphson found minimum instead of a maximum!\n", fprime );
	}
      
      for( i=1; i<num_states; i++ )
	{
	  lambda[i]+=gamma*delta->lambda[i];
	  if( type==DIST_TRANSLOGISTIC )
	    for( j=0; j<num_states; j++ )
	      sigma[j][i]+=gamma*delta->sigma[j][i];
	  for( j=0; j<dim; j++ )
	    rho[i][j]+=gamma*delta->rho[i][j];
	}
    } /* Run finished fine */

  /* Deallocating the arrays */
  for( i=1; i<num_states; i++ )
    {
      if( type==DIST_LOGISTIC ) 
	{
	  for( n=0; n<input->num_seqs; n++ )
	    delete adj_fact[i][0][n];
	  delete adj_fact[i][0];
	}
      else
	for( j=0; j<num_states; j++ )
	  {
	    for( n=0; n<input->num_seqs; n++ )
	      delete adj_fact[i][j][n];
	    delete adj_fact[i][j];
	  }
      delete adj_fact[i];
    }
  delete adj_fact;
  
  for( i=1; i<num_states; i++ )
    {
      if( type==DIST_LOGISTIC )
	{
	  for( n=0; n<input->num_seqs; n++ )
	    delete prob[i][0][n];
	  delete prob[i][0];
	}
      else
	for( j=0; j<num_states; j++ )
	  {
	    for( n=0; n<input->num_seqs; n++ )
	      delete prob[i][j][n];
	    delete prob[i][j];
	  }
      delete prob[i];
    }
  delete prob;
  
  return;
}

void Distribution::linear_search_logistic_halving( Data* input, 
						   double*** Anti,
						   double**** Bntij,
						   double**** log_un_prob,
						   Distribution* delta,
						   long iter,
						   double gamma_negative,
						   double gamma_positive ) {
  /* Line search method (utilizing Newton Raphson) to find the maximum of the function */
  
  /* Variables passed:
     input -- input data
     Anti -- the array of probabilities P(S_nt=i|R_n,X_n,Theta)
     Bntij -- the array of probabilities P(S_nt=i,S_n,t-1=j|R_n,X_n,Theta)
     log_un_prob -- unnormalized log probabilities of S_nt=i|S_n,t-1=j,X_nt
     delta -- the direction along which the search is performed
     iter -- current iteration of the EM algorithm (if needed)
     gamma_negative -- value of gamma corresponding to the negative value of f
     gamma_positive -- value of gamma corresponding to the positive value of f
  */

  double f;                       // Current value of the function (partial derivative of the log-likelihood)
  double gamma;                   // Value of the scalar to be optimized
  double epsilon;                 // Sensitivity threshold

  /* Index variables */
  long i,j,n,t;
  
  if( !(type==DIST_TRANSLOGISTIC || type==DIST_LOGISTIC) )
    { /* !!! !!! */
      exit( -1 );
    } /* !!! !!! */

  epsilon=MIN_NR_EPSILON/(double)iter;

#ifdef NEWTONRAPHSON_VERBOSE
  f=logistic_derivative_along_vector( input, Anti, Bntij, delta, gamma_negative );
  fprintf( stdout, "Negative gamma=%0.12le; f=%0.12le\n", gamma_negative, f );
  f=logistic_derivative_along_vector( input, Anti, Bntij, delta, gamma_positive );
  fprintf( stdout, "Positive gamma=%0.12le; f=%0.12le\n", gamma_positive, f );
#endif  

  do
    {
      gamma=0.5*(gamma_negative+gamma_positive);
      f=logistic_derivative_along_vector( input, Anti, Bntij, delta, gamma );

#ifdef NEWTONRAPHSON_VERBOSE
      fprintf( stdout, "Current gamma=%0.12le; current f=%0.12le\n", gamma, f );
#endif
      if( f<0.0 )
	gamma_negative=gamma;
      else
	gamma_positive=gamma;
    }
  while( fabs( f )>epsilon && fabs(gamma_negative-gamma_positive)>COMP_EPSILON );

  /* Updating the set of parameters */
  for( i=1; i<num_states; i++ )
    {
      lambda[i]+=gamma*delta->lambda[i];
      if( type==DIST_TRANSLOGISTIC )
	for( j=0; j<num_states; j++ )
	  sigma[j][i]+=gamma*delta->sigma[j][i];
      for( j=0; j<dim; j++ )
	rho[i][j]+=gamma*delta->rho[i][j];
    }

  /* Recalculating log-unnormalized probabilities */
  for( i=1; i<num_states; i++ )
    for( n=0; n<input->num_seqs; n++ )
      {
	log_un_prob[i][0][n][0]=log_un_probLogistic( input->sequence[n]->entry[0], i, -1 );
	for( t=1; t<input->sequence[n]->seq_length; t++ )
	  if( type==DIST_LOGISTIC )
	    log_un_prob[i][0][n][t]=log_un_probLogistic( input->sequence[n]->entry[t], i, -1 );
	  else
	    for( j=0; j<num_states; j++ )
	      log_un_prob[i][j][n][t]=log_un_probLogistic( input->sequence[n]->entry[t], i, j );
      }
  
  return;
}

double Distribution::logistic_derivative_along_vector( Data* input, 
						       double*** Anti,
						       double**** Bntij,
						       Distribution* delta,
						       double gamma ) {
  double f;

  double* prob;
  double* adj_fact;

  /* Index variable(s) */
  long i,j,n,t;

  prob=new double[num_states];
  adj_fact=new double[num_states];

  
  f=0.0;
  for( n=0; n<input->num_seqs; n++ )
    {
      /* First entry */
      /* Calculating adjustments */
      for( i=1; i<num_states; i++ )
	adj_fact[i]=delta->log_un_probLogistic( input->sequence[n]->entry[0], i, -1 );

      /* Calculating probabilities P(Sn1=i|X_n1,new Theta) */
      for( i=1; i<num_states; i++ )
	prob[i]=log_un_probLogistic( input->sequence[n]->entry[0], i, -1 ) + gamma*adj_fact[i];

      normalize_logistic( prob, prob, num_states, 1 );

      for( i=1; i<num_states; i++ )
	f+=adj_fact[i]*(Anti[i][n][0]-prob[i]);

      /* Rest of the entries */
      for( t=1; t<input->sequence[n]->seq_length; t++ )
	if( type==DIST_LOGISTIC )
	  {
	    for( i=1; i<num_states; i++ )
	      {
		adj_fact[i]=delta->log_un_probLogistic( input->sequence[n]->entry[t], i, -1 );
		prob[i]=log_un_probLogistic( input->sequence[n]->entry[t], i, -1 ) + gamma*adj_fact[i];
	      }

	    normalize_logistic( prob, prob, num_states, 1 );

	    for( i=1; i<num_states; i++ )
	      f+=adj_fact[i]*(Anti[i][n][t]-prob[i]);
	  }
	else
	  for( j=0; j<num_states; j++ )
	    {
	      /* Calculating adjustments */
	      for( i=1; i<num_states; i++ )
		adj_fact[i]=delta->log_un_probLogistic( input->sequence[n]->entry[t], i, j );
	      
	      /* Calculating probabilities P(Snt=i|S_n,t-1=j, X_nt,new Theta) */
	      for( i=1; i<num_states; i++ )
		prob[i]=log_un_probLogistic( input->sequence[n]->entry[t], i, j ) + gamma*adj_fact[i];
	      
	      normalize_logistic( prob, prob, num_states, 1 );

	      for( i=1; i<num_states; i++ )
		f+=adj_fact[i]*(Bntij[i][j][n][t]-Anti[j][n][t-1]*prob[i]);
	    }
    }

  delete prob;
  delete adj_fact;

  return( f );
}

Envelope::Envelope( long max_num ) {
  node=new long[max_num];
  edge=new long[max_num];
  num_nodes=0;
  num_edges=0;
}

Envelope::~Envelope() {
  delete node;
  delete edge;
}

long Distribution::compute_envelopes( DataPoint* v, long** node_index ) {
  long num_envelopes=0;

  /* Index variable(s) */
  long i,e,i1,i2;

  /* Initializing */
  for( i=0; i<dim; i++ )
    node_index[i][0]=-1;

  for( e=0; e<num_edges; e++ ) 
    { /* For each of the edges */
      /* Shortcuts */
      i1=edge[e][0];
      i2=edge[e][1];

      if( !is_missing( v->ddata[dim_index[i1]] ) )
	{ /* First variable is instantiated */
	  if( !is_missing( v->ddata[dim_index[i2]] ) )
	    { /* Both variables are instantiated */
	      if( node_index[i1][0]!=-1 )
		{ /* First variable belongs to an envelope -- adding the second variable to the same envelope */
		  envelope[node_index[i1][0]]->edge[envelope[node_index[i1][0]]->num_edges]=e;
		  envelope[node_index[i1][0]]->num_edges++;
		  envelope[node_index[i1][0]]->node[envelope[node_index[i1][0]]->num_nodes]=i2;
		  envelope[node_index[i1][0]]->num_nodes++;
		  node_index[i2][0]=node_index[i1][0];
		  node_index[i2][1]=envelope[node_index[i1][0]]->num_nodes-1;
		} /* First variable belongs to an envelope -- adding the second variable to the same envelope */
	      else
		if( node_index[i2][0]!=-1 )
		  { /* Second variables belongs to an envelope -- ading the first variables to the same envelope */
		    envelope[node_index[i2][0]]->edge[envelope[node_index[i2][0]]->num_edges]=e;
		    envelope[node_index[i2][0]]->num_edges++;
		    envelope[node_index[i2][0]]->node[envelope[node_index[i2][0]]->num_nodes]=i1;
		    envelope[node_index[i2][0]]->num_nodes++;
		    node_index[i1][0]=node_index[i2][0];
		    node_index[i1][1]=envelope[node_index[i2][0]]->num_nodes-1;
		  } /* Second variables belongs to an envelope -- ading the first variables to the same envelope */
		else
		  { /* Neither variable is in an envelope -- creating a new envelope */
		    envelope[num_envelopes]=new Envelope( dim );
		    envelope[num_envelopes]->num_nodes=2;
		    envelope[num_envelopes]->num_edges=1;
		    envelope[num_envelopes]->edge[0]=e;
		    envelope[num_envelopes]->node[0]=edge[e][0];
		    envelope[num_envelopes]->node[1]=edge[e][1];
		    node_index[i1][0]=num_envelopes;
		    node_index[i2][0]=num_envelopes;
		    node_index[i1][1]=0;
		    node_index[i2][1]=1;
		    num_envelopes++;
		  } /* Neither variable is in an envelope -- creating a new envelope */
	    } /* Both variables are instantiated */
	  else
	    { /* First variable is instantiated; the second one is not. */
	      if( node_index[i2][0]!=-1 )
		{ /* Missing variable already belongs to an envelope; adding the edge to the same envelope. */
		  envelope[node_index[i2][0]]->edge[envelope[node_index[i2][0]]->num_edges]=e;
		  envelope[node_index[i2][0]]->num_edges++;
		} /* Missing variable already belongs to an envelope; adding the edge to the same envelope. */
	      else
		{ /* Creating an envelope for the uninstantiated variable */
		  envelope[num_envelopes]=new Envelope( dim );
		  envelope[num_envelopes]->num_nodes=1;
		  envelope[num_envelopes]->num_edges=1;
		  envelope[num_envelopes]->edge[0]=e;
		  envelope[num_envelopes]->node[0]=i2;
		  node_index[i2][0]=num_envelopes;
		  node_index[i2][1]=0;
		  num_envelopes++;
		} /* Creating an envelope for the uninstantiated variable */
	    } /* First variable is instantiated; the second one is not. */
	} /* First variable is instantiated */
      else
	{ /* First variable is missing */
	  if( !is_missing( v->ddata[dim_index[i2]] ) )
	    { /* First variable is missing; the second one is present. */
	      if( node_index[i1][0]!=-1 )
		{ /* Missing variable already belongs to an envelope.  Adding the edge to that envelope. */
		  envelope[node_index[i1][0]]->edge[envelope[node_index[i1][0]]->num_edges]=e;
		  envelope[node_index[i1][0]]->num_edges++;
		} /* Missing variable already belongs to an envelope.  Adding the edge to that envelope. */
	      else
		{ /* Creating an envelope for the missing variable */
		  envelope[num_envelopes]=new Envelope( dim );
		  envelope[num_envelopes]->num_nodes=1;
		  envelope[num_envelopes]->num_edges=1;
		  envelope[num_envelopes]->edge[0]=e;
		  envelope[num_envelopes]->node[0]=i1;
		  node_index[i1][0]=num_envelopes;
		  node_index[i1][1]=0;
		  num_envelopes++;
		} /* Creating an envelope for the missing variable */
	    } /* First variable is missing; the second one is present. */
	  else
	    { /* Both variables are missing */
	      if( node_index[i1][0]!=-1 )
		{ /* First variable belongs to an envelope; adding the second variable to the same envelope. */
		  envelope[node_index[i1][0]]->edge[envelope[node_index[i1][0]]->num_edges]=e;
		  envelope[node_index[i1][0]]->num_edges++;
		  envelope[node_index[i1][0]]->node[envelope[node_index[i1][0]]->num_nodes]=i2;
		  envelope[node_index[i1][0]]->num_nodes++;
		  node_index[i2][0]=node_index[i1][0];
		  node_index[i2][1]=envelope[node_index[i1][0]]->num_nodes-1;
		} /* First variable belongs to an envelope; adding the second variable to the same envelope. */
	      else 
		if( node_index[i2][0]!=-1 )
		  { /* Second variable belongs to an envelope; adding the first variable to the same envelope. */
		    envelope[node_index[i2][0]]->edge[envelope[node_index[i2][0]]->num_edges]=e;
		    envelope[node_index[i2][0]]->num_edges++;
		    envelope[node_index[i2][0]]->node[envelope[node_index[i2][0]]->num_nodes]=i1;
		    envelope[node_index[i2][0]]->num_nodes++;
		    node_index[i1][0]=node_index[i2][0];
		    node_index[i1][1]=envelope[node_index[i2][0]]->num_nodes-1;
		  } /* Second variable belongs to an envelope; adding the first variable to the same envelope. */
		else
		  { /* Creating a new envelope */
		    envelope[num_envelopes]=new Envelope( dim );
		    envelope[num_envelopes]->num_nodes=2;
		    envelope[num_envelopes]->num_edges=1;
		    envelope[num_envelopes]->node[0]=i1;
		    envelope[num_envelopes]->node[1]=i2;
		    envelope[num_envelopes]->edge[0]=e;
		    node_index[i1][0]=num_envelopes;
		    node_index[i2][0]=num_envelopes;
		    node_index[i1][1]=0;
		    node_index[i2][1]=1;
		    num_envelopes++;
		  } /* Creating a new envelope */
	    } /* Both variables are missing */
	} /* First variable is missing */
    } /* For each of the edges */

  return( num_envelopes );
}

long Distribution::compute_envelopes_full() {
  long num_envelopes=0;

  long* node_index;

  /* Index variable(s) */
  long i,e,i1,i2;

  /* Initializing */
  node_index=new long[dim];
  for( i=0; i<dim; i++ )
    /* Nodes are not yet assigned to envelopes */
    node_index[i]=-1;

  for( e=0; e<num_edges; e++ ) 
    { /* For each of the edges */
      /* Shortcuts */
      i1=edge[e][0];
      i2=edge[e][1];

      if( node_index[i1]!=-1 )
	{ /* First variable belongs to an envelope; adding the second variable to the same envelope. */
	  envelope[node_index[i1]]->edge[envelope[node_index[i1]]->num_edges]=e;
	  envelope[node_index[i1]]->num_edges++;
	  envelope[node_index[i1]]->node[envelope[node_index[i1]]->num_nodes]=i2;
	  envelope[node_index[i1]]->num_nodes++;
	  node_index[i2]=node_index[i1];
	} /* First variable belongs to an envelope; adding the second variable to the same envelope. */
      else 
	if( node_index[i2]!=-1 )
	  { /* Second variable belongs to an envelope; adding the first variable to the same envelope. */
	    envelope[node_index[i2]]->edge[envelope[node_index[i2]]->num_edges]=e;
	    envelope[node_index[i2]]->num_edges++;
	    envelope[node_index[i2]]->node[envelope[node_index[i2]]->num_nodes]=i1;
	    envelope[node_index[i2]]->num_nodes++;
	    node_index[i1]=node_index[i2];
	  } /* Second variable belongs to an envelope; adding the first variable to the same envelope. */
	else
	  { /* Creating a new envelope */
	    envelope[num_envelopes]=new Envelope( dim );
	    envelope[num_envelopes]->num_nodes=2;
	    envelope[num_envelopes]->num_edges=1;
	    envelope[num_envelopes]->node[0]=i1;
	    envelope[num_envelopes]->node[1]=i2;
	    envelope[num_envelopes]->edge[0]=e;
	    node_index[i1]=num_envelopes;
	    node_index[i2]=num_envelopes;
	    num_envelopes++;
	  } /* Creating a new envelope */
    } /* For each edge */

  /* Making sure all nodes are assigned to some envelope */
  for( i=0; i<dim; i++ )
    if( node_index[i]==-1 )
      { /* Creating a single node envelope */
	envelope[num_envelopes]=new Envelope( 1 );
	envelope[num_envelopes]->num_nodes=1;
	envelope[num_envelopes]->num_edges=0;
	envelope[num_envelopes]->node[0]=i;
	node_index[i]=num_envelopes;
	num_envelopes++;
      } /* Creating a single node envelope */

  delete node_index;

  return( num_envelopes );
}

void Distribution::PostProcess() {
  /* Rearranging the parameters */

  long* visited;
  long best_edge;

  long** child;
  long* num_parents;
  long num_inserted;

  /* Index variable(s) */
  long i,j,e;

  /* Temporary variable(s) */
  double** temp_prob;
  double temp;

  switch( type ) {
  case DIST_FACTOR:
    for( i=0; i<dim; i++ )
      subdist[i]->PostProcess();
    break;
  case DIST_MIXTURE:
    for( i=0; i<dim; i++ )
      subdist[i]->PostProcess();
    break;
  case DIST_CLMIXTURE:
    for( i=0; i<dim; i++ )
      for( j=0; j<num_states; j++ )
	subdist2[i][j]->PostProcess();
    break;
  case DIST_CONDCHOWLIU:
    /* Need to rearrange the edges order to speed up the simulation */
    visited=new long[2*dim];
    for( i=0; i<dim; i++ )
      {
	visited[i]=1;
	visited[i+dim]=0;
      }

    for( e=0; e<num_edges-1; e++ )
      { /* Inserting the edge in position e */
	best_edge=e;
	for( i=e; i<num_edges; i++ )
	  if( visited[edge[i][0]] || visited[edge[i][1]] )
	    {
	      best_edge=i;
	      i=num_edges;
	    }

	visited[edge[best_edge][0]]=1;
	visited[edge[best_edge][1]]=1;

	/* Switch edge best_edge and e */
	if( e!=best_edge )
	  {
	    i=edge[e][0];
	    edge[e][0]=edge[best_edge][0];
	    edge[best_edge][0]=i;
	    i=edge[e][1];
	    edge[e][1]=edge[best_edge][1];
	    edge[best_edge][1]=i;

	    i=edge_rank[e];
	    edge_rank[e]=edge_rank[best_edge];
	    edge_rank[best_edge]=i;

	    temp_prob=edge_prob[e];
	    edge_prob[e]=edge_prob[best_edge];
	    edge_prob[best_edge]=temp_prob;
	    
	    temp=edge_MI[e];
	    edge_MI[e]=edge_MI[best_edge];
	    edge_MI[best_edge]=temp;
	  }
      } /* Inserting the edge in position e */

    delete visited;

    break;
    //  case DIST_CLMIXTURE:
  case DIST_CHOWLIU:
    //case DIST_CHOWLIU_MDL:
    /* Need to rearrange the edges order to speed up the simulation */
    visited=new long[dim];
    for( i=0; i<dim; i++ )
      visited[i]=0;

    for( e=0; e<num_edges-1; e++ )
      { /* Inserting the edge in position e */
	best_edge=e;
	for( i=e; i<num_edges; i++ )
	  if( visited[edge[i][0]] || visited[edge[i][1]] )
	    {
	      best_edge=i;
	      i=num_edges;
	    }

	visited[edge[best_edge][0]]=1;
	visited[edge[best_edge][1]]=1;

	/* Switch edge best_edge and e */
	if( e!=best_edge )
	  {
	    i=edge[e][0];
	    edge[e][0]=edge[best_edge][0];
	    edge[best_edge][0]=i;
	    i=edge[e][1];
	    edge[e][1]=edge[best_edge][1];
	    edge[best_edge][1]=i;

	    i=edge_rank[e];
	    edge_rank[e]=edge_rank[best_edge];
	    edge_rank[best_edge]=i;

	    temp_prob=edge_prob[e];
	    edge_prob[e]=edge_prob[best_edge];
	    edge_prob[best_edge]=temp_prob;
	    
	    temp=edge_MI[e];
	    edge_MI[e]=edge_MI[best_edge];
	    edge_MI[best_edge]=temp;
	  }
      } /* Inserting the edge in position e */

    delete visited;

    break;
  case DIST_BN_ME_BIVAR:
  case DIST_BN_CME_BIVAR:
    /* Determining the order of nodes to simulate the data */

    /* Partial order sorting */

    /* Building a matrix of children */
    num_parents=new long[dim];
    for( i=0; i<dim; i++ )
      num_parents[i]=0;

    child=new long*[dim];
    for( i=0; i<dim; i++ )
      child[i]=new long[dim];
    for( i=0; i<dim; i++ )
      for( j=0; j<dim; j++ )
	child[i][j]=0;

    for( i=0; i<dim; i++ )
      for( j=1; j<num_features[i]; j++ )
	{
	  if( feature_index[i][j]<dim )
	    {
	      child[feature_index[i][j]][i]=1;
	      num_parents[i]++;
	    }
	}

    visited=new long[dim];
    for( i=0; i<dim; i++ )
      visited[i]=0;

    /* Building an ordering */
    num_inserted=0;
    while( num_inserted<dim )
      {
	/* Looking for the non-visited node with 0 parents */
	i=0;
	while( visited[i] || num_parents[i]>0 )
	  i++;

	/* Adding i to the list */
	sim_order[num_inserted]=i;
	num_inserted++;

	visited[i]=1;
	for( j=0; j<dim; j++ )
	  if( child[i][j] )
	    num_parents[j]--;	
      }

    delete visited;
    delete num_parents;
    for( i=0; i<dim; i++ )
      delete child[i];
    delete child;

    break;
  default:;
  }
}

void Distribution::InformationGain( long num_points, long** data, double* w, double norm_const, double* sw, 
				    long* candidate, double* gain, double* alpha, long index ) {
  /* Implementation of the Information Gain score */
  long max_num_features;
  
  double* param_sum_f;
  double* p;
  
  double d1,d2; // First and second derivatives

  long iteration;

  /* Index variable(s) */
  long n,t,i;

  /* Determining the dimensionality of the candidate array */
  if( type==DIST_BN_ME_BIVAR )
    max_num_features=dim;
  else if( type==DIST_BN_CME_BIVAR )
    max_num_features=2*dim;

  /* Initializing the array of linear sums for each datum */
  param_sum_f=new double[num_points];
  for( n=0; n<num_points; n++ )
    param_sum_f[n]=sigma[index][0];

  /* Allocating the array of current probability values */
  p=new double[num_points];

  /* Computing the linear sums in the exponents for each vectors in the data set */
  for( n=0; n<num_points; n++ )
    for( i=1; i<num_features[index]; i++ )
      if( data[n][feature_index[index][i]] )
	param_sum_f[n]+=sigma[index][i];

  for( i=0; i<max_num_features; i++ )
    if( candidate[i] )
      { /* Computing Information Gain and corresponding coefficient for each of candidate features */
	alpha[i]=0;

	iteration=0;
	d1=norm_const; // Not to trigger compiler warnings
	while( (iteration==0 || fabs(d1)/norm_const>NR_EPSILON) && iteration<MAX_NR_ITERATIONS )
	  { /* Newton-Raphson */
	    iteration++;

	    /* Computing the probabilities P(1|params,new parameter) */
	    for( n=0; n<num_points; n++ )
	      if( data[n][i] )
		p[n]=1.0-1.0/(1.0+exp(alpha[i]+param_sum_f[n]));

	    /* Computing the first derivative */
	    d1=sw[i];
	    for( n=0; n<num_points; n++ )
	      if( data[n][i] )
		d1-=w[n]*p[n];

	    /* Computing the second derivative */
	    d2=0.0;
	    for( n=0; n<num_points; n++ )
	      if( data[n][i] )
		d2-=w[n]*p[n]*(1.0-p[n]);

	    /* Updating the parameter */
	    alpha[i]-=d1/d2;  
	  } /* Newton-Raphson */

	/* Computing gain */
	gain[i]=alpha[i]*sw[i];
	for( n=0; n<num_points; n++ )
	  if( data[n][i] )
	    gain[i]-=w[n]*(log(1.0+exp(alpha[i]+param_sum_f[n]))-log(1.0+exp(param_sum_f[n])));
	gain[i]/=norm_const;

      } /* Computing Information Gain and corresponding coefficient for each of candidate features */
  
  /* Deallocating the arrays */
  delete p;
  delete param_sum_f;

  return;
}

double learn_univariate_conditional_maxent( long num_points, long** data, double* w, double norm_const, long index, 
					    double* sw,long num_features, double* delta, long* feature ) {

  double wll=0.0;                        // Weighted log-likelihood
  double prev_wll=NEG_INF;               // Previous value of weighted log-likelihood 
  double* delta_sum_f;                   // Linear sums in the exponents (contribution from base)
  long iteration=0;                      // Iteration of the conjugate gradient algorithm
  double *gradient, *old_gradient;       // Gradient of the weighted log-likelihood
  double* xi;                            // Direction of the ascent
  double gamma, gamma_pr, gamma_fr;      // Coefficient used in determining the direction of the ascent
  double* xi_sum_f;                      // Linear sums in the exponents (contribution from the direction)
  double nu;                             // The solution of the linear search problem
  long iteration_nr;                     // Newton-Raphson iteration index
  double d1,d2;                          // First and second derivative of the log-likelihood
  double* p;                             // Probability P(1|other variables, parameters)

  /* Index variable(s) */
  long n,i;

  /* Temporary variables */
  double gg, og, oo;

  /*
  for( i=0; i<num_features; i++ )
    delta[i]=0.0;
  */

  /* Initializing the arrays of linear sums for each datum */
  delta_sum_f=new double[num_points];
  for( n=0; n<num_points; n++ )
    delta_sum_f[n]=delta[0];

  xi_sum_f=new double[num_points];

  xi=new double[num_features];

  /* Allocating the array of current probability values */
  p=new double[num_points];

  /* Precomputing values */
  /* Computing the linear sums in the exponents */
  for( n=0; n<num_points; n++ )
    for( i=1; i<num_features; i++ )
      if( data[n][feature[i]] )
	delta_sum_f[n]+=delta[i];

  /* Computing weighted log-likelihood */
  for( n=0; n<num_points; n++ )
    if( data[n][index] )
      wll+=w[n]*(delta_sum_f[n]-log(1.0+exp(delta_sum_f[n])));
    else
      wll-=w[n]*log(1.0+exp(delta_sum_f[n]));

  old_gradient=NULL; gradient=NULL; // Not to trigger compiler warnings
  while( (wll-prev_wll)/norm_const>cg_epsilon )
    {
      prev_wll=wll;

      /* Computing the gradient */
      if( iteration>0 )
	{
	  if( iteration>1 )
	    delete old_gradient;
	  old_gradient=gradient;
	}

      gradient=new double[num_features];
      for( i=0; i<num_features; i++ )
	gradient[i]=sw[feature[i]];

      for( n=0; n<num_points; n++ )
	gradient[0]-=w[n]*exp(delta_sum_f[n])/(1.0+exp(delta_sum_f[n]));
      gradient[0]/=norm_const;

      for( i=1; i<num_features; i++ )
	{
	  for( n=0; n<num_points; n++ )
	    if( data[n][feature[i]] )
	      gradient[i]-=w[n]*exp(delta_sum_f[n])/(1.0+exp(delta_sum_f[n]));
	  gradient[i]/=norm_const;
	}

      /* Computing the new direction */
      if( iteration==0 )
	for( i=0; i<num_features; i++ )
	  xi[i]=-gradient[i];
      else
	{
	  gg=0.0;
	  og=0.0;
	  oo=0.0;
	  for( i=0; i<num_features; i++ )
	    {
	      gg+=gradient[i]*gradient[i];
	      og+=gradient[i]*old_gradient[i];
	      oo+=old_gradient[i]*old_gradient[i];
	    }

	  gamma_pr=(gg-og)/oo;  // Polak-Ribiere
	  gamma_fr=gg/oo;       // Fletcher-Reeves

	  if( gamma_pr<-gamma_fr )
	    gamma=-gamma_fr;
	  else if( gamma_pr > gamma_fr )
	    gamma=gamma_fr;
	  else
	    gamma=gamma_pr;

	  for( i=0; i<num_features; i++ )
	    xi[i]=gradient[i]-gamma*old_gradient[i];
	}

      /* Line search optimization algorithm */

      /* Pre-computing commonly used values */
      /* Exponent contribution from the new direction */
      for( n=0; n<num_points; n++ )
	xi_sum_f[n]=xi[0];

      for( n=0; n<num_points; n++ )
	for( i=1; i<num_features; i++ )
	  if( data[n][feature[i]] )
	    xi_sum_f[n]+=xi[i];

      nu=0.0;
      iteration_nr=0;

      /* Newton-Raphson */
      d1=norm_const;   // Not to trigger the compiler warning
      while( (iteration_nr==0 || fabs(d1)/norm_const>NR_EPSILON) && iteration_nr<MAX_NR_ITERATIONS )
	{
	  iteration_nr++;
     
	  /* First derivative */
	  d1=0.0;
	  for( i=0; i<num_features; i++ )
	    d1+=sw[feature[i]]*xi[i];
	  
	  for( n=0; n<num_points; n++ )
	    {
	      p[n]=exp(delta_sum_f[n]+nu*xi_sum_f[n])/(1.0+exp(delta_sum_f[n]+nu*xi_sum_f[n]));
	      d1-=w[n]*xi_sum_f[n]*p[n];
	    }

	  d2=0.0;
	  for( n=0; n<num_points; n++ )
	    d2-=w[n]*xi_sum_f[n]*xi_sum_f[n]*p[n]*(1.0-p[n]);
     
	  nu-=d1/d2;
	}

      /* Updating the parameters */
      for( i=0; i<num_features; i++ )      
	delta[i]+=nu*xi[i];
      
      /* Updating sums of features */
      for( n=0; n<num_points; n++ )
	delta_sum_f[n]+=nu*xi_sum_f[n];
  
      wll=0.0;
      /* Computing weighted log-likelihood */
      for( n=0; n<num_points; n++ )
	if( data[n][index] )
	  wll+=w[n]*(delta_sum_f[n]-log(1.0+exp(delta_sum_f[n])));
	else
	  wll-=w[n]*log(1.0+exp(delta_sum_f[n]));
      
      iteration++;
    }

  /* Deallocating the arrays */
  if( iteration>0 )
    {
      delete gradient;
      if( iteration>1 )
	delete old_gradient;
    }

  delete p;
  delete xi;
  delete xi_sum_f;
  delete delta_sum_f;

  return( wll );
}

void normalize_logistic( double* log_un_prob, double* prob, long num_states, long first ) {
  /* Calculating normalized softmax function given unnormalized exponent expressions */

  long min_index;            // Which entry to start

  /* Index variable(s) */
  long i;

  /* Temporary variable(s) */
  double sum;
  double max;
  
  if( first )
    /* First exponent is assumed 0.0 */
    max=0.0;
  else
    max=NEG_INF;
      
  for( i=first; i<num_states; i++ )
    if( max<log_un_prob[i] )
      max=log_un_prob[i];

  if( first )
    sum=exp(-max);      
  else
    sum=0.0;

  for( i=first; i<num_states; i++ )
    {
      prob[i]=exp(log_un_prob[i]-max);
      sum+=prob[i];
    }

  /* Normalizing */
  for( i=first; i<num_states; i++ )
    prob[i]/=sum;
  
  return;
}

Distribution* Distribution::Project() {
  /* Converting HMM models with inputs to models without inputs */

  Distribution* transition;

  /* Index variable(s) */
  long i,j;

  /* Temporary variable(s) */
  double sum;

  /* Converting the structures */
  switch( type ) {
  case DIST_TRANSLOGISTIC:
    /* Allocating collapsed transition distribution */
    transition=new Distribution( DIST_CONDBERNOULLI, num_states, 1 );

    /* Retrieving the distribution values from the base transition distribution */
    sum=0.0;
    for( i=0; i<num_states; i++ )
      { /* First state value i */
	transition->state_prob[i]=exp(lambda[i]);
	sum+=transition->state_prob[i];
      } /* First state value i */

    for( i=0; i<num_states; i++ )
      transition->state_prob[i]/=sum;

    for( i=0; i<num_states; i++ )
      { /* Previous state i */
	sum=0.0;
	for( j=0; j<num_states; j++ )
	  {
	    transition->cond_state_prob[i][j]=exp(sigma[i][j]);
	    sum+=transition->cond_state_prob[i][j];
	  }

	/* Normalizing */
	for( j=0; j<num_states; j++ )
	  transition->cond_state_prob[i][j]/=sum;
      } /* Previous state i */

    break;
  case DIST_LOGISTIC:
    /* Allocating collapsed transition distribution */
    transition=new Distribution( DIST_BERNOULLI, num_states, 1 );

    /* Retrieving the distribution values from the base transition distribution */
    sum=0.0;
    for( i=0; i<num_states; i++ )
      { /* First state value i */
	transition->state_prob[i]=exp(lambda[i]);
	sum+=transition->state_prob[i];
      } /* First state value i */

    for( i=0; i<num_states; i++ )
      transition->state_prob[i]/=sum;

    break;
  default:
    transition=NULL;
  }
  
  return( transition );
}

void Distribution::Expand( Distribution* transition ) {
  /* Expands a model without inputs to a model with inputs of dimension d */

  /* Index variable(s) */
  long i,j;

  /* Temporary variable(s) */
  double temp;

  /* Converting the structures */
  switch( type ) {
  case DIST_CONDBERNOULLI:
    /* Converting parameter values */
    
    /* First entry in the sequence */
    temp=log(state_prob[0]);
    transition->lambda[0]=0.0;
    for( i=1; i<num_states; i++ )
      transition->lambda[i]=log(state_prob[i])-temp;

    for( i=0; i<num_states; i++ )
      { /* Transition from state i */
	temp=log(cond_state_prob[i][0]);
	transition->sigma[i][0]=0.0;
	for( j=1; j<num_states; j++ )
	  transition->sigma[i][j]=log(cond_state_prob[i][j])-temp;
      } /* Transition from state i */
    
    /* Zeroing out the linear term */
    for( i=0; i<num_states; i++ )
      for( j=0; j<transition->dim; j++ )
	transition->rho[i][j]=0.0;

    break;
  case DIST_BERNOULLI:
    /* Converting parameter values */

    /* First entry in the sequence */
    temp=log(state_prob[0]);
    transition->lambda[0]=0.0;
    for( i=1; i<num_states; i++ )
      transition->lambda[i]=log(state_prob[i])-temp;

    /* Zeroing out the linear term */
    for( i=0; i<num_states; i++ )
      for( j=0; j<transition->dim; j++ )
	transition->rho[i][j]=0.0;

    break;
  default:;
  }

  return;
}

























/* Code acknowledgement */
/* gammaln, digamma, and trigammma code by Thomas Minka */
/* http://www.stat.cmu.edu/~minka/papers/dirichlet/minka-digamma.zip */

/* Logarithm of the gamma function.

   References:

   1) W. J. Cody and K. E. Hillstrom, 'Chebyshev Approximations for
      the Natural Logarithm of the Gamma Function,' Math. Comp. 21,
      1967, pp. 198-203.

   2) K. E. Hillstrom, ANL/AMD Program ANLC366S, DGAMMA/DLGAMA, May,
      1969.
 
   3) Hart, Et. Al., Computer Approximations, Wiley and sons, New
      York, 1968.

   From matlab/gammaln.m
*/
double gammaln(double x)
{
  double result, y, xnum, xden;
  int i;
  static double d1 = -5.772156649015328605195174e-1;
  static double p1[] = { 
    4.945235359296727046734888e0, 2.018112620856775083915565e2, 
    2.290838373831346393026739e3, 1.131967205903380828685045e4, 
    2.855724635671635335736389e4, 3.848496228443793359990269e4, 
    2.637748787624195437963534e4, 7.225813979700288197698961e3 
  };
  static double q1[] = {
    6.748212550303777196073036e1, 1.113332393857199323513008e3, 
    7.738757056935398733233834e3, 2.763987074403340708898585e4, 
    5.499310206226157329794414e4, 6.161122180066002127833352e4, 
    3.635127591501940507276287e4, 8.785536302431013170870835e3
  };
  static double d2 = 4.227843350984671393993777e-1;
  static double p2[] = {
    4.974607845568932035012064e0, 5.424138599891070494101986e2, 
    1.550693864978364947665077e4, 1.847932904445632425417223e5, 
    1.088204769468828767498470e6, 3.338152967987029735917223e6, 
    5.106661678927352456275255e6, 3.074109054850539556250927e6
  };
  static double q2[] = {
    1.830328399370592604055942e2, 7.765049321445005871323047e3, 
    1.331903827966074194402448e5, 1.136705821321969608938755e6, 
    5.267964117437946917577538e6, 1.346701454311101692290052e7, 
    1.782736530353274213975932e7, 9.533095591844353613395747e6
  };
  static double d4 = 1.791759469228055000094023e0;
  static double p4[] = {
    1.474502166059939948905062e4, 2.426813369486704502836312e6, 
    1.214755574045093227939592e8, 2.663432449630976949898078e9, 
    2.940378956634553899906876e10, 1.702665737765398868392998e11, 
    4.926125793377430887588120e11, 5.606251856223951465078242e11
  };
  static double q4[] = {
    2.690530175870899333379843e3, 6.393885654300092398984238e5, 
    4.135599930241388052042842e7, 1.120872109616147941376570e9, 
    1.488613728678813811542398e10, 1.016803586272438228077304e11, 
    3.417476345507377132798597e11, 4.463158187419713286462081e11
  };
  static double c[] = {
    -1.910444077728e-03, 8.4171387781295e-04, 
    -5.952379913043012e-04, 7.93650793500350248e-04, 
    -2.777777777777681622553e-03, 8.333333333333333331554247e-02, 
    5.7083835261e-03
  };
  static double a = 0.6796875;

  if((x <= 0.5) || ((x > a) && (x <= 1.5))) {
    if(x <= 0.5) {
      result = -log(x);
      /*  Test whether X < machine epsilon. */
      if(x+1 == 1) {
	return result;
      }
    }
    else {
      result = 0;
      x = (x - 0.5) - 0.5;
    }
    xnum = 0;
    xden = 1;
    for(i=0;i<8;i++) {
      xnum = xnum * x + p1[i];
      xden = xden * x + q1[i];
    }
    result += x*(d1 + x*(xnum/xden));
  }
  else if((x <= a) || ((x > 1.5) && (x <= 4))) {
    if(x <= a) {
      result = -log(x);
      x = (x - 0.5) - 0.5;
    }
    else {
      result = 0;
      x -= 2;
    }
    xnum = 0;
    xden = 1;
    for(i=0;i<8;i++) {
      xnum = xnum * x + p2[i];
      xden = xden * x + q2[i];
    }
    result += x*(d2 + x*(xnum/xden));
  }
  else if(x <= 12) {
    x -= 4;
    xnum = 0;
    xden = -1;
    for(i=0;i<8;i++) {
      xnum = xnum * x + p4[i];
      xden = xden * x + q4[i];
    }
    result = d4 + x*(xnum/xden);
  }
  /*  X > 12  */
  else {
    y = log(x);
    result = x*(y - 1) - y*0.5 + .9189385332046727417803297;
    x = 1/x;
    y = x*x;
    xnum = c[6];
    for(i=0;i<6;i++) {
      xnum = xnum * y + c[i];
    }
    xnum *= x;
    result += xnum;
  }
  return result;
}

/* The digamma function is the derivative of gammaln.

   Reference:
    J Bernardo,
    Psi ( Digamma ) Function,
    Algorithm AS 103,
    Applied Statistics,
    Volume 25, Number 3, pages 315-317, 1976.

    From http://www.psc.edu/~burkardt/src/dirichlet/dirichlet.f
    (with modifications for negative numbers and extra precision)
*/
double digamma(double x)
{
  //  static const double neginf = -1.0/0,
  static const double neginf = -1e+300,
    c = 12,
    d1 = -0.57721566490153286,
    d2 = 1.6449340668482264365, /* pi^2/6 */
    s = 1e-6,
    s3 = 1./12,
    s4 = 1./120,
    s5 = 1./252,
    s6 = 1./240,
    s7 = 1./132,
    s8 = 691/32760,
    s9 = 1/12,
    s10 = 3617/8160;
  double result;
#if 0
  static double cache_x = 0;
  static int hits = 0, total = 0;
  total++;
  if(x == cache_x) {
    hits++;
  }
  if(total % 1000 == 1) {
    printf("hits = %d, total = %d, hits/total = %g\n", hits, total, 
	   ((double)hits)/total);
  }
  cache_x = x;
#endif
  /* Illegal arguments */
  if((x == neginf) || isnan(x)) {
    return( neginf );
    //    return nan("");
    //    return 0.0/0;
  }
  /* Singularities */
  if((x <= 0) && (floor(x) == x)) {
    return neginf;
  }
  /* Negative values */
  /* Use the reflection formula (Jeffrey 11.1.6):
   * digamma(-x) = digamma(x+1) + pi*cot(pi*x)
   *
   * This is related to the identity
   * digamma(-x) = digamma(x+1) - digamma(z) + digamma(1-z)
   * where z is the fractional part of x
   * For example:
   * digamma(-3.1) = 1/3.1 + 1/2.1 + 1/1.1 + 1/0.1 + digamma(1-0.1)
   *               = digamma(4.1) - digamma(0.1) + digamma(1-0.1)
   * Then we use
   * digamma(1-z) - digamma(z) = pi*cot(pi*z)
   */
  if(x < 0) {
    return digamma(1-x) + M_PI/tan(-M_PI*x);
  }
  /* Use Taylor series if argument <= S */
  if(x <= s) return d1 - 1/x + d2*x;
  /* Reduce to digamma(X + N) where (X + N) >= C */
  result = 0;
  while(x < c) {
    result -= 1/x;
    x++;
  }
  /* Use de Moivre's expansion if argument >= C */
  /* This expansion can be computed in Maple via asympt(Psi(x),x) */
  if(x >= c) {
    double r = 1/x;
    result += log(x) - 0.5*r;
    r *= r;
    result -= r * (s3 - r * (s4 - r * (s5 - r * (s6 - r * s7))));
  }
  return result;
}

/* The trigamma function is the derivative of the digamma function.

   Reference:

    B Schneider,
    Trigamma Function,
    Algorithm AS 121,
    Applied Statistics, 
    Volume 27, Number 1, page 97-99, 1978.

    From http://www.psc.edu/~burkardt/src/dirichlet/dirichlet.f
    (with modification for negative arguments and extra precision)
*/
double trigamma(double x)
{
  //  double neginf = -1.0/0,
  double neginf = -1e+300,
    small = 1e-4,
    large = 8,
    c = 1.6449340668482264365, /* pi^2/6 = Zeta(2) */
    c1 = -2.404113806319188570799476,  /* -2 Zeta(3) */
    b2 =  1./6,
    b4 = -1./30,
    b6 =  1./42,
    b8 = -1./30,
    b10 = 5./66;
  double result;
  /* Illegal arguments */
  if((x == neginf) || isnan(x)) {
    //    return 0.0/0;
    //    return nan("");
    return( neginf );
  }
  /* Singularities */
  if((x <= 0) && (floor(x) == x)) {
    return -neginf;
  }
  /* Negative values */
  /* Use the derivative of the digamma reflection formula:
   * -trigamma(-x) = trigamma(x+1) - (pi*csc(pi*x))^2
   */
  if(x < 0) {
    result = M_PI/sin(-M_PI*x);
    return -trigamma(1-x) + result*result;
  }
  /* Use Taylor series if argument <= small */
  if(x <= small) {
    return 1/(x*x) + c + c1*x;
  }
  result = 0;
  /* Reduce to trigamma(x+n) where ( X + N ) >= B */
  while(x < large) {
    result += 1/(x*x);
    x++;
  }
  /* Apply asymptotic formula when X >= B */
  /* This expansion can be computed in Maple via asympt(Psi(1,x),x) */
  if(x >= large) {
    double r = 1/(x*x);
    result += 0.5*r + (1 + r*(b2 + r*(b4 + r*(b6 + r*(b8 + r*b10)))))/x;
  }
  return result;
}

double exp_dist( double x, double alpha ) {
  return( alpha*exp(-alpha*x) );
}

double gamma_dist( double x, double alpha, double beta ) {
  return( exp((alpha-1.0)*log(x)-beta*x+alpha*log(beta)-gammaln(alpha)) );
}

double lognormal_dist( double x, double M, double S2) {
  return( exp(-(log(x)-M)*(log(x)-M)/(2.0*S2))/(x*sqrt(2*M_PI*S2)) );
}
